/* VDP routine for the TI-99/4a Emulator */
/* Ami99VDP by Mike Brent */

/* TI Palette */

#define BORDER 40
/* index of color of border to emulate call screen() */
#define TIPALETTE 41
/* first index of a fixed copy of the TI Palette (used for sprites) */

RGB TIPAL[16];
PALETTE vdp_pal;

int red[16]={ 0,0,8,24,8,16,40,16,56,56,48,48,8,48,40,56 };
int green[16]= { 0,0,48,56,8,24,8,48,8,24,48,48,32,16,40,56 };
int blue[16]= { 0,0,8,24,56,56,8,56,8,24,8,32,8,40,40,56 };

BITMAP *VDPBUFFER, *SPRITEBUFFER, *SCREENBUFFER;

int SIT, CT, PDT, SAL, SDT;
int SITsize, CTsize, PDTsize, SALsize, SDTsize;

int col = 41;
int i1,i2;

Word p_add;

Byte chr,fgcol,bgcol,x;

int Video_Debug;

unsigned int ad;

FILE *fp;

void gettables()
{
  if (v9938) {
	// default table base addresses
	/* Screen Image Table / Pattern Name Table */
	SIT=(VDPREG[2]<<10)&0x1ffff;
	/* Colour Table */
	CT=((VDPREG[3]<<6)|(VDPREG[10]<<14))&0x1ffff;
	/* Pattern Descriptor Table / Pattern Generator Table */
	PDT=(VDPREG[4]<<11)&0x1ffff;
	/* Sprite Attribute List / Table */
	SAL=((VDPREG[5]<<7)|(VDPREG[11]<<15))&0x1ffff;
	/* Sprite Descriptor Table / Sprite Pattern Generator Table */
	SDT=(VDPREG[6]<<11)&0x1ffff;

	// default sizes (G1)
	SITsize = 0x300;
	CTsize = 0x20;
	PDTsize = 0x800;
	SALsize = 0x80;
	SDTsize = 0x400;

	// corrections
	switch (mode) {
	case G2:
	case G3:					/* FIXME: sprite mode 2 */
	  CTsize = 0x1800;
	  PDTsize = 0x1800;
	  break;
	case G4:					/* FIXME: sprite mode 2 */
	case G5:					/* FIXME: sprite mode 2 */
	  SIT &= 0x18000;
	  SAL &= 0x1fc00;
	  PDTsize = 0x6a00;
	  SITsize = CTsize = 0;		/* bitmapped */
	  SALsize = SDTsize = 0;	/* FIXME: no sprites */
	  break;
	case G6:					/* FIXME: sprite mode 2 */
	case G7:					/* FIXME: sprite mode 2 */
	  SIT = (VDPREG[2]<<11) & 0x10000;
	  SAL &= 0x1fc00;
	  PDTsize = 0xd400;
	  SITsize = CTsize = 0;		/* bitmapped */
	  SALsize = SDTsize = 0;	/* FIXME: no sprites */
	  break;
	case T1:
	  SITsize = 0x3c0;
	  CTsize = 0;
	  SALsize = SDTsize = 0;	/* no sprites */
	  break;
	case T2:
	  SIT &= 0x1f000;
	  CT &= 0x1fe00;
	  SITsize = 0x870;
	  CTsize = 0x10e;
	  SALsize = SDTsize = 0;	/* no sprites */
	  break;
	default:
	  break;
	} /* switch mode */
#ifdef v9938debug
	if (SIT>0x3fff) fprintf(stderr, "SIT 0x%4X\n", SIT);
	if (CT>0x3fff) fprintf(stderr, "CT 0x%4X\n", CT);
	if (PDT>0x3fff) fprintf(stderr, "PDT 0x%4X\n", PDT);
#endif
  } else {						/* !v9938) */
	/* Screen Image Table */
	SIT=(VDPREG[2]<<10)&0x3fff;
	/* Colour Table */
	CT=(VDPREG[3]<<6)&0x3fff;
	/* Pattern Descriptor Table */
	PDT=(VDPREG[4]<<11)&0x3fff;
	/* Sprite Attribute List */
	SAL=(VDPREG[5]<<7)&0x3fff;
	/* Sprite Descriptor Table */
	SDT=(VDPREG[6]<<11)&0x3fff;
  }
}

void VDPsetpalette() {
  vdp_pal[BORDER].r=TIPAL[0].r;
  vdp_pal[BORDER].g=TIPAL[0].g;
  vdp_pal[BORDER].b=TIPAL[0].b;

  switch (mode) {
  case G1: {
	int i1, i2, t, t1, t2;
	for (i1=0; i1<32; i1++) {
	  t=VDP[CT+i1];
	  t1=t&0xf;
	  t2=t>>4;
	  i2=i1+128;
	  vdp_pal[i1].r=TIPAL[t1].r;
	  vdp_pal[i1].g=TIPAL[t1].g;
	  vdp_pal[i1].b=TIPAL[t1].b;
	  vdp_pal[i2].r=TIPAL[t2].r;
	  vdp_pal[i2].g=TIPAL[t2].g;
	  vdp_pal[i2].b=TIPAL[t2].b;
	}
  }
  break;
  case G2:
  case G3:
  case G4:
  case G5:
  case G6:
	for (i1=0; i1<16; i1++) {	
	  vdp_pal[i1].r=TIPAL[i1].r;
	  vdp_pal[i1].g=TIPAL[i1].g;
	  vdp_pal[i1].b=TIPAL[i1].b;
	}
	break;
  case G7:
	for (i1=0; i1<256; i1++) {
	  vdp_pal[i1].r = (i1&0x1c)<<1;
	  vdp_pal[i1].g = (i1&0xe0)>>2;
	  vdp_pal[i1].b = (i1&0x03)<<4;
	}
	break;
  case T1: {
	int i1, i2, t, t1, t2;
	i1=0;
	t=VDPREG[7];
	t1=t&0xf;
	t2=t>>4;
	i2=i1+128;
	vdp_pal[i1].r=TIPAL[t1].r;
	vdp_pal[i1].g=TIPAL[t1].g;
	vdp_pal[i1].b=TIPAL[t1].b;
	vdp_pal[i2].r=TIPAL[t2].r;
	vdp_pal[i2].g=TIPAL[t2].g;
	vdp_pal[i2].b=TIPAL[t2].b;
  }
  break;
  case T2: {
	int t=VDPREG[7];
	int t1=t&0xf;
	int t2=t>>4;
	vdp_pal[0].r=TIPAL[t1].r;
	vdp_pal[0].g=TIPAL[t1].g;
	vdp_pal[0].b=TIPAL[t1].b;
	vdp_pal[1].r=TIPAL[t2].r;
	vdp_pal[1].g=TIPAL[t2].g;
	vdp_pal[1].b=TIPAL[t2].b;
	t=VDPREG[12];					/* blink color */
	t1=t&0xf;
	t2=t>>4;
	vdp_pal[2].r=TIPAL[t1].r;
	vdp_pal[2].g=TIPAL[t1].g;
	vdp_pal[2].b=TIPAL[t1].b;
	vdp_pal[3].r=TIPAL[t2].r;
	vdp_pal[3].g=TIPAL[t2].g;
	vdp_pal[3].b=TIPAL[t2].b;
  }
  break;
  default:
	break;
  }	/* switch mode */
  
  set_palette(vdp_pal);
  setpalette_needed = 0;
}

void VDPdisplay()
{

  if (v9938) {

	if (!(VDPREG[1] & 0x40)) {
	  // screen is blanked
	  rectfill(screen, 0, 0, wti + 39, hti + 39, BORDER);
	  return;
	}

	
	if (mode!=oldmode || debug!=olddebug) {	/* graphics mode or debug state changed */

	  gettables();

	  fprintf(stderr, "entering ");
	  
	  switch (mode) {
	  case G1:
		  fprintf(stderr, "Graphics 1");
		  break;		   
	  case G2:			   
		  fprintf(stderr, "Graphics 2");
		  break;		   
	  case G3:			   
		  fprintf(stderr, "Graphics 3");
		  break;		   
	  case G4:			   
		  fprintf(stderr, "Graphics 4");
		  break;		   
	  case G5:			   
		  fprintf(stderr, "Graphics 5");
		  break;		   
	  case G6:			   
		  fprintf(stderr, "Graphics 6");
		  break;		   
	  case G7:			   
		  fprintf(stderr, "Graphics 7");
		  break;		   
	  case T1:			   
		  fprintf(stderr, "Text 1");
		  break;		   
	  case T2:			   
		  fprintf(stderr, "Text 2");
		  break;		   
	  case MC:			   
		  fprintf(stderr, "Multicolor");
		  break;		   
	  default:			   
		  fprintf(stderr, "unknown 0x%2x", mode);
		  break;
	  } /* switch */

	  fprintf(stderr, " mode, debug o%s\n", debug ? "n" : "ff");

	  if (mode==T2 || mode==G5 || mode==G6 || mode==G7) {
		wti = 512;											/* width of TI screen */
		hti = (VDPREG[9]&0x80) ? 424 : 384;					/* interlace mode not yet supported */
		wscr = debug ? 1024 : 640;							/* width of window / hor. resolution */
		hscr = debug ? 768  : 480;
	  }	else {
		wti = 256;											/* width of TI screen */
		hti = (VDPREG[9]&0x80) ? 212 : 192;					/* interlace mode not yet supported */
		wscr = debug ? 640 : 320;							/* width of window / hor. resolution */
		hscr = debug ? 480 : 200;
	  }	

	  // change resolution if necessary
	  if (wscr != oldwscr) {
		if (set_gfx_mode(GFX_AUTODETECT, wscr, hscr, 0, 0) < 0)
		  fprintf(stderr, "Can't set %dx%dx8 graphics mode!", wscr, hscr);
		clear_keybuf();	/* maybe didnt get release event */
	  }

	  VDPsetpalette();
	  rectfill(screen, 0, 0, wscr-1, hscr-1, 0); /* clear screen (border!) */

	  oldmode = mode;
	  olddebug = debug;
	  oldwscr = wscr;

	} /* if mode!=oldmode... */

	if (setpalette_needed)
	  VDPsetpalette();

	switch (mode) {
	default:					/* all other modes not yet supported */
								/* use G1 so that at least something is visible... */
	case G1:
	  VDPgraphics();
	  break;
	case G2:
	case G3:												/* cant handle sprite mode 2 anyway :) */
	  VDPbitmap();
	  break;
	case G4:
	  VDPgraphics4(0, 0, wti-1, hti-1);
	  break;
	case G6:
	  VDPgraphics6(0, 0, wti-1, hti-1);
	  break;
	case G7:
	  VDPgraphics7(0, 0, wti-1, hti-1);
	  break;
	case T1:
	  VDPtext();
	  break;
	case T2:
	  VDPtext2();
	  break;
	}

    redraw_needed=0;
	
  } else {						/* !v9938 */

	/* determine which video mode to display */
	/* we only support Bitmap, Text, and Graphics, not Multicolor */
	/* right or not, I give Bitmap priority, then text, then graphics */

	int i1;

	gettables();
	VDPsetpalette();
	rebuild_patterns_needed = 1;

	if (!(VDPREG[1] & 0x40))
	{
		// screen is blanked
		for (i1=0; i1<16; i1++)
		{	
			vdp_pal[i1].r=TIPAL[i1].r;
			vdp_pal[i1].g=TIPAL[i1].g;
			vdp_pal[i1].b=TIPAL[i1].b;
		}
		vdp_pal[BORDER].r=TIPAL[0].r;
		vdp_pal[BORDER].g=TIPAL[0].g;
		vdp_pal[BORDER].b=TIPAL[0].b;

		set_palette(vdp_pal);

		rectfill(screen, 0, 0, 319, 199, 0);

		return;
	}

    redraw_needed=0;

	if (VDPREG[0] & 0x02)	// MODE BIT 3
	{
		VDPbitmap();
		return;
	}

	if (VDPREG[1] & 0x10)	// MODE BIT 2
	{
		VDPtext();
		return;
	}

	if (VDPREG[1] & 0x80)	// MODE BIT 1
	{
//		fail("Video mode MULTICOLOR not supported");
// just ignore this mode
	}

	VDPgraphics();
  }
}

void VDPmain()
{	
  set_color_depth(8);
  
  /* To increase compatibility with Windows DirectX. Should let user choose */
  /* will implement scaling after it works */

  if (debug)
  {  if (set_gfx_mode(GFX_DIRECTX,640,480,0,0)<0)
		  fail("Can't set 640x480x8 graphics mode!");
  }
  else
  {
	if (set_gfx_mode(GFX_DIRECTX,320,200,0,0)<0)
		fail("Can't set 320x200x8 graphics mode!");
  }

	set_display_switch_mode(SWITCH_BACKGROUND);

	set_display_switch_callback(SWITCH_IN, restore_video);

	/* VDPBUFFER points to a bitmap sized to be 256 characters by */
	/* 1 character high. It's used to draw the character set for  */
	/* blitting to the screen. We draw 256 chars instead of 768.  */
	/* Gives a bit of speed boost, & reduces duplication of work  */
  
	/* Now drawing only needed chars */

  VDPBUFFER=create_bitmap(4096,16);
  if (VDPBUFFER==NULL) fail("Can't create VDP workspace.");

  SPRITEBUFFER=create_bitmap(16,16);
  if (SPRITEBUFFER==NULL) fail("Can't create VDP Sprite workspace.");

  SCREENBUFFER=create_bitmap(256,192);
  if (SCREENBUFFER==NULL) fail("Can't create VDP screen buffer.");

  acquire_screen();
  clear_to_color(screen,BORDER);
  release_screen();

  for (i1=0; i1<16; i1++)
  { TIPAL[i1].r=red[i1];
    TIPAL[i1].b=blue[i1];
    TIPAL[i1].g=green[i1];
  }

  for (i1=0; i1<256; i1++)
  { vdp_pal[i1].r=0x80;
    vdp_pal[i1].g=0x80;
	vdp_pal[i1].b=0x80;
  }

  for (i1=0; i1<16; i1++)
  {	vdp_pal[TIPALETTE+i1].r=TIPAL[i1].r;
	vdp_pal[TIPALETTE+i1].g=TIPAL[i1].g;
	vdp_pal[TIPALETTE+i1].b=TIPAL[i1].b;
  }

  redraw_needed=1;

  /* screen should be set up and open */
  
  VDPREG[8]=0xff;   /* clear hacked-in keyboard buffer */

  /* When set, display the character set at the top of the screen */
  Video_Debug=0;
}


/*****************************************************************************
 *
 *	 9900dasm.c
 *	 TMS 9900 disassembler
 *
 *	 Copyright (c) 1998 John Butler, all rights reserved.
 *	 Based on 6502dasm.c 6502/65c02/6510 disassembler by Juergen Buchmueller
 *
 *	 - This source code is released as freeware for non-commercial purposes.
 *	 - You are free to use and redistribute this code in modified or
 *	   unmodified form, provided you list me in the credits.
 *	 - If you modify this source code, you must add a notice to each modified
 *	   source file that it has been changed.  If you're a nice person, you
 *	   will clearly mark each change too.  :)
 *	 - The author of this copywritten work reserves the right to change the
 *     terms of its usage and license at any time, including retroactively
 *   - This entire notice must remain in the source code.
 *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define RDOP(A) (CPU[A] << 8) + (CPU[(A+1) & 0xffff])
#define RDWORD(A) (CPU[A] << 8) + (CPU[(A+1) & 0xffff])

#define BITS_0to3	((OP>>12) & 0xf)
#define BITS_2to5	((OP>>10) & 0xf)
#define BITS_5to9	((OP>>6) & 0x1f)
#define BITS_3to7	((OP>>8) & 0x1f)
#define BITS_6to10	((OP>>5) & 0x1f)

#define BITS_0to1	((OP>>14) & 0x3)
#define BITS_0to4	((OP>>11) & 0x1f)
#define BITS_0to2	((OP>>13) & 0x7)
#define BITS_0to5	((OP>>10) & 0x3f)

#define	MASK	0x0000ffff
#define OPBITS(n1,n2)	((OP>>(15-(n2))) & (MASK>>(15-((n2)-(n1)))))

enum opcodes {
	_a=0,	_ab,	_c,		_cb,	_s,		_sb,	_soc,	_socb,	_szc,	_szcb,
	_mov,	_movb,	_coc,	_czc,	_xor,	_mpy,	_div,	_xop,	_b,		_bl,
	_blwp,	_clr,	_seto,	_inv,	_neg,	_abs,	_swpb,	_inc,	_inct,	_dec,
	_dect,	_x,		_ldcr,	_stcr,	_sbo,	_sbz,	_tb,	_jeq,	_jgt,	_jh,
	_jhe,	_jl,	_jle,	_jlt,	_jmp,	_jnc,	_jne,	_jno,	_joc,	_jop,
	_sla,	_sra,	_src,	_srl,	_ai,	_andi,	_ci,	_li,	_ori,	_lwpi,
	_limi,	_stst,	_stwp,	_rtwp,	_idle,	_rset,	_ckof,	_ckon,	_lrex,	_ill
};


static const char *token[]=
{
	"a",	"ab",	"c",	"cb",	"s",	"sb",	"soc",	"socb",	"szc",	"szcb",
	"mov",	"movb",	"coc",	"czc",	"xor",	"mpy",	"div",	"xop",	"b",	"bl",
	"blwp",	"clr",	"seto",	"inv",	"neg",	"abs",	"swpb",	"inc",	"inct",	"dec",
	"dect",	"x",	"ldcr",	"stcr",	"sbo",	"sbz",	"tb",	"jeq",	"jgt",	"jh",
	"jhe",	"jl",	"jle",	"jlt",	"jmp",	"jnc",	"jne",	"jno",	"joc",	"jop",
	"sla",	"sra",	"src",	"srl",	"ai",	"andi",	"ci",	"li",	"ori",	"lwpi",
	"limi",	"stst",	"stwp",	"rtwp",	"idle",	"rset",	"ckof",	"ckon",	"lrex",	"ill"
};


static const enum opcodes ops0to3[16]=
{
	_ill,	_ill,	_ill,	_ill,	_szc,	_szcb,	_s,		_sb,	/*0000-0111*/
	_c,		_cb,	_a,		_ab,	_mov,	_movb,	_soc,	_socb	/*1000-1111*/
};


static const enum opcodes ops2to5[16]=
{
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*0000-0111*/
	_coc,	_czc,	_xor,	_xop,	_ldcr,	_stcr,	_mpy,	_div	/*1000-1111*/
};


static const enum opcodes ops5to9[32]=
{
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*00000-00111*/
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*01000-01111*/
	_blwp,	_b,		_x,		_clr,	_neg,	_inv,	_inc,	_inct,	/*10000-10111*/
	_dec,	_dect,	_bl,	_swpb,	_seto,	_abs,	_ill,	_ill	/*11000-11111*/
};


static const enum opcodes ops3to7[32]=
{
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*00000-00111*/
	_sra,	_srl,	_sla,	_src,	_ill,	_ill,	_ill,	_ill,	/*01000-01111*/
	_jmp,	_jlt,	_jle,	_jeq,	_jhe,	_jgt,	_jne,	_jnc,	/*10000-10111*/
	_joc,	_jno,	_jl,	_jh,	_jop,	_sbo,	_sbz,	_tb		/*11000-11111*/
};


static const enum opcodes ops6to10[32]=
{
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*00000-00111*/
	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	_ill,	/*01000-01111*/
	_li,	_ai,	_andi,	_ori,	_ci,	_stwp,	_stst,	_lwpi,	/*10000-10111*/
	_limi,	_ill,	_idle,	_rset,	_rtwp,	_ckon,	_ckof,	_lrex	/*11000-11111*/
};

static int myPC;

static char *print_arg (int mode, int arg)
{
	static char temp[20];
	int	base;

	switch (mode)
	{
		case 0x0:	/* workspace register */
			sprintf (temp, "R%d", arg);
			break;
		case 0x1:	/* workspace register indirect */
			sprintf (temp, "*R%d", arg);
			break;
		case 0x2:	/* symbolic|indexed */
			base = RDWORD(myPC); myPC+=2;
			if (arg) 	/* indexed */
				sprintf (temp, "@>%04x(R%d)", base, arg);
			else		/* symbolic (direct) */
				sprintf (temp, "@>%04x", base);
			break;
		case 0x3:	/* workspace register indirect auto increment */
			sprintf (temp, "*R%d+", arg);
			break;
	}
	return temp;
}


/*****************************************************************************
 *	Disassemble a single command and return the number of bytes it uses.
 *****************************************************************************/
int Dasm9900 (char *buffer, int pc)
{
	int	OP, opc;
	int sarg, darg, smode, dmode;

	myPC = pc;
	OP = RDOP(myPC); myPC+=2;

	if ((opc = ops0to3[BITS_0to3]) != _ill)
	{
		smode = OPBITS(10,11);
		sarg = OPBITS(12,15);
		dmode = OPBITS(4,5);
		darg = OPBITS(6,9);

 		sprintf (buffer, "%-4s ", token[opc]);
		strcat (buffer, print_arg (smode, sarg));
		strcat (buffer, ",");
		strcat (buffer, print_arg (dmode, darg));
	}
	else if (BITS_0to1==0 && (opc = ops2to5[BITS_2to5]) != _ill)
	{
		smode = OPBITS(10,11);
		sarg = OPBITS(12,15);
		darg = OPBITS(6,9);

		if (darg==0 && (opc==_ldcr || opc==_stcr))
			darg = 16;

		if (opc==_xop || opc==_ldcr || opc==_stcr)
			sprintf (buffer, "%-4s %s,%d", token[opc], print_arg (smode, sarg), darg);
		else	/* _coc, _czc, _xor, _mpy, _div */
			sprintf (buffer, "%-4s %s,R%d", token[opc], print_arg (smode, sarg), darg);
	}
	else if (BITS_0to2==0 && (opc = ops3to7[BITS_3to7]) != _ill)
	{
		switch (opc)
		{
			case _sra: case _srl: case _sla: case _src:
				sarg = OPBITS(12,15);
				darg = OPBITS(8,11);

				sprintf (buffer, "%-4s R%d,%d", token[opc], sarg, darg);
				break;
			case _jmp: case _jlt: case _jle: case _jeq: case _jhe: case _jgt:
			case _jne: case _jnc: case _joc: case _jno: case _jl:  case _jh: case _jop:
				{
					signed char displacement;

					displacement = (signed char)OPBITS(8,15);
					sprintf (buffer, "%-4s >%04x", token[opc], 0xffff & (myPC + displacement * 2));
				}
				break;
			case _sbo: case _sbz: case _tb:
				{
					signed char displacement;

					displacement = (signed char)OPBITS(8,15);
					sprintf (buffer, "%-4s >%04x", token[opc], 0xffff & displacement);
				}
				break;
		}
	}
	else if (BITS_0to4==0 && (opc = ops5to9[BITS_5to9]) != _ill)
	{
		smode = OPBITS(10,11);
		sarg = OPBITS(12,15);

		sprintf (buffer, "%-4s %s", token[opc], print_arg (smode, sarg));
	}
	else if (BITS_0to5==0 && (opc = ops6to10[BITS_6to10]) != _ill)
	{
		switch (opc)
		{
			case _li:   case _ai:   case _andi: case _ori:  case _ci:
				darg = OPBITS(12,15);
				sarg = RDWORD(myPC); myPC+=2;

				sprintf (buffer, "%-4s R%d,>%04x", token[opc], darg, sarg);
				break;
			case _lwpi: case _limi:
				sarg = RDWORD(myPC); myPC+=2;

				sprintf (buffer, "%-4s >%04x", token[opc], sarg);
				break;
			case _stwp: case _stst:
				sarg = OPBITS(12,15);

				sprintf (buffer, "%-4s R%d", token[opc], sarg);
				break;
			case _idle: case _rset: case _rtwp: case _ckon: case _ckof: case _lrex:
				sprintf (buffer, "%-4s", token[opc]);
				break;
		}
	}
	else
		sprintf (buffer, "data >%04x", OP);

	return myPC - pc;
}

void draw_debug()
{
	// draw a debug screen as well as the primary
	// screen is assumed to be set to 640x480 now

	char buf1[256], buf2[256];
	int myPC, idx, i;
	int tmpPC;
	static int curline=0;

	if (max_ipf>0 || instruction_count<0) {
	  // trace simulation: draw only the current line and line below (which may be replaced)
	  acquire_screen();
	  myPC=PC;
	  tmpPC = myPC;
	  myPC += Dasm9900(buf2, myPC);
	  /*  	sprintf(buf1,"%04x   %-32s",tmpPC,buf2); */
	  sprintf(buf1, "  %04x   ", tmpPC, buf2);
	  for (i=0; i<3; i++)
		if (tmpPC+2*i < myPC)
		sprintf(buf1+9+5*i, "%04x  ", romword(tmpPC+2*i));
	  else
		sprintf(buf1+9+5*i, "      ");
	sprintf(buf1+25, "%-32s", buf2);
	textout(screen, font, buf1, wti+40, curline<<3, col);
	if (++curline > 59) curline=0;
	tmpPC = myPC;
	myPC += Dasm9900(buf2, myPC);
	sprintf(buf1, "  %04x n ", tmpPC, buf2);
	for (i=0; i<3; i++)
	  if (tmpPC+2*i < myPC)
		sprintf(buf1+9+5*i, "%04x  ", romword(tmpPC+2*i));
	  else
		sprintf(buf1+9+5*i, "      ");
	sprintf(buf1+25, "%-32s", buf2);
	textout(screen, font, buf1, wti+40, curline<<3, col);
	release_screen();
	}

	// draw the current register settings
	idx=0;

	sprintf(buf1,"PC     - %04x",PC);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	sprintf(buf1,"WP     - %04x",WP);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	sprintf(buf1,"ST     - %04x",ST);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	sprintf(buf1,"VDPST  - ");
	for (i=0; i < (v9938 ? 10 : 1); i++)
	  sprintf(buf1+9+3*i, "%02x ", VDPS[i]);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	sprintf(buf1,"VDPADD - %04x",VDPADD);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	sprintf(buf1,"GRMADD - %04x",GRMADD);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+40+(idx++)*8, col);
	release_screen();

	for (idx=0; idx<16; idx++) {
	  if (v9938) {
		if (idx<15)
		  sprintf(buf1,"R%-2d=%04x - VDPREG %-2d=%02x, %-2d=%02x, %-2d=%02x",
				  idx, romword(WP+idx*2),
				  idx, VDPREG[idx],
				  idx+16, VDPREG[idx+16],
				  idx+32, VDPREG[idx+32]
				  );
		else
		  sprintf(buf1,"R%-2d=%04x - VDPREG %-2d=%02x, %-2d=%02x",
				  idx, romword(WP+idx*2),
				  idx, VDPREG[idx],
				  idx+16, VDPREG[idx+16]
				  );
	  } else {												/* !v9938 */
		if (idx<8)
			sprintf(buf1,"R%-2d    - %04x     VDPREG %1d    - %02x",idx, romword(WP+idx*2), idx, VDPREG[idx]);
		else
			sprintf(buf1,"R%-2d    - %04x",idx, romword(WP+idx*2));
	  }
	  acquire_screen();
	  textout(screen, font, buf1, 0, hti+96+idx*8, col);
	  release_screen();
	}

	// realtime watches
	sprintf(buf1,"M%04X = ", debugadrm);
	for (i=0; i < 6; i++)
	  sprintf(buf1+8+5*i, "%04x ", romword(debugadrm+2*i));
	acquire_screen();
	textout(screen, font, buf1, 0, hti+232, col);
	release_screen();

	sprintf(buf1,"V%05X= ", debugadrv);
	for (i=0; i < 8; i++)
	  sprintf(buf1+8+3*i, "%02x ", VDP[debugadrv+i]);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+240, col);
	release_screen();

	sprintf(buf1,"G%04X = ", debugadrg);
	for (i=0; i < 8; i++)
	  sprintf(buf1+8+3*i, "%02x ", GROM[debugadrg+i]);
	acquire_screen();
	textout(screen, font, buf1, 0, hti+248, col);
	release_screen();


/*  	if (key[KEY_F1]) */
/*  	{	while (key[KEY_F1]); */
/*  		while ((!key[KEY_F1])&&(!key[KEY_F2])); */
/*  	} */

	if (max_ipf == 0) {										/* singlestepping: memory monitor */

	  if (key[KEY_0]) {
		debugadr = (debugadr << 4);
		key[KEY_0] = 0;
	  }
	  if (key[KEY_1]) {
		debugadr = (debugadr << 4) | 1;
		key[KEY_1] = 0;
	  }
	  if (key[KEY_2]) {
		debugadr = (debugadr << 4) | 2;
		key[KEY_2] = 0;
	  }
	  if (key[KEY_3]) {
		debugadr = (debugadr << 4) | 3;
		key[KEY_3] = 0;
	  }
	  if (key[KEY_4]) {
		debugadr = (debugadr << 4) | 4;
		key[KEY_4] = 0;
	  }
	  if (key[KEY_5]) {
		debugadr = (debugadr << 4) | 5;
		key[KEY_5] = 0;
	  }
	  if (key[KEY_6]) {
		debugadr = (debugadr << 4) | 6;
		key[KEY_6] = 0;
	  }
	  if (key[KEY_7]) {
		debugadr = (debugadr << 4) | 7;
		key[KEY_7] = 0;
	  }
	  if (key[KEY_8]) {
		debugadr = (debugadr << 4) | 8;
		key[KEY_8] = 0;
	  }
	  if (key[KEY_9]) {
		debugadr = (debugadr << 4) | 9;
		key[KEY_9] = 0;
	  }
	  if (key[KEY_A]) {
		debugadr = (debugadr << 4) | 0x0a;
		key[KEY_A] = 0;
	  }
	  if (key[KEY_B]) {
		debugadr = (debugadr << 4) | 0x0b;
		key[KEY_B] = 0;
	  }
	  if (key[KEY_C]) {
		debugadr = (debugadr << 4) | 0x0c;
		key[KEY_C] = 0;
	  }
	  if (key[KEY_D]) {
		debugadr = (debugadr << 4) | 0x0d;
		key[KEY_D] = 0;
	  }
	  if (key[KEY_E]) {
		debugadr = (debugadr << 4) | 0x0e;
		key[KEY_E] = 0;
	  }
	  if (key[KEY_F]) {
		debugadr = (debugadr << 4) | 0x0f;
		key[KEY_F] = 0;
	  }

	  debugadr &= 0x3ffff;
	  sprintf(buf1, "adr=0x%05x  -  breakpoint=0x%04x", debugadr, breakpoint);
	  acquire_screen();
	  textout(screen, font, buf1, 0, hti+256, col);
	  release_screen();

	  if (key[KEY_M]) {
		debugadrm = debugadr;
		key[KEY_M] = 0;
	  }
	  if (key[KEY_V]) {
		debugadrv = debugadr;
		key[KEY_V] = 0;
	  }
	  if (key[KEY_G]) {
		debugadrg = debugadr;
		key[KEY_G] = 0;
	  }
	  if (key[KEY_R]) {			/* bReak */
		breakpoint = debugadr;
		key[KEY_R] = 0;
	  }
	  if (key[KEY_U]) {			/* Unbreak */
		breakpoint = 0;
		key[KEY_R] = 0;
	  }
	  if (key[KEY_I]) {			/* dIsassemble */
	    myPC = debugadr & 0xffff;
	    curline = 0;
	    acquire_screen();
	    for (idx=0; idx<60; idx++)
	      {
			tmpPC = myPC;
			myPC += Dasm9900(buf2, myPC);
			sprintf(buf1, "  %04x   ", tmpPC, buf2);
			for (i=0; i<3; i++)
			  if (tmpPC+2*i < myPC)
				sprintf(buf1+9+5*i, "%04x  ", romword(tmpPC+2*i));
			  else
				sprintf(buf1+9+5*i, "      ");
			sprintf(buf1+25, "%-32s", buf2);
			textout(screen, font, buf1, wti+40, idx*8, col);
	      }
	    release_screen();
	    key[KEY_I] = 0;
	  }
	} /* max_ipf==0 */
}

void restore_video()
{

	// DOESN'T WORK

//	if (debug)
//	{
//		if (set_gfx_mode(GFX_DIRECTX,640,480,0,0)<0)
//			fail("Can't set 640x480x8 graphics mode!");
//	}
//	else
//	{
//		if (set_gfx_mode(GFX_DIRECTX,320,200,0,0)<0)
//			fail("Can't set 320x200x8 graphics mode!");
//	}
}

void VDPgraphics()
{ /* this function to display the graphics once */
int xx,yy,t,pat,col,o;
int i1,i2,i3,i4,t1,t2;
int vdp_drawn[256];		/* flags for which characters have been drawn */

/* do a simple display of graphics mode only for now */

/* clear the vdp_drawn[] array */
 if (rebuild_patterns_needed) {
   for (i1=0; i1<256; i1++)
	 vdp_drawn[i1]=0;
   rebuild_patterns_needed = 0;
 }

/* now blit each tile to the screen buffer */
/* calculate position in VDPBUFFER of base of table */

// slight speed increase at the expense of sprite flicker?

acquire_screen();
i4=0;

rectfill(screen, 0, 0, 31, 199, 0);
rectfill(screen, 288, 0, 319, 199, 0);
rectfill(screen, 0, 196, 319, 199, 0);

o=0;

for (i1=0; i1<192; i1+=8)         /* y loop */
{ for (i2=0; i2<256; i2+=8)       /* x loop */
  { if (vdp_drawn[VDP[SIT+o]]==0)
	{	/* Now we draw a single character into VDPBUFFER. Then we set a flag		*/
		/* that is has been done, so that we don't do it twice this frame. We only	*/
		/* need to draw the characters we actually use this way.					*/

		xx=VDP[SIT+o]<<3;
		p_add=PDT+xx;
		
		for (i3=0; i3<8; i3++)
		{
		  Byte c = VDP[SIT+o]>>3;
		  t=VDP[p_add++];
		
		    putpixel(VDPBUFFER,xx,i3,c|(t&0x80 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+1,i3,c|(t&0x40 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+2,i3,c|(t&0x20 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+3,i3,c|(t&0x10 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+4,i3,c|(t&0x08 ? 0x80 : 0 ));
			putpixel(VDPBUFFER,xx+5,i3,c|(t&0x04 ? 0x80 : 0 ));
			putpixel(VDPBUFFER,xx+6,i3,c|(t&0x02 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+7,i3,c|(t&0x01 ? 0x80 : 0 ));
		}
		
		vdp_drawn[VDP[SIT+o]]=1;
	}
	if ((Video_Debug)&&(i4<256))
		blit(VDPBUFFER,screen/*SCREENBUFFER*/,(i4++)<<3,0,i2+32,i1+4,8,8);
	else
		blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++]<<3),0,i2+32,i1+4,8,8);
  }
}

/* now draw the sprites - each sprite drawn because of color and shape changes */

o=0;

for (i1=0; i1<32; i1++)		/* 32 sprites */
{	yy=VDP[SAL+o++]+1;		/* sprite Y, it's stupid, cause 255 is line 0 */
	if (yy>255) yy=0;
	xx=VDP[SAL+o++];			/* sprite X */
	pat=VDP[SAL+o++];			/* sprite pattern */
	col=VDP[SAL+o++]&0xf;		/* sprite early clock (not using yet) and color */
	
	if (yy==0xd1)
		i1=32;			/* 0xd0 ends sprite list, add 1 cause I did above */

	if ((yy<192)&&(col&0xf))				/* if on-screen and not transparent */
	{	p_add=SDT+(pat<<3);
		
		clear(SPRITEBUFFER);

		for (i3=0; i3<8; i3++)
		{	t=VDP[p_add++];
			putpixel(SPRITEBUFFER,0,i3,(t&0x80 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,1,i3,(t&0x40 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,2,i3,(t&0x20 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,3,i3,(t&0x10 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,4,i3,(t&0x08 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,5,i3,(t&0x04 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,6,i3,(t&0x02 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,7,i3,(t&0x01 ? col+TIPALETTE : 0 ));

			if (VDPREG[1]&0x02)
			{	/* double-size sprites, need to draw 3 more chars */
				t=VDP[p_add+7];
				putpixel(SPRITEBUFFER,0,i3+8,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,1,i3+8,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,2,i3+8,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,3,i3+8,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,4,i3+8,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,5,i3+8,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,6,i3+8,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,7,i3+8,(t&0x01 ? col+TIPALETTE : 0 ));

				t=VDP[p_add+15];
				putpixel(SPRITEBUFFER,8,i3,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,9,i3,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,10,i3,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,11,i3,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,12,i3,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,13,i3,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,14,i3,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,15,i3,(t&0x01 ? col+TIPALETTE : 0 ));

				t=VDP[p_add+23];
				putpixel(SPRITEBUFFER,8,i3+8,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,9,i3+8,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,10,i3+8,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,11,i3+8,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,12,i3+8,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,13,i3+8,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,14,i3+8,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,15,i3+8,(t&0x01 ? col+TIPALETTE : 0 ));
			}

		}
		
		if (VDPREG[1]&0x01)
		{	/* magnified sprites */
			stretch_sprite(screen/*SCREENBUFFER*/, SPRITEBUFFER, xx+32, yy+4, 32, 32);
		}
		else
		{	draw_sprite(screen/*SCREENBUFFER*/,SPRITEBUFFER,xx+32,yy+4);
		}
	}
}

//acquire_screen();
//blit(SCREENBUFFER,screen,0,0,32,4,256,192);
release_screen();
 
}

void VDPgraphics4(int x1, int y1, int x2, int y2) { /* this function to display the graphics once */
  int i1,i2;
  Byte *p1;

  acquire_screen();
  
  for (i1=y1+4; i1<=y2+4; i1++) {	/* y loop */
	p1 = &VDP[SIT&0x18000] + ((i1-4)<<7) + (x1>>1);
	for (i2=x1+32; i2<=x2+32; i2+=2) {	/* x loop */
	  putpixel(screen, i2, i1, *p1 >> 4);
	  putpixel(screen, i2+1, i1, *p1++ & 0x0f);
	}
  }
		
  release_screen();
 
}


void VDPgraphics6(int x1, int y1, int x2, int y2) { /* this function to display the graphics once */
  int i1,i2;
  Byte *p1, *p2;

  acquire_screen();
  
  for (i1=y1+4; i1<=y2+4; i1+=2) {	/* y loop */
	p1 = &VDP[SIT&0x10000] + ((i1-4)<<7) + (x1>>1);
	if (VDPREG[9] & 0x0c == 0x0c) /* interlace */
	  p2 = p1 - 0x10000;		/* &VDP[0]; FIXME: handle p1=&VDP[0] */
	else
	  p2 = p1;

	for (i2=x1+32; i2<=x2+32; i2+=2) {	/* x loop */
	  putpixel(screen, i2, i1, *p1 >> 4);
	  putpixel(screen, i2+1, i1, *p1++ & 0x0f);
	  putpixel(screen, i2, i1+1, *p2 >> 4);
	  putpixel(screen, i2+1, i1+1, *p2++ & 0x0f);
	}
  }
		
  release_screen();
 
}


void VDPgraphics7(int x1, int y1, int x2, int y2) { /* this function to display the graphics once */
  int i1,i2;
  Byte *p1, *p2;

  acquire_screen();
  
  for (i1=y1+4; i1<=y2+4; i1+=2) {	/* y loop */
	p1 = &VDP[SIT&0x10000] + ((i1-4)<<7) + (x1>>1);
	if (VDPREG[9] & 0x0c == 0x0c) /* interlace */
	  p2 = p1 - 0x10000;		/* &VDP[0]; FIXME: handle p1=&VDP[0] */
	else
	  p2 = p1;

	for (i2=x1+32; i2<=x2+32; i2+=2) {	/* x loop */
	  putpixel(screen, i2, i1, *p1);
	  putpixel(screen, i2+1, i1, *p1++);
	  putpixel(screen, i2, i1+1, *p2);
	  putpixel(screen, i2+1, i1+1, *p2++);
	}
  }
		
  release_screen();
 
}


void VDPtext()
{ /* this function to display the text once */
int xx,t, o=0;
int i1,i2,i3,i4,t1,t2;
int vdp_drawn[256];		/* flags for which characters have been drawn */

/* clear the vdp_drawn[] array */
 if (rebuild_patterns_needed) {
   for (i1=0; i1<256; i1++)
	 vdp_drawn[i1]=0;
   rebuild_patterns_needed = 0;
 }

/* now blit each tile to the screen buffer */
/* calculate position in VDPBUFFER of base of table */

// slight speed increase at the expense of sprite flicker?

acquire_screen();
i4=0;

rectfill(screen, 0, 0, 39, 199, 0);
rectfill(screen, 280, 0, 319, 199, 0);
rectfill(screen, 0, 196, 319, 199, 0);

for (i1=0; i1<192; i1+=8)         /* y loop */
{ for (i2=0; i2<240; i2+=6)       /* x loop */
  { if (vdp_drawn[VDP[SIT+o]]==0)
	{	/* Now we draw a single character into VDPBUFFER. Then we set a flag		*/
		/* that is has been done, so that we don't do it twice this frame. We only	*/
		/* need to draw the characters we actually use this way.					*/

		xx=VDP[SIT+o]<<3;
		p_add=PDT+xx;
		
		for (i3=0; i3<8; i3++)		// 6 pixels wide
		{	t=VDP[p_add++];
		    putpixel(VDPBUFFER,xx,i3,(t&0x80 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+1,i3,(t&0x40 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+2,i3,(t&0x20 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+3,i3,(t&0x10 ? 0x80 : 0 ));
		    putpixel(VDPBUFFER,xx+4,i3,(t&0x08 ? 0x80 : 0 ));
			putpixel(VDPBUFFER,xx+5,i3,(t&0x04 ? 0x80 : 0 ));
		}
		
		vdp_drawn[VDP[SIT+o]]=1;
	}
	if ((Video_Debug)&&(i4<256))
		blit(VDPBUFFER,screen/*SCREENBUFFER*/,(i4++)<<3,0,i2+40,i1+4,6,8);
	else
		blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++])<<3,0,i2+40,i1+4,6,8);
  }
}

// no sprites in text mode

//acquire_screen();
//blit(SCREENBUFFER,screen,0,0,32,4,256,192);
release_screen();
  
}


void VDPtext2()
{ /* this function to display the text once */
int xx,t, o=0;
int i1,i2,i3,i4,t1,t2;
int vdp_drawn[256];		/* flags for which characters have been drawn */

/* clear the vdp_drawn[] array */
 if (rebuild_patterns_needed) {
   // initialize VDPBUFFER
   rectfill(VDPBUFFER, 0, 0, 2047, 15, 0);
   rectfill(VDPBUFFER, 2048, 0, 4095, 15, 2);
   for (i1=0; i1<256; i1++)
	 vdp_drawn[i1]=0;
   rebuild_patterns_needed = 0;
 }

/* now blit each tile to the screen buffer */
/* calculate position in VDPBUFFER of base of table */

acquire_screen();
i4=0;

/*  rectfill(screen, 0, 0, 39, 199, 0); */
/*  rectfill(screen, 280, 0, 319, 199, 0); */
/*  rectfill(screen, 0, 196, 319, 199, 0); */

for (i1=0; i1< (VDPREG[9]&0x80 ? 212 : 192); i1+=8)         /* y loop */
{ for (i2=0; i2<480; i2+=6)       /* x loop */
  { if (vdp_drawn[VDP[SIT+o]]==0)
	{	/* Now we draw a single character into VDPBUFFER. Then we set a flag		*/
		/* that is has been done, so that we don't do it twice this frame. We only	*/
		/* need to draw the characters we actually use this way.					*/

		xx=VDP[SIT+o]<<3;
		p_add=PDT+xx;
		
		for (i3=0; i3<8; i3++)		// 6 pixels wide
		{	t=VDP[p_add++];
		    t1 = i3<<1;

			if (t&0x80) {
			  putpixel(VDPBUFFER,xx,t1,1);
			  putpixel(VDPBUFFER,xx,t1+1,1);
			}
			if (t&0x40) {
			  putpixel(VDPBUFFER,xx+1,t1,1);
			  putpixel(VDPBUFFER,xx+1,t1+1,1);
			}
			if (t&0x20) {
			  putpixel(VDPBUFFER,xx+2,t1,1);
			  putpixel(VDPBUFFER,xx+2,t1+1,1);
			}
			if (t&0x10) {
			  putpixel(VDPBUFFER,xx+3,t1,1);
			  putpixel(VDPBUFFER,xx+3,t1+1,1);
			}
			if (t&0x08) {
			  putpixel(VDPBUFFER,xx+4,t1,1);
			  putpixel(VDPBUFFER,xx+4,t1+1,1);
			}
			if (t&0x04) {
			  putpixel(VDPBUFFER,xx+5,t1,1);
			  putpixel(VDPBUFFER,xx+5,t1+1,1);
			}
			/* same character with blink bit on */
			if (t&0x80) {
			  putpixel(VDPBUFFER,xx+2048,t1,3);
			  putpixel(VDPBUFFER,xx+2048,t1+1,3);
			}
			if (t&0x40) {
			  putpixel(VDPBUFFER,xx+2048+1,t1,3);
			  putpixel(VDPBUFFER,xx+2048+1,t1+1,3);
			}
			if (t&0x20) {
			  putpixel(VDPBUFFER,xx+2048+2,t1,3);
			  putpixel(VDPBUFFER,xx+2048+2,t1+1,3);
			}
			if (t&0x10) {
			  putpixel(VDPBUFFER,xx+2048+3,t1,3);
			  putpixel(VDPBUFFER,xx+2048+3,t1+1,3);
			}
			if (t&0x08) {
			  putpixel(VDPBUFFER,xx+2048+4,t1,3);
			  putpixel(VDPBUFFER,xx+2048+4,t1+1,3);
			}
			if (t&0x04) {
			  putpixel(VDPBUFFER,xx+2048+5,t1,3);
			  putpixel(VDPBUFFER,xx+2048+5,t1+1,3);
			}
		} /* for i3 */
		
		vdp_drawn[VDP[SIT+o]]=1;
	} /* if !drawn */

  i3 = VDP[CT + (o>>3)];				/* significant byte in color table */
  i3 &= 1 << (7 - (o&0x07));	/* mask significant bit */
  if (i3)
	i3 = 256;					/* char offset if CT bit set */

  if ((Video_Debug)&&(i4<256))
	blit(VDPBUFFER,screen/*SCREENBUFFER*/,(i4++)<<3,0,i2+40,2*i1+4,6,16);
  else
	blit(VDPBUFFER,screen/*SCREENBUFFER*/,(i3+VDP[SIT+o++])<<3,0,i2+40,2*i1+4,6,16);
  }	/* for i2 */
} /* for i1 */

// no sprites in text mode

//acquire_screen();
//blit(SCREENBUFFER,screen,0,0,32,4,256,192);
release_screen();
  
}


void VDPbitmap()
{ /* this function to display the bitmap once */
int xx,yy,t,pat,col, bank, o;
int i1,i2,i3,t1,t2;
int vdp_drawn[256];		/* flags for which characters have been drawn */
Word c_add;

// slight speed increase at the expense of sprite flicker?

acquire_screen();

rectfill(screen, 0, 0, 31, 199, 0);
rectfill(screen, 288, 0, 319, 199, 0);
rectfill(screen, 0, 196, 319, 199, 0);

// The normal math for table addresses isn't quite right in bitmap mode
// adjust here

PDT=((VDPREG[4]&0x04) << 11);
CT=((VDPREG[3]&0x80) << 6);

for (bank=0; bank<3; bank++)	// 3 banks in bitmap mode
{

/* clear the vdp_drawn[] array */

for (i1=0; i1<256; i1++)
  vdp_drawn[i1]=0;

/* now blit each tile to the screen buffer */
/* calculate position in VDPBUFFER of base of table */

o=0;

for (i1=0; i1<64; i1+=8)         /* y loop */
{ for (i2=0; i2<256; i2+=8)       /* x loop */
  { if (vdp_drawn[VDP[SIT+o]]==0)
	{	/* Now we draw a single character into VDPBUFFER. Then we set a flag		*/
		/* that is has been done, so that we don't do it twice this frame. We only	*/
		/* need to draw the characters we actually use this way.					*/

		xx=VDP[SIT+o]<<3;
		p_add=(PDT+xx+(bank<<11))&0x3fff;
		c_add=(CT+xx+(bank<<11))&0x3fff;

		for (i3=0; i3<8; i3++)
		{	t=VDP[p_add++];
			t1=(VDP[c_add])>>4;
			t2=(VDP[c_add++]) & 0xf;
		    putpixel(VDPBUFFER,xx,i3,(t&0x80 ? t1 : t2 ));
		    putpixel(VDPBUFFER,xx+1,i3,(t&0x40 ? t1 : t2 ));
		    putpixel(VDPBUFFER,xx+2,i3,(t&0x20 ? t1 : t2 ));
		    putpixel(VDPBUFFER,xx+3,i3,(t&0x10 ? t1 : t2 ));
		    putpixel(VDPBUFFER,xx+4,i3,(t&0x08 ? t1 : t2 ));
			putpixel(VDPBUFFER,xx+5,i3,(t&0x04 ? t1 : t2 ));
			putpixel(VDPBUFFER,xx+6,i3,(t&0x02 ? t1 : t2 ));
		    putpixel(VDPBUFFER,xx+7,i3,(t&0x01 ? t1 : t2 ));
		}
		
		vdp_drawn[VDP[SIT+o]]=1;
	}
	blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++])<<3,0,i2+32,i1+4+(bank<<6),8,8);
  }
}
}

/* now draw the sprites - each sprite drawn because of color and shape changes */

o=0;

for (i1=0; i1<32; i1++)		/* 32 sprites */
{	yy=VDP[SAL+o++]+1;		/* sprite Y, it's stupid, cause 255 is line 0 */
	if (yy>255) yy=0;
	xx=VDP[SAL+o++];			/* sprite X */
	pat=VDP[SAL+o++];			/* sprite pattern */
	col=VDP[SAL+o++]&0xf;		/* sprite early clock (not using yet) and color */
	
	if (yy==0xd1)
		i1=32;			/* 0xd0 ends sprite list, add 1 cause I did above */

	if ((yy<192)&&(col&0xf))				/* if on-screen and not transparent */
	{	p_add=SDT+(pat<<3);
		
		clear(SPRITEBUFFER);

		for (i3=0; i3<8; i3++)
		{	t=VDP[p_add++];
			putpixel(SPRITEBUFFER,0,i3,(t&0x80 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,1,i3,(t&0x40 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,2,i3,(t&0x20 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,3,i3,(t&0x10 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,4,i3,(t&0x08 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,5,i3,(t&0x04 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,6,i3,(t&0x02 ? col+TIPALETTE : 0 ));
			putpixel(SPRITEBUFFER,7,i3,(t&0x01 ? col+TIPALETTE : 0 ));

			if (VDPREG[1]&0x02)
			{	/* double-size sprites, need to draw 3 more chars */
				t=VDP[p_add+7];
				putpixel(SPRITEBUFFER,0,i3+8,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,1,i3+8,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,2,i3+8,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,3,i3+8,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,4,i3+8,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,5,i3+8,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,6,i3+8,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,7,i3+8,(t&0x01 ? col+TIPALETTE : 0 ));

				t=VDP[p_add+15];
				putpixel(SPRITEBUFFER,8,i3,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,9,i3,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,10,i3,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,11,i3,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,12,i3,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,13,i3,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,14,i3,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,15,i3,(t&0x01 ? col+TIPALETTE : 0 ));

				t=VDP[p_add+23];
				putpixel(SPRITEBUFFER,8,i3+8,(t&0x80 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,9,i3+8,(t&0x40 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,10,i3+8,(t&0x20 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,11,i3+8,(t&0x10 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,12,i3+8,(t&0x08 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,13,i3+8,(t&0x04 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,14,i3+8,(t&0x02 ? col+TIPALETTE : 0 ));
				putpixel(SPRITEBUFFER,15,i3+8,(t&0x01 ? col+TIPALETTE : 0 ));
			}

		}
		
		if (VDPREG[1]&0x01)
		{	/* magnified sprites */
			stretch_sprite(screen/*SCREENBUFFER*/, SPRITEBUFFER, xx+32, yy+4, 32, 32);
		}
		else
		{	draw_sprite(screen/*SCREENBUFFER*/,SPRITEBUFFER,xx+32,yy+4);
		}
	}
}

//acquire_screen();
//blit(SCREENBUFFER,screen,0,0,32,4,256,192);
release_screen();
 
}

// Local Variables: //
// tab-width:4 //
// comment-column:32 //
// fill-column:100 //
// End: //
