//*****************************************************
//* Ami 99 - TI Emulator for Win32					  *
//* by M.Brent                                        *
//*                                                   *
//* Version DXL2.0	-	9 Feb 00					  *
//*                                                   *
//* Thanks to:	Jeff Brown ('X' instruction)		  *
//*				Frank Palazolo, Ralph Nebet (speech)  *
//*			    Roland Meier (Linux, Disk, 9938 code) *
//*													  *
//* It should be noted that this Emulator is Dependant*
//* on running the original ROMs and expects to find  *
//* ROM routines at certain addresses. Custom ROMs    *
//* are very likely to act strangely or outright fail *
//*****************************************************
// Linux code merge is experimental
// WIN32 code is intended for MSVC

// ifdefs are specifically written to target WIN32 and __linux__
// code won't compile without one of them

// Warning: Scratchpad RAM is now located at 0x8000, not 0x8300
// any patches that want to access it directly (not through ROMWORD or RCPUBYTE)
// must take note of this or they will fail

#ifdef WIN32
#pragma warning (disable: 4113 4761)

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0400
#endif

#define VERSION "DXL2.0"

////////////////////////////////////////////
// Includes
////////////////////////////////////////////

#ifdef WIN32
#include <stdio.h>
#include <windows.h>
#include <mmsystem.h>
#include <process.h>
#include <winbase.h>
#include <ddraw.h>
#include <dsound.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "resource.h"
#pragma hdrstop
#endif

#ifdef __linux__
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <allegro.h>
#include <signal.h>
#include <sys/time.h>
#include <math.h>
/*  #include <vga.h> */
/*  #include <vgagl.h> */
/*  #include <vgakeyboard.h> */
/*  #include "tiemul-linux.h" */
#endif

#include "tiemul.h"

////////////////////////////////////////////
// Globals
// These don't all NEED to be globals, but I'm only cleaning up the code, not re-writing it
// all from scratch.
////////////////////////////////////////////

// Memory
Byte CPU[65536];									// Main CPU Memory
Byte CPU2[8192];									// Cartridge space bank-switched
Byte GROM[65536];									// GROM space
Byte VDP[0x40000];									// Video RAM
Byte ROMMAP[65536];									// Write-protect map of CPU space
Byte CRU[4096];										// CRU space
Word GRMADD;										// GROM Address counter
Byte grmaccess,grmdata;								// GROM Prefetch emulation

// Keyboard
#ifdef WIN32
int KEYS[8][8]= { 
/* Joy 2 */		VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,

/* 0 */			'M', 'J', 'U', '7', '4', 'F', 'R', 'V',
/* 1 */			191, 186, 'P', '0', '1', 'A', 'Q', 'Z',
/* 2 */			190, 'L', 'O', '9', '2', 'S', 'W', 'X',

/* Joy 1 */		VK_TAB, VK_RIGHT, VK_LEFT, VK_UP, VK_DOWN, VK_ESCAPE, VK_ESCAPE, VK_ESCAPE,
	
/* 3 */			188, 'K', 'I', '8', '3', 'D', 'E', 'C',
/* 4 */			'N', 'H', 'Y', '6', '5', 'G', 'T', 'B',
/* 5 */			187, VK_SPACE, VK_RETURN, VK_ESCAPE, VK_MENU, VK_SHIFT, VK_CONTROL, VK_ESCAPE 
};
char key[256];										// keyboard state buffer
#endif

#ifdef __linux__
int KEYS[8][8]= { 

/* Joy 2 */		KEY_ESC, KEY_ESC, KEY_ESC, KEY_ESC, KEY_ESC, KEY_ESC, KEY_ESC, KEY_ESC,

/* 0 */			KEY_M, KEY_J, KEY_U, KEY_7, KEY_4, KEY_F, KEY_R, KEY_V,
/* 1 */			KEY_SLASH, KEY_COLON, KEY_P, KEY_0, KEY_1, KEY_A, KEY_Q, KEY_Z,
/* 2 */			KEY_STOP, KEY_L, KEY_O, KEY_9, KEY_2, KEY_S, KEY_W, KEY_X,

/* Joy 1 */		KEY_TAB, KEY_RIGHT, KEY_LEFT, KEY_UP, KEY_DOWN, KEY_ESC, KEY_ESC, KEY_ESC,
	
/* 3 */			KEY_COMMA, KEY_K, KEY_I, KEY_8, KEY_3, KEY_D, KEY_E, KEY_C,
/* 4 */			KEY_N, KEY_H, KEY_Y, KEY_6, KEY_5, KEY_G, KEY_T, KEY_B,
/* 5 */			KEY_EQUALS, KEY_SPACE, KEY_ENTER, KEY_ESC, KEY_ALT, KEY_LSHIFT, KEY_LCONTROL, KEY_ESC 
};
#endif

// VDP
int redraw_needed;									// redraw flag
int interrupt_needed;								// interrupt flag
int setpalette_needed;								// set after write access to color definitions
int rebuild_patterns_needed;						// set after write access to PDT
Byte VDPREG[64];									// VDP read-only registers
Word VDPADD;										// VDP Address counter
Byte vdpaccess;										// VDP Prefetch
unsigned long hVideoThread;							// thread handle
Byte hzRate;										// flag for 50 or 60hz
Byte Recording;										// Flag for AVI recording
extern Byte Recording;
Byte RecordFrame;
int MaintainAspect;									// Flag for Aspect ratio

// v9938
Byte VDPS[16];										// v9938 status registers
int v9938 = 0;										// emulate V9938 video display processor
int mode, oldmode=-1;								// current graphics mode and old value
int wti=256, hti=192, wscr, hscr, oldwscr=-1;		// Screen sizes
int mx=0, my=0, oldmx, oldmy;						// mouse position, old values
int firstcolorbyte = -1;							// first of two bytes defining a color
#ifdef WIN32
int MouseX, MouseY, MouseB;							// tracking of Windows mouse and buttons
#endif

// Audio
int need_byte;										// flag for whether waiting on a second byte
int last_byte;										// copy of last byte

#ifdef WIN32
extern LPDIRECTSOUNDBUFFER voice[3];				// extern for voice structures
extern LPDIRECTSOUNDBUFFER noise[8];				// noise structures
#endif
#ifdef __linux__
SAMPLE *spl[9];										// 8 noises and a tone
int voice[4];										// 4 voices
#endif

// Assorted
char qw[80];										// temp string
char quitflag;										// quit flag
extern char quitflag;
extern int drawspeed;								// Frame skip
char lines[30][100];								// debug lines
int bank=0;											// Cartridge bank switch
int xb = 0;											// Is second bank (XB) loaded?
unsigned int index1;									// General counter variable
int debug,d2;										// debugging flags
int olddebug=-1;		
int debugadr=0, debugadrv=0;
Word debugadrm=0, debugadrg=0, breakpoint=0;
int drawspeed=0;									// flag used in display updating
int max_ipf=9999999;								// Maximum instructions per frame
int oldmax=9999999;									// copy of same
int speedup_float = 0;								// speedup floating point operations
int speedup_fr213 = 0;								// speedup fractal calculations
int instruction_count;								// Instruction counter for speed throttle
int cpucount, cpuframes;							// CPU counters for timing
int last_count=0;									// Separates video redraw and video interrupts
int timercount;										// Used to estimate runtime

#ifdef WIN32
int retrace_count=0;								// count on the 60hz timer
#endif

int PauseInactive, DrawInactive;					// what to do when the window is inactive
time_t STARTTIME, ENDTIME;
long ticks;

#ifdef WIN32
ATOM myClass;										// Window Class
HWND myWnd;											// Handle to window
extern HWND myWnd;
HINSTANCE hInstance;								// global program instance
UINT myTimer;										// 60hz timer
HANDLE TimerEvent;									// timer event
char AVIFileName[256]="C:\\AMI99AVI.AVI";			// AVI Filename
#endif

char diskpath[10][256];								// path to disk directories or images
char disktype[10];									// i=image, v=files with v9t9 header
													// t=files with TIFILES header,
													// r=raw program files without header
unsigned long myThread;								// timer thread

// Console Interrupt Routine Simulator
int num_sprites;									// Number of active sprites
int y_speed;										// X pixel speed
int x_speed;										// Y pixel speed
int t1, t1a, t1b, t2, t2a, t2b;						// temp variables
int sprite_y, sprite_x, offset;						// Sprite location
int ty, tx;											// Temp location
int snd_address;									// Sound chip data address
Byte *sndlist;										// The list itself
Byte data;											// temp data

///////////////////////////////////
// CPU Opcodes
///////////////////////////////////
#include "cpu9900.cpp"

///////////////////////////////////
// VDP Routines
///////////////////////////////////
#include "tivdp.cpp"

///////////////////////////////////
// Window routine
///////////////////////////////////
#ifdef WIN32
#include "WindowProc.cpp"
#endif

///////////////////////////////////
// Main
// Startup and shutdown system
///////////////////////////////////
#ifdef __linux__
void timer_handler(int s) 
{
  retrace_count++;
  interrupt_needed=1;
  VDPS[2] ^= 0x62;									// toggle HR, VW, and EO
  VDPS[8]++;										// FIXME: test
  //  signal (s, timer_handler);					// reinstall (only SYSV)
}

int main(int argc, char *argv[]) {
	struct itimerval timerval;
	char temp[1024];
	int err;
	int idx;

	if (signal(SIGVTALRM, timer_handler)==SIG_ERR)
		fail("Can't install signal handler");

	/* FIXME: using virtual timer, real timer used by allegro */
	timerval.it_interval.tv_usec = (long int) 20000;	// 20ms virtual time
	timerval.it_interval.tv_sec = 0;
	timerval.it_value.tv_usec = (long int) 50000;		// 50ms
	timerval.it_value.tv_sec = 0;
	if (setitimer (ITIMER_VIRTUAL, &timerval, NULL) < 0)
		fail("Can't install timer");
  
	allegro_init();

	set_window_title("Ami99");
	    
	install_keyboard();
	install_timer();
#endif

#ifdef WIN32
int WINAPI WinMain( HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	WNDCLASS aclass;
	char temp[1024];
	int err;
	int idx;

	hInstance = hInst;

	// Null the pointers
	myTimer=NULL;
	TimerEvent=NULL;
	myClass=0;
	myWnd=NULL;
	lpdd=NULL;
	lpds=NULL;
	voice[0]=NULL;
	voice[1]=NULL;
	voice[2]=NULL;

	for (idx=0; idx<8; idx++)
		noise[idx]=NULL;

	ScreenBuffer=BackBuffer=VDPBuffer=SpriteBuffer=NULL;
	framedata=NULL;

	// create and register a class and open a window

	if (NULL == hPrevInstance)
	{
		aclass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
		aclass.lpfnWndProc = myproc;
		aclass.cbClsExtra = 0;
		aclass.cbWndExtra = 0;
		aclass.hInstance = hInst;
		aclass.hIcon = NULL;   //(struct HICON__ *)MAKEINTRESOURCE(IDI_ICON2); How do I make this work?
		aclass.hCursor = LoadCursor(NULL, IDC_ARROW);
		aclass.hbrBackground = NULL;
		aclass.lpszMenuName = MAKEINTRESOURCE(IDR_MENU1);
		aclass.lpszClassName = "TIWndClass";
		myClass = RegisterClass(&aclass);
		if (0 == myClass)
		{	
			err=GetLastError();
			sprintf(temp, "Can't create class: %x", err);
			fail(temp);
		}
	}

	myWnd = CreateWindow("TIWndClass", "Ami99", WS_OVERLAPPEDWINDOW | WS_SIZEBOX | WS_VISIBLE, CW_USEDEFAULT, 0, 264, 238, NULL, NULL, hInstance, NULL);
	if (NULL == myWnd)
	{	
		err=GetLastError();
		sprintf(temp, "Can't open window: %x", err);
		fail(temp);
	}
	ShowWindow(myWnd, SW_SHOWNORMAL);

	// set up 60hz timer
	myThread=NULL;
	myThread=_beginthread((void (__cdecl *)(void*))TimerThread, 0, NULL);
	if (myThread != -1)
		debug_write("Thread began...");
	else
		debug_write("Thread failed.");

	Sleep(100);			// time for thread to start
#endif

	grmaccess=0;		// No GROM Access yet
	vdpaccess=0;		// No VDP address writes yet 
	interrupt_needed=0;	// No interrupt missed yet
	quitflag=0;			// no quit yet

	// set memory banks to zero. Real TIs don't do this
	memset(CPU, 0, 65536);
	memset(CPU2, 0, 8192);
	memset(GROM, 0, 65536);
	memset(VDP, 0, 0x40000);
	memset(CRU, 1, 4096);
#ifdef WIN32
	memset(key, 0, 256);
#endif

	VDPS[7]=1;			// FIXME: for a test to make it different from #0

	// set interrupt rate
	hzRate=60;

	// video recording
	Recording=0;

	// window inactive
	DrawInactive=0;
	PauseInactive=0;

	// clear debugging strings and state
	debug=0;
	memset(lines, 0, 3000);

	// Print some initial debug
	sprintf(temp, "\n\n---\nAmi99 version %s (c)1999 M.Brent", VERSION);
	debug_write(temp);

	// read the TIDATA file and load up the ROMs
	debug_write("Reading ROMs");
	readroms();

	// build the CPU table
	debug_write("Building CPU");
	buildcpu();

	// start sound
	debug_write("Starting Sound");
	startsound();
	
	// start the video processor
	debug_write("Starting Video");
	startvdp();

	// v9938 mouse setup
#ifdef __linux__
	if (v9938)
	{
		install_mouse();
	}
#endif

	// prepare the emulation...
	cpucount=0;
	cpuframes=0;
	timercount=0;

	// begin emulation - returns when it's time to exit
	debug_write("Starting Emulation");
	emulti();

	// Fail is the full exit
	debug_write("Shutting down");
	fail("Normal Termination");

	// good bye
	return 0;
}


//////////////////////////////////////////////////////
// start up the sound system
//////////////////////////////////////////////////////
void startsound()
{ /* start up the sound files */

#ifdef WIN32
	DSBUFFERDESC dsbd;
	WAVEFORMATEX pcmwf;
	unsigned int idx, idx2;
	UCHAR c;
	UCHAR *ptr1, *ptr2;
	unsigned long len1, len2, len;
	char buf[80];
	FILE *fp;
	
	if (DirectSoundCreate(NULL, &lpds, NULL) != DS_OK)
	{
		lpds=NULL;		// no sound
		return;
	}
	
	if (lpds->SetCooperativeLevel(myWnd, DSSCL_NORMAL) != DS_OK)
	{
		lpds->Release();
		return;
	}

	for (idx=0; idx<4; idx++)
	{
		ZeroMemory(&pcmwf, sizeof(pcmwf));
		pcmwf.wFormatTag = WAVE_FORMAT_PCM;		// wave file
		pcmwf.nChannels=1;						// 1 channel (mono)
		pcmwf.nSamplesPerSec= 22050;			// 22khz
		pcmwf.nBlockAlign=1;					// 1 byte per sample * 1 channel
		pcmwf.nAvgBytesPerSec=pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
		pcmwf.wBitsPerSample=8;					// 8 bit samples
		pcmwf.cbSize=0;							// always zero;

		ZeroMemory(&dsbd, sizeof(dsbd));
		dsbd.dwSize=sizeof(dsbd);
		dsbd.dwFlags=DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC;
		dsbd.dwBufferBytes=1000;				// the sample is 1000 bytes long
		dsbd.lpwfxFormat=&pcmwf;

		if (lpds->CreateSoundBuffer(&dsbd, &voice[idx], NULL) != DS_OK)
		{
			lpds->Release();
			return;
		}
	
		// I have calculated that the square wav changed state every 20 samples. so, this
		// just generates the square wave. Apologies to those who liked the sine or triangle
		// versions, but they could be coded in
		if (voice[idx]->Lock(0, 1000, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER) == DS_OK)
		{
			// since we haven't started the sound, hopefully the second pointer is nil
			if (len2 != 0)
				MessageBox(NULL, "Why can't things just be easy?", "D'oh!", MB_OK);

			// unsigned
			c=255;
			for (idx2=0; idx2<len1; idx2++)
			{
				if ((idx2/20)*20 == idx2)
				{
					if (c==0)
						c=240;
					else
						c=0;
				}

				*(ptr1 + idx2)=c;
			}

			voice[idx]->Unlock(ptr1, len1, ptr2, len2);
		}

		if (voice[idx]->SetVolume(-10000) != DS_OK)
			debug_write("Set volume failed");

		if (voice[idx]->Play(0, 0, DSBPLAY_LOOPING) != DS_OK)
			debug_write("Voice DID NOT START");
	}

	// each noise channel now has it's own voice, to eliminate the need to change buffers
	// and mess with repeat lengths and the like

	for (idx=0; idx<8; idx++)
	{
		sprintf(buf,"noise%d.voc",idx+1);
		ZeroMemory(&pcmwf, sizeof(pcmwf));

		fp=fopen(buf,"rb");
		if (NULL==fp)
		{
			debug_write(buf);
		}
		else
		{
			fread(buf,1,25,fp);
			// in a demonstration of bad programming, I'm just going to read the darn thing ;)
			// if it's not a VOC, you may be in trouble
			idx2=buf[20] + (256*buf[21]);			// header length
			idx2-=25;								// we've already read 25 bytes
			while (idx2--)
				fgetc(fp);

			// it's simple. I expect it to be a 22khz, 8 bit mono VOC with 1 part
			// good luck with anything else
			fgetc(fp);								// skip block type
			len=fgetc(fp)+(256*fgetc(fp))+(65536*fgetc(fp));
			fgetc(fp); fgetc(fp);					// two more bytes to skip

			pcmwf.wFormatTag = WAVE_FORMAT_PCM;		// wave file
			pcmwf.nChannels=1;						// 1 channel (mono)
			pcmwf.nSamplesPerSec= 22050;			// 22khz
			pcmwf.nBlockAlign=1;					// 1 byte per sample * 1 channel
			pcmwf.nAvgBytesPerSec=pcmwf.nSamplesPerSec * pcmwf.nBlockAlign;
			pcmwf.wBitsPerSample=8;					// 8 bit samples
			pcmwf.cbSize=0;							// always zero;

			ZeroMemory(&dsbd, sizeof(dsbd));
			dsbd.dwSize=sizeof(dsbd);
			dsbd.dwFlags=DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_STATIC;
			dsbd.dwBufferBytes=len;
			dsbd.lpwfxFormat=&pcmwf;

			if (lpds->CreateSoundBuffer(&dsbd, &noise[idx], NULL) != DS_OK)
			{
				lpds->Release();
				return;
			}
	
			if (noise[idx]->Lock(0, len, (void**)&ptr1, &len1, (void**)&ptr2, &len2, DSBLOCK_ENTIREBUFFER) == DS_OK)
			{
				// since we haven't started the sound, hopefully the second pointer is nil
				if (len2 != 0)
					MessageBox(NULL, "Why can't things just be easy?", "D'oh!", MB_OK);

				fread(ptr1, 1, len, fp);
			}
			noise[idx]->Unlock(ptr1, len1, ptr2, len2);
		
			if (noise[idx]->SetVolume(-10000) != DS_OK)
				debug_write("Set volume failed");

			if (noise[idx]->Play(0, 0, DSBPLAY_LOOPING) != DS_OK)
				debug_write("Noise DID NOT START");

			fclose(fp);
		}
	}
#endif

#ifdef __linux__
	int i;
	char buf[80];

	int detect_digi_driver(int driver_id);

	if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) != 0)
	{	if (install_sound(DIGI_NONE, MIDI_NONE, NULL) != 0)
			fail("Can't install sound");
	}
	
	set_volume(255, -1);							/* max digi volume, ignore midi */

	spl[0]=load_sample("tone.wav");					/* this is a short wav at 22khz, playing a 550hz tone */
	if (spl[0] == 0)
	  fail("Can't load sound");
	
	for (i=1; i<9; i++)
	{
		sprintf(buf,"noise%d.voc",i);
		spl[i]=load_sample(buf);
		if (spl[i] == 0)
		  fail("Can't load noise");
	}

	for (i=0; i<3; i++)
	{	voice[i]=allocate_voice(spl[0]);			/* start the sample */
		voice_set_playmode(voice[i],PLAYMODE_LOOP); /* loop it */
		voice_set_volume(voice[i], 0);				/* mute the voice */
		voice_start(voice[i]);						/* start playing */
	}
	voice[3]=allocate_voice(spl[1]);
	voice_set_playmode(voice[3], PLAYMODE_LOOP);
	voice_set_volume(voice[3], 0);					/* mute volume */
	voice_start(voice[3]);							/* join in */
#endif

	need_byte=0;
	last_byte=0;
}

//////////////////////////////////////////////////////////
// Start up the video system 
//////////////////////////////////////////////////////////
void startvdp()
{ 
#ifdef WIN32
	// call VDP Startup
	hVideoThread=NULL;
	hVideoThread=_beginthread((void (__cdecl *)(void*))VDPmain, 0, NULL);
	if (hVideoThread != -1)
		debug_write("Video Thread began...");
	else
		debug_write("Video Thread failed.");

	Sleep(100);
#endif

#ifdef __linux__
	VDPmain();
#endif

	// first retrace
	retrace_count=0;
}

//////////////////////////////////////////////////////////
// Non-fatal recoverable (?) error
// Used to say the EMULATOR is ok, but the TI is toast ;)
//////////////////////////////////////////////////////////
void warn(char *x)
{ 
	// currently all errors are fatal. Eventually should offer a chance to continue
	fail(x);
}

//////////////////////////////////////////////////////////
// Fatal error - clean up and exit
// Note that normal exit is a fatal error ;)
//////////////////////////////////////////////////////////
void fail(char *x)
{ 
	// fatal error
	char buffer[1024];
	char buf2[256];

#ifdef __linux__
	while (keypressed())
	{ readkey();
	}

	if (VDPBUFFER) destroy_bitmap(VDPBUFFER);
	if (SPRITEBUFFER) destroy_bitmap(SPRITEBUFFER);
	if (SCREENBUFFER) destroy_bitmap(SCREENBUFFER);

	allegro_exit();
#endif

	// just in case it's not set yet
	quitflag=1;

	// add to the log - not useful now, but maybe in the future when it writes to disk
	debug_write(x);

#ifdef WIN32
	Sleep(100);			// pause to let threads catch up

	timeEndPeriod(1);
#endif

	sprintf(buffer,"\n%s\n",x);
	sprintf(buf2,"PC-%.4X  WP-%.4X  ST-%.4X  OP-%.4X\nGROM-%.4X VDP-%.4X\n",PC,WP,ST,in,GRMADD,VDPADD);
	strcat(buffer,buf2);
	sprintf(buf2,"Run Duration  : %d seconds\n",timercount/60);
	strcat(buffer,buf2);
	sprintf(buf2,"Operation time: %d instructions processed.\n",cpucount);
	strcat(buffer,buf2);
	sprintf(buf2,"Display frames: %d video frames displayed.\n",cpuframes);
	strcat(buffer,buf2);

	if (timercount<60) timercount=60;	// avoid divide by zero
	
	sprintf(buf2,"Average speed : %d instructions per second.\n",cpucount/(timercount/60));
	strcat(buffer,buf2);
	sprintf(buf2,"Frameskip     : %d\n",drawspeed);
	strcat(buffer,buf2);

	// the messagebox fails during a normal exit in WIN32.. why is that?
	MessageBox(myWnd, buffer, "Ami99 Exit", MB_OK);

#ifdef WIN32
	Sleep(600);			// give the threads a little time to shut down

	if (Recording)
		CloseAVI();

	if (myTimer)
	{
		timeKillEvent(myTimer);
		CloseHandle(TimerEvent);
	}

	if (voice[0])
	{
		voice[0]->Stop();
		voice[0]->Release();
	}

	if (lpds)
		lpds->Release();

	if (ScreenBuffer)
		ScreenBuffer->Release();

	if (BackBuffer)
		BackBuffer->Release();

	if (VDPBuffer)
		VDPBuffer->Release();

	if (SpriteBuffer)
		SpriteBuffer->Release();

	if (lpdd)
		lpdd->Release();

	if (myWnd) 
		DestroyWindow(myWnd);
	
	if (myClass)
		UnregisterClass("TIWndClass", hInstance);
#endif

	exit(0);
}

/////////////////////////////////////////////////////////
// Return a Word from CPU memory
/////////////////////////////////////////////////////////
Word romword(Word x)
{ 
	x&=0xfffe;		// drop LSB
	return((rcpubyte(x)<<8)+rcpubyte(x+1));
}

/////////////////////////////////////////////////////////
// Write a Word to CPU memory
/////////////////////////////////////////////////////////
void wrword(Word x, Word y)
{ 
	x&=0xfffe;		// drop LSB
	wcpubyte(x,(Byte)(y>>8));
	wcpubyte(x+1,(Byte)(y&0xff));
}

/////////////////////////////////////////////////////////
// Main loop for Emulation
/////////////////////////////////////////////////////////
void emulti()
{
#ifdef WIN32
	MSG msg;
#endif

	WP=romword(0);		// >0000 is the reset vector - read WP
	PC=romword(2);		// and then read the PC
	X_flag=0;			// not currently executing an X instruction
	quitflag=0;			// Don't quit

	while (!quitflag)
	{ 
		do1();			// execute one opcode

#ifdef WIN32
		// check for messages
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			if (msg.message == WM_QUIT)
				quitflag=1;
			
			TranslateMessage(&msg);
			DispatchMessage(&msg);		// this will push it to the Window procedure
		}
#endif

	}

#ifdef __linux__
	fail("STOPPED!");
#endif

}

//////////////////////////////////////////////////////////
// Read and process the TIDATA file
//////////////////////////////////////////////////////////
void readroms()
{ 
	FILE *fp, *fp2;										// file pointers for TIDATA and binary file
	int flag, idx;										// condition flag
	char type;											// type of data being load
	unsigned short address, length;						// address and length for data
	char temp[1024], temp2[1024];						// temp string

	memset(ROMMAP, 0, 65536);							// set all CPU memory to r/w
	
	fp=fopen("tidata","r");								// open TIDATA file
	
	if (fp==NULL) 
		fail("Can't open TIDATA file!");
	
	flag=1;
	while (flag)										// till we're done reading
	{
		flag=(int)fgets(qw,80,fp);						// read the next line
		
		if (qw[0]=='#')									// check for Emulator options
		{
			if (strncmp(qw,"#FRAMESKIP",10) == 0)		// Frameskip
			{
				drawspeed=atoi(&qw[10]);
				sprintf(temp, "FrameSkip: %d", drawspeed);
				debug_write(temp);
			}

			if (strncmp(qw,"#MAXIPF",7) == 0)			// maximum speed
			{
				max_ipf=atoi(&qw[7]);
				oldmax=max_ipf;
				sprintf(temp, "Max IPF: %d", max_ipf);
				debug_write(temp);
			}

			if (strncmp(qw,"#DEBUG",6) == 0)			// debug mode
			{
				debug=atoi(&qw[6]);
				if (debug)
					debug_write("Debug Mode On");
			}

#ifdef WIN32
			if (strncmp(qw, "#AVI", 4) == 0)			// AVI filename
			{
				strncpy(AVIFileName, &qw[5], 255);
				idx=0;
				while (AVIFileName[idx])
				{
					if (AVIFileName[idx]<' ')
					{
						AVIFileName[idx]='\0';
						idx--;
					}
					idx++;
				}
				sprintf(temp, "AVI Filename: %s", AVIFileName);
				debug_write(temp);
			}
#endif

			if (strncmp(qw,"#v9938",6) == 0)			// v9938 support
			{
				v9938=atoi(&qw[6]);
				if (0==v9938)
					debug_write("9938 Disabled");
			}

			if (strncmp(qw,"#speedup_float",14) == 0)	// ROM BIOS floating point speedups
			{
				speedup_float=atoi(&qw[14]);
				if (0==speedup_float)
					debug_write("Floating point speedups disabled.");
			}

			if (strncmp(qw,"#speedup_fr213",14) == 0)	// Roland's FR213 acceleration ;)
			{
				speedup_fr213=atoi(&qw[14]);
				if (0 == speedup_fr213)
					debug_write("Disabled FR213 speedup");
			}

			if (strncmp(qw,"#disk",5) == 0)				// disk support (paths)
			{
				int drive;
				char path[256];
				char dtype;
				
				strcpy(path, "");
				sscanf(qw+6, "%d %1s %s", &drive, &dtype, &path);
				if ((strlen(path)) && (drive>=1 && drive<=9) && (dtype == 'i' || dtype == 'v' || dtype == 't' || dtype == 'r')) 
				{
					strcpy(&diskpath[drive][0], path);
					disktype[drive] = dtype;
					sprintf(temp, "Using path %s type %c as DSK%d.", path, dtype, drive);
					debug_write(temp);
				}
			}
		}
		else
		{ 
			if ((flag!=(int)NULL)&&(qw[0]!=';'))		// if we have a line and it's not a comment
			{ 
				type=qw[0];								// read the type
				if ((type!='C')&&(type!='X')&&(type!='G')&&(type!='V')&&(type!='S'))
				{ 
					goto skipload;						// skip unrecognized lines
				}
      
				address=hex2dec(&qw[2]);				// get the address
				length=hex2dec(&qw[7]);					// and the length
      
				index1=11;					
				while (qw[++index1] >= ' ')	// get the last character of the filename
#ifdef __linux__
				  if (qw[index1] == '\\')
					qw[index1] = '/'; /* convert backslashes to slashes */
#endif
				;
				qw[index1]=0;							// zero terminate it (in case of spaces)

				fp2=fopen(&qw[12],"rb");				// open the memory file

				if (fp2==NULL) 
					fail("Can't open specified memory file!");
      
				sprintf(temp, "Loading %s to ",&qw[12]);// debug note
				switch (type)
				{ 
				case 'C': strcat(temp, "CPU ROM"); break;
				case 'X': strcat(temp, "XB ROM bank 2"); break;
				case 'G': strcat(temp, "GROM"); break;
				case 'V': strcat(temp, "VDP RAM"); break;
				case 'S': strcat(temp, "SPEECH ROM"); break;
				}
				sprintf(temp2, " at >%x, length >%x",address,length);
				strcat(temp, temp2);
				debug_write(temp);
      
				if (qw[6]=='-')							// TIFILES skip 128 bytes
				{ 
					debug_write("TIFILES header skip...");
					fseek(fp2, 128, SEEK_CUR);
				}

				if (qw[11]=='-')						// GRAM Kracker skip 6 bytes
				{ 
					debug_write("GRAM Kracker header skip...");
					fseek(fp2, 6, SEEK_CUR);
				}

				switch(type)
				{ 
				case 'C':	fread(&CPU[address], 1, length, fp2);
							memset(&ROMMAP[address], 1, length);
							break;
				case 'X':	fread(&CPU2[address-0x6000], 1, length, fp2);
							xb=1;
							break;
				case 'G':	fread(&GROM[address], 1, length, fp2);
							break;
				case 'V':	fread(&VDP[address], 1, length, fp2);
							break;
				case 'S':	//fread(&SPEECH[address], 1, length, fp2);	// ignore speech for now
							break;
				}

				fclose(fp2);							// done this ROM
skipload: ;
			}
		}
	}
fclose(fp);												// done all ROMS
}

//////////////////////////////////////////////////////////
// Convert hex to decimal
//////////////////////////////////////////////////////////
Word hex2dec(char *x)
{
	Word z;

	z=0;
	sscanf(x, "%4hx", &z);

	return z;
}

//////////////////////////////////////////////////////////
// Speedup routine - convert from TI floating point
//////////////////////////////////////////////////////////
double tifloat2double(Byte *p) 
{
	short int firstword;
	int i, exp, sign=1;
	double res;
  
	firstword = *p++ << 8;
	firstword |= *p++;
	if (firstword==0)
	{
		return 0;
	}
	else 
	{
		if (firstword<0) 
		{								/* negative */
			sign=-1;
			firstword = -firstword;
		}
	}
	exp = ((firstword >> 8) - 0x40 - 6) << 1;
	res = firstword & 0xff;
	for (i=0; i<6; i++)
	{
		res = 100*res + *p++;
	}
	
	return (sign * res * pow(10,exp));
}
	
//////////////////////////////////////////////////////////
// Speedup routine - convert to TI floating point
//////////////////////////////////////////////////////////
void double2tifloat(double d, Byte *p)
{
	short int firstword;
	int i, exp, sign=1;
  
	if (d==0)
	{
		*p++ = 0;
		*p = 0;
		return;
	} 
	else
	{
		if (d<0) 
		{								/* negative */
			sign=-1;
			d = -d;
		}
	}
	exp = (int) log10(d) >> 1;
	d /= pow(10,exp<<1);
	firstword = (exp + 0x40) << 8 | (Byte) d;
	if (sign<0)
	{
		firstword = -firstword;
		*p++ = firstword >> 8;
		*p++ = firstword & 0xff;
		for (i=0; i<6; i++) 
		{
			d = (d - (int) d) * 100;
		}
		*p++ = (Byte) d;
	}
}

//////////////////////////////////////////////////////////
// Interpret a single instruction
//////////////////////////////////////////////////////////
void do1()
{
#ifdef WIN32
	int x, y;
	unsigned short *col2;
	DDSURFACEDESC buf;
	long address;
	int offset;

	if ((PauseInactive)&&(myWnd != GetForegroundWindow()))
		return;										// exit if window inactive and we're supposed to pause
#endif

#ifdef __linux__
	Byte x1;
	int t1,t2;

	if (debug)
		draw_debug();								// draw the screen if in debug mode
#endif

	cpucount++;										// increment instruction counter

	// check for video interrupt or redraw
	if (interrupt_needed)
	{
		if (retrace_count>drawspeed && setpalette_needed)
		{
			VDPsetpalette();
		}

		if ((retrace_count>drawspeed)&&(redraw_needed))	// only draw if VDP RAM has been changed, and it's time
		{ 
#ifdef __linux__
			VDPdisplay();									/* this skews the accounting a bit, but */
			VDPS[0] |= 0x80;
		}
#endif
#ifdef WIN32
			if (DisplayEvent)
				SetEvent(DisplayEvent);

			redraw_needed=0;
		}
	
		if (debug)
			SetEvent(BlitEvent);						// draw debug screen
#endif

		timercount+=retrace_count;						// update timer statistics

		if (retrace_count>drawspeed)
		{
			cpuframes++;								// for accounting
			retrace_count=0;							// reset retrace counter
			instruction_count=0;						// for throttling
			last_count=0;								// to verify changes above

#ifdef WIN32
			if (Recording)								// for AVI recording
			{
				// write a frame to the AVI
				// I want only every 4th frame.
				RecordFrame++;
				if (RecordFrame >= 4/(drawspeed+1))
				{
					RecordFrame=0;
					
					// don't single step this code!! Well, NOSYSLOCK might make it possible (cause it's not the primary surface)
					ZeroMemory(&buf, sizeof(buf));
					buf.dwSize=sizeof(buf);
					if (DD_OK == BackBuffer->Lock(NULL, &buf, DDLOCK_WAIT | DDLOCK_NOSYSLOCK | DDLOCK_READONLY | DDLOCK_SURFACEMEMORYPTR, NULL))
					{
						col2=(unsigned short *)buf.lpSurface;
						// scan and copy to the frame buffer
						for (y=0; y<192; y++)
						{
							// NOTE! Assumes 16 bit, but gets alignment right
							address=(y+4)*(buf.lPitch >> 1) + 32;		// it's in words, not bytes
							offset=(191-y)*256;

							for (x=0; x<256; x++)
							{
								(unsigned short)framedata[offset+x]=col2[address+x];
							}
						}
						BackBuffer->Unlock(NULL);
						WriteFrame();
					}
					// safe to single step again here
				}
			}
#endif
		}

		if ((ST&0x0002)&&(VDPREG[1]&0x20))				// if interrupt mask on and VDP ints on
		{	
			if (interrupt_needed)						// now flagged by the timer routine
			{	
				if (instruction_count < max_ipf-1)		// speed throttling
					ConsoleInterrupt();					// Call console interrupt routine
			}
		}
	}

	////////////////// System Patches

	// Quick hack to force joysticks to work 
	if ((PC==0x478) && ((CPU[0x8074]==2)||(CPU[0x8074]==1)))		// fill in joystick settings during kscan(1|2)
	{
		if (key[VK_UP])
			CPU[0x8076]=0x04;
		if (key[VK_DOWN])
			CPU[0x8076]=0xfc;
		if (key[VK_LEFT])
			CPU[0x8077]=0xfc;
		if (key[VK_RIGHT])
			CPU[0x8077]=0x04;
	}

	// intercept patches for Ami99 DSK1 emulation
	if (PC==0x4800)									// special address to start DSK1 emulation
	{
		do_dsrlnk();
		PC=romword(WP+22)+2;						// return address in R11.. TI is a bit silly
													// and if we don't increment by 2, it's considered
													// an error condition
		return;
	}

	if (PC == 0x4810) 
	{												// powerup 
		wrword(0x8358, 0x3ef5);						// volume ID block
		wrword(0x8366, 0x3eea);						// VDP stack
		do_files(3);
		PC=romword(WP+22);
		return;
	}

	if (PC>=0x4820 && PC<=0x482c)					// special intercept address for emulated ROM
	{
		do_sbrlnk();
		PC=romword(WP+22)+2;						// return address in R11.. as above
		return;
	}

/************************for the keyboard processor */

	// debug mode
	if (!debug && key[VK_HOME]) 
	{
		debug=1;
		redraw_needed=1;
#ifdef __linux__
		if (!v9938) {
		  fprintf(stderr, "entering debug mode\n");
		  if (set_gfx_mode(GFX_DIRECTX,640,480,0,0)<0)
			fail("Can't set 640x480x8 graphics mode!");
		  clear_keybuf();	/* maybe didnt get release event */
		}
#endif
	}
	if (debug && key[VK_END]) 
	{
		debug=0;
		redraw_needed=1;
#ifdef __linux__
		if (!v9938) {
		  fprintf(stderr, "leaving debug mode\n");
		  if (set_gfx_mode(GFX_DIRECTX,320,200,0,0)<0)
			if (set_gfx_mode(GFX_DIRECTX,640,480,0,0)<0)
			  fail("Can't set 320x200x8 or 640x480x8 graphics mode!");
		  clear_keybuf();	/* maybe didnt get release event */
		}
#endif
	}
	
    // breakpoint handling
    if (breakpoint==PC)
	  max_ipf=0;

	// Single-stepping
	if (key[VK_F1])
	{	
		if (0 == max_ipf)
		{
			max_ipf=oldmax;
#ifdef WIN32
			SetWindowText(myWnd, "Ami99");
#endif
		}
		else
		{
			max_ipf=0;
#ifdef WIN32
			SetWindowText(myWnd, "Ami99 - Paused. F1 to Continue, F2 to Step");
#endif
		}
		key[VK_F1]=0;					// I shouldn't do this. Naughty me.
	}
	if (key[VK_F2])
	{
		instruction_count=-2;			// this will allow 1 instruction to run
		key[VK_F2]=0;					// naughty me again
	}									// negative 2 because it counts before it checks

/******************************************/

	if (speedup_fr213)					// speedup fractal calculation in Rolands program ;)
	{
		if (PC==0xbcd8 && CPU[PC] == 0xc8) 
		{								// Mandel 
			double xu = tifloat2double(&CPU[0xf64e]);
			double yu = tifloat2double(&CPU[0xf65e]);
			double n = tifloat2double(&CPU[0x8040]);
			double m = tifloat2double(&CPU[0x8038]);
			double dx = tifloat2double(&CPU[0x8080]);
			double dy = tifloat2double(&CPU[0x8088]);
			double xc = xu + n * dx;
			double yc = yu + m * dy;
			double xz=0, yz = 0, xx, yy;
			int k=0;
			int kx = (CPU[WP+14]<<8) + CPU[WP+15]; /* R7 */
	   
			do 
			{
				k++;
				xx = xz * xz;
				yy = yz * yz;
				if (k > kx)
					break;
				yz = 2 * xz * yz - yc;
				xz = xx - yy - xc;
			} while (xx+yy < 100);

			CPU[WP+12] = k >> 8;
			CPU[WP+13] = k & 0xff;
			PC=romword(WP+22);			// return address in R11.
			return;
		}

		if (PC==0xd198 && CPU[PC] == 0x04) 
		{								// Iter
			int r0 = romword(WP);
			int i0 = romword(WP+2);
			int kx = romword(WP+12);
			int k=0, rn, in, rq, iq;

			if (r0 & 0x8000)			// negative
				r0 -= 0x10000;
			if (i0 & 0x8000)
				i0 -= 0x10000;
			rn = r0;
			in = i0;
			while(1) 
			{
				rq = (rn * rn) >> 12;
				iq = (in * in) >> 12;
				if (rq + iq >= 0x4000)
					break;
				if (++k >= kx)
					break;
				in = i0 + ((rn * in) >> 11);
				rn = r0 + rq - iq;
			}
			CPU[WP+12] = k >> 8;
			CPU[WP+13] = k & 0xff;
			PC += 4;
			return;
		}

		if (PC==0xd28a && CPU[PC] == 0x04) 
		{								// Iter32 
			int r0 = romword(WP+4);
			int i0 = romword(WP);
			int kx = romword(WP+12);
			int rn, in, rq, iq;
			int k=0;

			r0 = (r0 << 16) + romword(WP+6);
			i0 = (i0 << 16) + romword(WP+2);
			if (r0 <0)
			  r0 = -r0;
			if (i0 <0)
			  i0 = -i0;
			rn = r0;
			in = i0;
			while(1) 
			{
				rq = ((__int64)rn * rn) >> 28;
				iq = ((__int64)in * in) >> 28;
				if ((__int64)rq + iq >= 0x40000000)
					break;
				if (++k >= kx)
					break;
				in = i0 + (((__int64)rn * in) >> 27);
				rn = r0 + rq - iq;
			}
			CPU[WP+12] = k >> 8;
			CPU[WP+13] = k & 0xff;
			PC += 4;
			return;
		}
	} /* speedup_fr213 */

	if (speedup_float) 
	{									/* speedup floating point operations */
		if (PC==0x0d80) {					/* ADD */
			double2tifloat(tifloat2double(&CPU[0x805c]) + tifloat2double(&CPU[0x804a]), &CPU[0x804a]);
			PC=romword(WP+22);			// return address in R11.
			return;
		}
		if (PC==0x0d7c) {					/* SUBT */
			double2tifloat(tifloat2double(&CPU[0x805c]) - tifloat2double(&CPU[0x804a]), &CPU[0x804a]);
			PC=romword(WP+22);			// return address in R11.
			return;
		}
		if (PC==0xe88) {					/* MULT */
			double2tifloat(tifloat2double(&CPU[0x805c]) * tifloat2double(&CPU[0x804a]), &CPU[0x804a]);
			PC=romword(WP+22);			// return address in R11.
			return;
		}
		if (PC==0x0ff4) {					/* DIV */
			double2tifloat(tifloat2double(&CPU[0x805c]) / tifloat2double(&CPU[0x804a]), &CPU[0x804a]);
			PC=romword(WP+22);			// return address in R11.
			return;
		}
		if (PC==0x12b8) {					/* CFI */
			wrword(0x834a, (Word) tifloat2double(&CPU[0x804a]));
			PC=romword(WP+22);			// return address in R11.
			return;
		}
	} /* speedup_float */

///////////////// End of patches 

	instruction_count++;

	if (instruction_count<max_ipf)					// speed throttling
	{
		if (X_flag==0)
		{ 
			in=romword(PC);							// ie: not an 'X' command
		    PC=PC+2;								// thanks to Jeff Brown for explaining that!
		}

		(*opcode[in])();							// 'in' contains the opcode to execute
	}
#ifdef __linux__
	else
	{
		vsync();									// wait for next
	}
#endif

}

//////////////////////////////////////////////////////
// Read a single byte from CPU memory
//////////////////////////////////////////////////////
Byte rcpubyte(Word x)
{
	// TI CPU memory map
	// >0000 - >1fff  Console ROM
	// >2000 - >3fff  Low bank RAM
	// >4000 - >5fff  DSR ROMs
	// >6000 - >7fff  Cartridge expansion
	// >8000 - >9fff  Memory-mapped devices & CPU scratchpad
	// >a000 - >ffff  High bank RAM
	// 
	// All is fine and dandy, except the memory-mapped devices, and the
	// fact that writing to the cartridge port with XB in places causes
	// a bank switch. In this emulator, that will only happen if bank 2
	// has been loaded. A custom DSR will be written to be loaded which
	// will appear to the "TI" to support all valid devices, and it will
	// be loaded into the DSR space, which then will not be paged.

	// ordered roughly by how often each bank is accessed, give or take ;)

	if ((x>=0x8000)&&(x<=0x83ff))
	{ 
		x&=0xf0ff;										// scratchpad RAM - 256 bytes repeating.
		return(CPU[x]);									// I map it all to >80xx
	}

	if ((x<0x4000)||(x>0x9fff)) 
		return(CPU[x]);									// normal CPU RAM

	// XB is supposed to only page the upper 4k, but some other carts seem to like it all
	// paged? Dunno.. weird.

	if ((x>=0x6000)&&(x<0x8000))
		if ((xb)&&(bank)) 
			return(CPU2[x-0x6000]);						// cartridge bank 2
		else 
			return(CPU[x]);								// cartridge bank 1

	if ((x>=0x9800)&&(x<=0x9fff))
	{
		return(rgrmbyte(x));							// read GROM data
	}

	if ((x>=0x8800)&&(x<=0x8fff))
	{
		return(rvdpbyte(x));							// VDP read data
	}

	if ((x>=0x4000) && (x<=0x5fff)) 
		return(CPU[x]);									// DSR ROM (no bank switching)

	if ((x>=0x9000)&&(x<=0x97ff))
	{
		return(rspeechbyte(x));							// Speech read data
	}

	if ((x>=0x8400)&&(x<=0x87ff))
	{
		return((Byte)0);								// Don't read the sound chip (can hang a real TI)
	}

	// We shouldn't get here, but just in case...
	return(0);
}

//////////////////////////////////////////////////////////
// Write a byte to CPU memory
//////////////////////////////////////////////////////////
void wcpubyte(Word x, Byte c)
{

	if (ROMMAP[x]) 
	{
		return;											// can't write to ROM!
	}

	if ((x>=0x8000)&&(x<=0x83ff))
	{ 
		x&=0xf0ff;										// Scratchpad RAM
		CPU[x]=c;
		return;
	}

	if ((x<0x8000)||(x>0x9fff))
	{ 
		CPU[x]=c;										// CPU RAM/ROM
		return;
	}

	if ((x>=0x8800)&&(x<=0x8fff))
	{ 
		wvdpbyte(x,c);									// Write to VDP
		return;
	}

	if ((x>=0x8400)&&(x<=0x87ff))
	{ 
		wsndbyte(c);									// Sound chip
		return;
	}

	if ((x>=0x6000)&&(x<0x8000)&&(xb))		
	{ 
		bank=(x&0x0002)>>1;								// XB bank switch
		return; 
	}
 
	if ((x>=0x9000)&&(x<=0x97ff))
	{	
		wspeechbyte(x,c);								// Write to speech chip
		return; 
	}

	if ((x>=0x9800)&&(x<=0x9fff))
	{ 
		wgrmbyte(x,c);									// Write to GROM
		return;
	}
}

//////////////////////////////////////////////////////
// Read a byte from the speech synthesizer
//////////////////////////////////////////////////////
Byte rspeechbyte(Word x)
{
	return 0;											// no speech
}

//////////////////////////////////////////////////////
// Write a byte to the speech synthesizer
//////////////////////////////////////////////////////
void wspeechbyte(Word x, Byte c)
{
	// that's nice :)
}

//////////////////////////////////////////////////////
// Increment VDP Address
//////////////////////////////////////////////////////
void increment_vdpadd() 
{
	char temp[1024];

	if ((++VDPADD & 0x3fff) == 0) 
	{ /* increment, handle overflow */
		VDPADD ^= 0x4000;			/* restore bit 14 */
		/* workaround for compatibility problem: inc. bank only in new modes */
		if (mode==G1 || mode==G2 || mode==T1 || mode==MC) 
		{
			sprintf(temp, "VDP access: suppessed bank increment at PC 0x%04X", PC);
			debug_write(temp);
		} else 
		{
			VDPADD &= 0x4000;
			VDPREG[14] = (VDPREG[14]+1) & 0x07;
			sprintf(temp, "VDP access: bank increment at PC 0x%04X", PC);
			debug_write(temp);
		}
	}
}

//////////////////////////////////////////////////////////////
// Read from VDP chip
//////////////////////////////////////////////////////////////
Byte rvdpbyte(Word x)
{ 
#ifdef WIN32
	int mouse_b;

	mouse_b=MouseB;											// TODO - get state of Windows mouse buttons
															// bit 0=left, bit 1=right
#endif

	if ((x>=0x8c00) || (x&1))
	{
		return(0);											// write address
	}

	if (v9938) 
	{
		if (x & 2) 
		{													// read VDP status register
			Byte z;
			Byte s = VDPREG[15] & 0x0f;

			if (v9938) 
			{
				VDPS[1] = VDPS[1] & 0x3f | (mouse_b << 6);	// update mouse buttons - FIXME: check enable
				if (s==3 || s==5) 
				{											// get mouse position - FIXME: check enable
#ifdef __linux__
					int dx, dy;								// difference of mouse position since last call
					get_mouse_mickeys(&dx, &dy);			// get differences since last query 
					mx+=dx; my+=dy;							// update positions 
#endif
#ifdef WIN32
					mx=MouseX;								// Clipped to 256, just cheat...
					my=MouseY;
#endif
					if (s==3) 
					{										// get x difference
						z = mx - oldmx;
						oldmx = mx;
						return z;
					} else 
					{										// get y difference 
						z = my - oldmy;
						oldmy = my;
						return z;
					}
				}
			}	/* v9938 */
	  
			z = VDPS[s];
			VDPS[s] = 0;
			return(z);
		} else 
		{													// read VDP memory
			int adr = ((VDPREG[45] & 0x40) << (17-5)) | ((VDPREG[14] & 0x07) << 14) | VDPADD & 0x3fff;
			Byte dat = VDP[adr];

			increment_vdpadd();
			return(dat);
		}
	} else 
	{		/* v9929 */
		unsigned short z;
		if (x&0x0002)
		{	/* read status */
			z=VDPS[0];
			VDPS[0]=0;
			return((Byte)z);
		}
		else
		{ /* read data */
			z=VDP[VDPADD&0x3fff];
			VDPADD++;
			return ((Byte)z);
		}
	}
}

///////////////////////////////////////////////////////////////
// Write to VDP chip
///////////////////////////////////////////////////////////////
void wvdpbyte(Word x, Byte c)
{
	if (x<0x8c00 || (x&1)) 
	{
		return;							/* not going to write at that block */
	}

	if (v9938) 
	{
		if ((x&6) == 2) 
		{			/* write VDP address */
			VDPADD = (VDPADD>>8) | (c<<8);
			vdpaccess++;
			if (vdpaccess==2) 
			{
				if (VDPADD & 0x8000) 
				{	/* write VDP register */
					wVDPreg((Byte)((VDPADD&0x3f00)>>8),(Byte)(VDPADD&0x00ff));
					// redraw_needed=1;
				}
				vdpaccess=0;
			}
		} else 
		{	
			if ((x&6) == 0) 
			{		/* write VDP data */
				int adr;

				// undocumented feature: if address not set for writing, inc. address before writing
				if ((VDPADD & 0x4000) == 0) 
				{
					VDPADD |= 0x4000;
					increment_vdpadd();
				}
				adr = ((VDPREG[45] & 0x40) << (17-5)) | ((VDPREG[14] & 0x07) << 14) | VDPADD & 0x3fff;
				VDP[adr] = c;
				increment_vdpadd();
				// trigger necessary updates
				if (CTsize)
				{
					if (adr >= CT && adr < CT+CTsize) 
					{
						redraw_needed = 1;
						if (mode == G1)
						{
							setpalette_needed = 1;
						}
					}
				}

				if (SITsize)
				{
					if (adr >= SIT && adr < SIT+SITsize)
					{
						redraw_needed = 1;
					}
				}

				if (PDTsize)
				{
					if (adr >= PDT && adr < PDT+PDTsize) 
					{
						redraw_needed = 1;
						rebuild_patterns_needed = 1;
					}
				}

				if (SALsize)
				{
					if (adr >= SAL && adr < SAL+SALsize)
					{
						redraw_needed = 1;
					}
				}

				if (SDTsize)
				{
					if (adr >= SDT && adr < SDT+SDTsize)
					{
						redraw_needed = 1;
					}
				}

				rebuild_patterns_needed = 1;						/* FIXME: doesnt work :( */
				redraw_needed = 1;									/* for some reason I need this.. especially in Demo cartridge */
			} else 
			{
				if ((x&6) == 4) 
				{		/* write VDP color palette */
					if (firstcolorbyte<0)
					{
						firstcolorbyte = c;
					}
					else 
					{
						TIPAL[VDPREG[16]&0x0f].r = (firstcolorbyte & 0x70) >> 1;
						TIPAL[VDPREG[16]&0x0f].b = (firstcolorbyte & 0x07) << 3;
						TIPAL[VDPREG[16]&0x0f].g = (c & 0x07) << 3;
						firstcolorbyte = -1;
						setpalette_needed = 1;
					}
				} else 
				{					/* write VDP register indirect */
					Byte r = VDPREG[17] & 0x3f;
					if (r != 17) 
					{
						wVDPreg(r, c);
						if ((VDPREG[17] & 0x80) == 0)
						{
							VDPREG[17] = (VDPREG[17]+1) & ~0x40;
						}
					}
				}
			}
		}
	} else 
	{						/* v9929 */
		if (x&0x0002)
		{	/* write address */
			VDPADD=(VDPADD>>8)|(c<<8);
			vdpaccess++;
			if (vdpaccess==2)
			{ 
				if (VDPADD&0x8000)
				{ 
					wVDPreg((Byte)((VDPADD&0x0700)>>8),(Byte)(VDPADD&0x00ff));
					redraw_needed=1;
				}
				vdpaccess=0;
			}
		}
		else
		{	/* write data */
			VDP[(VDPADD++)&0x3fff]=c;
			redraw_needed=1;
		}
	}
}

////////////////////////////////////////////////////////////////
// Set a pixel in video memory
////////////////////////////////////////////////////////////////
void pset(int dx, int dy, int c, int a, int l) 
{	/* set pixel */
	int p, mask, val;			/* VDP base address, mask, value */

	switch (mode) 
	{
	case G4: 
		if (a & 0x20)			/* MXD */
		{
			dy += 1024;
		}
		p = (dy<<7) + (dx>>1);
		mask = (dx&1) ? 0x0f : 0xf0;
		val = (dx&1) ? c : c<<4;
		break;
	
	case G5:
		if (a & 0x20)			/* MXD */
		{
			dy += 1024;
		}
		p = (dy<<7) + (dx>>2);
		mask = 3 << (6-2*(dx&3));
		val = c << (6-2*(dx&3));
		break;

	case G6:
		if (a & 0x20)			/* MXD */
		{
			dy += 512;
		}
		p = (dy<<8) + (dx>>1);
		mask = (dx&1) ? 0x0f : 0xf0;
		val = (dx&1) ? c : c<<4;
		break;
	
	case G7:
		if (a & 0x20)			/* MXD */
		{
			dy += 512;
		}
		p = (dy<<8) + dx;
		mask = 0xff;
		val = c;
		break;
	
	default:					/* ignore command in other modes */
		return;
	} /* switch */

	val &= mask;

	if (val || (l&0x08==0))
	{
		switch (l & 0x07) 
		{
		case 0:					/* IMP */
			VDP[p] = VDP[p] & ~mask | val;
			break;
		case 1:					/* AND */
			VDP[p] &= (~mask | val);
			break;
		case 2:					/* OR */
			VDP[p] |= val;
			break;
		case 3:					/* EOR */
			VDP[p] ^= val;
			break;
		case 4:					/* NOT */
			VDP[p] &= (~mask | ~val);
			break;
		default:					/* ignore illegal logic */
			break;
		} /*  switch l */
	}

	if (!redraw_needed)
	{
		if (VDPREG[1] & 0x40)
		{
			// redraw only necessary area immediately, if screen on and no redraw coming anyway
			switch (mode)
			{
			case G4:	VDPgraphics4(dx&0xfe, dy&0xff, dx&0xfe, dy&0xff); break;
			case G5:	/* VDPgraphics5(dx&0x1fc, dy&0xff, dx&0x1fc, dy&0xff); */ break;	// FIXME: G5 not yet implemented
			case G6:	VDPgraphics6(dx&0x1fe, (dy&0xff)<<1, dx&0x1fe, (dy&0xff)<<1); break;
			case G7:	VDPgraphics7((dx&0xff)<<1, (dy&0xff)<<1, (dx&0xff)<<1, (dy&0xff)<<1); break;
			}
	   }
	}
}

////////////////////////////////////////////////////////////////
// Write to VDP Register
////////////////////////////////////////////////////////////////
void wVDPreg(Byte r, Byte v)
{ 
	int t;
	char temp[1024];

	VDPREG[r]=v;

	if (r==7)
	{	/* color of screen, set color 0 (trans) to match */
		t=v&0xf;
		TIPAL[0].r=TIPAL[t].r;
		TIPAL[0].g=TIPAL[t].g;
		TIPAL[0].b=TIPAL[t].b;
		// TODO WIN32 - Update PenBuffer for color 0
	}

	if (r<=1)						/* mode switched */
	{
		mode = ((VDPREG[0]&0x0e)<<1) | ((VDPREG[1]&0x18)>>3);
	}

	/* workaround for compatibility problem: higher banks only from user programs */
	if (v9938 && r==2 && PC<0x2000 && v&0xf0) 
	{
		sprintf(temp, "PC 0x%04X: writing 0x%02x instead of 0x%02x to VDP register 2", PC, v&0x0f, v);
		debug_write(temp);
		VDPREG[r] = v & 0x0f;
	}

	if (v9938 && r==4 && PC<0x2000 && v&0xf8) 
	{
		sprintf(temp, "PC 0x%04X: writing 0x%02x instead of 0x%02x to VDP register 4", PC, v&0x07, v);
		debug_write(temp);
		VDPREG[r] = v & 0x07;
	}

	if (r<=13) 
	{
		gettables();
		redraw_needed=1;
		setpalette_needed = 1;
		rebuild_patterns_needed = 1;
	}

	if (r==46 && ((mode==G4) || (mode==G5) || (mode==G6) || (mode==G7))) 
	{	/* execute command */
		int sx = ((VDPREG[33]&0x01)<<8) | VDPREG[32];
		int sy = ((VDPREG[35]&0x03)<<8) | VDPREG[34];
		int dx = ((VDPREG[37]&0x01)<<8) | VDPREG[36];
		int dy = ((VDPREG[39]&0x03)<<8) | VDPREG[38];
		int nx = ((VDPREG[41]&0x03)<<8) | VDPREG[40]; /* Bit 1+8 also */
		int ny = ((VDPREG[43]&0x03)<<8) | VDPREG[42];
		int c = VDPREG[44];			/* color */
		int a = VDPREG[45];			/* argument */
		int l = v & 0x0f;			/* logical operation */

		int p, ps, pd, dp, n, mask, count;				/* VDP base address source/dest, increment, number of bytes */

		switch (v & 0xf0) 
		{			/* command */

		// POINT
		case 0x40: 						/* POINT */
			switch (mode) 
			{
			case G4: 
				if (a & 0x10)			/* MXD */
				{
					sy += 1024;
				}
				p = (sy<<7) + (sx>>1);
				mask = 0x0f;
				count = (sx&1) ? 0 : 4;
			break;

			case G5:
				if (a & 0x10)				/* MXD */
				{
					sy += 1024;
				}
				p = (sy<<7) + (sx>>2);
				mask = 0x03;
				count = 6-2*(sx&3);
			break;
		
			case G6: 
				if (a & 0x10)				/* MXD */
				{
					sy += 512;
				}
				p = (sy<<8) + (sx>>1);
				mask = 0x0f;
				count = (sx&1) ? 0 : 4;
			break;

			case G7: 
				if (a & 0x10)				/* MXD */
				{
					sy += 512;
				}
				p = (sy<<8) + sx;
				mask = 0xff;
				count = 0;
			break;

			default:						/* ignore command in other modes */
				return;
			} /* switch mode */

			VDPS[7] = (VDP[p] >> count) & mask;
		break;
   
		// PSET
		case 0x50: 							/* PSET */
			pset(dx, dy, c, a, l);
		break;
	 
		// LINE
		case 0x70:							/* LINE */
			if (nx) 
			{								/* ignore invalid count */
				int x, y;
	   
				if (a & 0x01) 
				{							/* y side longer, ny <-> nx! */
					if (a & 0x08)			/* DY */
					{
						dy -= (nx-1);
					}

					if (a & 0x04)			/* DX */
					{
						dx -= (ny-1);
					}

					for (y=dy; y<dy+nx; y++)
					{
						pset(dx + ny * (y-dy) / nx, y, c, a, l);
					}
				} else 
				{							/* x side longer */
					if (a & 0x08)			/* DY */
					{
						dy -= (ny-1);
					}

					if (a & 0x04)			/* DX */
					{
						dx -= (nx-1);
					}

					for (x=dx; x<dx+nx; x++)
					{
						pset(x, dy + ny * (x-dx) / nx, c, a, l);
					}
				}
			} /* if nx */
			break;
	 
		// HMMV
		case 0xc0:							/* HMMV */
			if (a & 0x08)					/* DY */
			{
				dy -= (ny-1);
			}

			if (a & 0x04)					/* DX */
			{
				dx -= (nx-1);
			}

			switch (mode) 
			{
				case G4: 
					p = (dy<<7) + (dx>>1);
					dp = 128;
					n = nx>>1;
				break;

				case G5: 
					p = (dy<<7) + (dx>>2);
					dp = 128;
					n = nx>>2;
				break;

				case G6: 
					p = (dy<<8) + (dx>>1);
					dp = 256;
					n = nx>>1;
				break;

				case G7: 
					p = (dy<<8) + dx;
					dp = 256;
					n = nx;
				break;

				default:					/* ignore command in other modes */
					return;
			} /* switch mode */
	 
			if (a & 0x20)					/* MXD */
			{
				p += 0x20000;
			}

			sprintf(temp,"PC 0x%4X: HMMV dx=0x%02x dy=0x%02x nx=0x%02x ny=0x%02x c=0x%02x", PC, dx, dy, nx, ny, c);
			debug_write(temp);

			for ( ; ny; ny-- , p+=dp)
			{
				memset(&VDP[p], c, n);
			}

			redraw_needed=1;
		break;

		// HMMM
		case 0xd0:							/* HMMM */
			if (a & 0x08) 
			{								/* DIY */
				sy -= (ny-1);
				dy -= (ny-1);
			}

			if (a & 0x04) 
			{								/* DIX */
				sx -= (nx-1);
				dx -= (nx-1);
			}

			switch (mode) 
			{
				case G4: 
					ps = (sy<<7) + (sx>>1);
					pd = (dy<<7) + (dx>>1);
					dp = 128;
					n = nx>>1;
				break;

				case G5:
					ps = (sy<<7) + (sx>>2);
					pd = (dy<<7) + (dx>>2);
					dp = 128;
					n = nx>>2;
				break;
	 
				case G6:
					ps = (sy<<8) + (sx>>1);
					pd = (dy<<8) + (dx>>1);
					dp = 256;
					n = nx>>1;
				break;

				case G7: 
					ps = (sy<<8) + sx;
					pd = (dy<<8) + dx;
					dp = 256;
					n = nx;
				break;
	 
				default:					/* ignore command in other modes */
					return;
			} /* switch mode */

			if (a & 0x10)				/* MXS */
			{
				ps += 0x20000;
			}

			if (a & 0x20)				/* MXD */
			{
				pd += 0x20000;
			}

			sprintf(temp,"PC 0x%4X: HMMM sx=0x%02x sy=0x%02x dx=0x%02x dy=0x%02x nx=0x%02x ny=0x%02x", PC, sx, sy, dx, dy, nx, ny);
			debug_write(temp);
	 
			for ( ; ny; ny-- , ps+=dp, pd+=dp)
			{
				memmove(&VDP[pd], &VDP[ps], n);
			}
	 
			redraw_needed=1;
		break;

		// unknown
		default:
			sprintf(temp,"PC 0x%4X: ignoring unsupported command 0x%02x", PC, v);
			debug_write(temp);
		break;
		} /* switch v */
	} /* if r==46 */
}

////////////////////////////////////////////////////////////////
// Write a byte to the sound chip
////////////////////////////////////////////////////////////////
void wsndbyte(Byte c)
{
	unsigned int x, idx;								// temp variable
	static int noise_volume=-10000;						// volume of noise
	static int noise_type=0;							// type of noise
	static int freq3=22000;								// frequency calculator

	#define mult -500

#ifdef WIN32
	if (NULL == lpds) return;
#endif

	// 'c' contains the byte currently being written to the sound chip
	// all functions are 1 or 2 bytes long, as follows					
	//
	// BYTE		BIT		PURPOSE											
	//	1		0		always '1'										
	//			1-3		Operation:	000 - tone 1 frequency				
	//								001 - tone 1 volume					
	//								010 - tone 2 frequency				
	//								011 - tone 2 volume					
	//								100 - tone 3 frequency				
	//								101 - tone 3 volume					
	//								110 - noise control					
	//								111 - noise volume					
	//			4-7		Least sig. frequency bits for tone, or volume	
	//					setting (0-F), or type of noise.				
	//					(volume F is off)								
	//					Noise set:	4 - always 0						
	//								5 - 0=periodic noise, 1=white noise 
	//								6-7 - shift rate from table, or 11	
	//									to get rate from voice 3.		
	//	2		0-1		Always '0'. This byte only used for frequency	
	//			2-7		Most sig. frequency bits						
	//
	// Commands are instantaneous

	switch (c&0xf0)										// check command
	{	
	case 0x80:											// Voice 1 frequency
		last_byte=c;
		need_byte=1;									// needs a second byte
		break;

	case 0x90:
#ifdef WIN32
		x=(c&0x0f);										// Voice 1 volume
		voice[0]->SetVolume(x*(mult));
#endif
#ifdef __linux__
		x=(c&0x0f)<<4;
		voice_set_volume(voice[0],0xf0-x);
#endif
		break;

	case 0xa0:
		last_byte=c;									// Voice 2 frequency
		need_byte=2;									// needs a second byte
		break;

	case 0xb0:
#ifdef WIN32
		x=(c&0x0f);										// Voice 2 volume
		voice[1]->SetVolume(x*(mult));
#endif
#ifdef __linux__
		x=(c&0x0f)<<4;
		voice_set_volume(voice[1],0xf0-x);
#endif
		break;

	case 0xc0:
		last_byte=c;									// Voice 3 frequency
		need_byte=3;									// needs a second byte
		break;

	case 0xd0:
#ifdef WIN32
		x=(c&0x0f);										// Voice 3 volume
		voice[2]->SetVolume(x*(mult));
#endif
#ifdef __linux__
		x=(c&0x0f)<<4;
		voice_set_volume(voice[2],0xf0-x);
#endif
		break;

	case 0xe0:
		x=(c&0x07);										// Noise - get type
#ifdef WIN32
		for (idx=0; idx<8; idx++)
		{
			if (idx==x)
				noise[idx]->SetVolume(noise_volume);
			else
				noise[idx]->SetVolume(-10000);			// turn off noises, except the one we want
		}
#endif
#ifdef __linux__
		deallocate_voice(voice[3]);
		voice[3]=allocate_voice(spl[1+x]);
		voice_set_playmode(voice[3], PLAYMODE_LOOP);
		noise_type=x+1;
		voice_start(voice[3]);							// join in
		if ((noise_type==4) || (noise_type==8))
		{
			// variable frequency noise
			voice_set_frequency(voice[3],freq3);
		}
#endif
		noise_type=x;
		break;

	case 0xf0:											// Noise volume
#ifdef WIN32
		x=(c&0x0f);
		noise[noise_type]->SetVolume(x*(mult));
		noise_volume=x*(mult);
#endif
#ifdef __linux__
		x=(c&0x0f)<<4;
		voice_set_volume(voice[3],0xf0-x);
#endif
		break;

	default:											// unknown byte
		if (need_byte)									// if we were waiting for it...
		{
			x=((last_byte&0x0f) | ((c&0x3f)<<4)) + 1;	// Get frequency code
			x=(111860/x)*22050;							// Make sample rate in HZ (22khz samples)
			x/=550;										// adjust to balance (sample recorded as 550hz)
#ifdef WIN32
			voice[need_byte-1]->SetFrequency(x);
#endif
#ifdef __linux__
			voice_set_frequency(voice[need_byte-1], x);
#endif
			if (need_byte==3)							// If voice 3, save and adjust for noise channel
			{
				freq3=x;								
#ifdef WIN32
				noise[3]->SetFrequency(freq3);
				noise[7]->SetFrequency(freq3);
#endif
#ifdef __linux__
				if ((noise_type==4) || (noise_type==8))
				{
					// also set noise channel
					voice_set_frequency(voice[3], freq3);
				}
#endif
			}
			need_byte=0;								// don't need another byte
		}
		break;
	}
}

//////////////////////////////////////////////////////////////////
// Read a byte from GROM
//////////////////////////////////////////////////////////////////
Byte rgrmbyte(Word x)
{
	unsigned int z;										// temp variable

	if (x>=0x9c00)
	{
		return(0);										// write address
	}

	if (x&0x0002)
	{
		grmaccess=2;									// read GROM address
		z=(GRMADD&0xff00)>>8;
		GRMADD=(((GRMADD&0xff)<<8)|(GRMADD&0xff));		// read is destructive
		return((Byte)z);
	}
	else
	{
		grmaccess=2;									// read data
		z=grmdata;
		grmdata=GROM[GRMADD++];
		return((Byte)z);
	}
}

//////////////////////////////////////////////////////////////////
// Write a byte to GROM
//////////////////////////////////////////////////////////////////
void wgrmbyte(Word x, Byte c)
{
	if (x<0x9c00) 
	{
		return;											// read address
	}

	if (x&0x0002)
	{
		GRMADD=(GRMADD<<8)|(c&0x00ff);					// write GROM address
		grmaccess--;
		if (grmaccess==0)
		{ 
			grmaccess=2;								// prefetch emulation
			grmdata=GROM[GRMADD++];
		}
	}
	else
	{
		warn("Writing to GROM!!");						// write to GROM not supported
		GROM[GRMADD++]=c;								// write data
	}
}

//////////////////////////////////////////////////////////////////
// Write a bit to CRU
//////////////////////////////////////////////////////////////////
void wcru(Word ad, int bt)
{
	ad=(ad&0x0fff);										// get actual CRU line
	
	if (bt)												// write the data
		CRU[ad]=1;
	else
		CRU[ad]=0;
}

//////////////////////////////////////////////////////////////////
// Read a bit from CRU
// The CRU is kinda like black magic and is weird to emulate
//////////////////////////////////////////////////////////////////
int rcru(Word ad)
{
	int ret,col;										// temp variables

	// only certain CRU bits are implemented, check them

	ad=(ad&0x0fff);										// get actual CRU line
	ret=1;												// default return code (false)

	// keyboard/joystick. PC Keyboard is currently *NOT* remapped		
	// so that it's keys work.	Joysticks are very badly whacked		
	
	// keyboard reads as an array. Bits 24, 26 and 28 set the line to	
	// scan (columns). Bits 6-14 are used for return. 0 means on.		
	// The address was divided by 2 before being given to the routine	

	if ((ad>=0x03)&&(ad<=0x0a))
	{	
		if ((ad==0x07)&&(CRU[0x15]==0))					// is it ALPHA LOCK?
		{	
			ret=1;
#ifdef __linux__
			if (key_shifts & KB_CAPSLOCK_FLAG) /* check CAPS LOCK */
#endif
#ifdef WIN32
			if (GetKeyState(VK_CAPITAL) & 0x01)			// check CAPS LOCK (on?)
#endif
			{	
				ret=0;									// set Alpha Lock on
			}
		}
		else											// other key
		{
			col=(CRU[0x14]==0 ? 1 : 0) | (CRU[0x13]==0 ? 2 : 0) | (CRU[0x12]==0 ? 4 : 0);	// get column

			if ((col == 0) || (col == 4))				// reading joystick
			{	
				if (ad == 3)
				{	
					if (key[KEYS[col][0]])				// button reads normally
					{
						ret=0;
					}
				}
				else
				{
					if (!key[KEYS[col][ad-3]])			// but the stick return is inverted
					{	
						ret=0;
					}
				}
			}
			else
			{
				if (key[KEYS[col][ad-3]])				// normal key
				{	
					ret=0;
				}
			}

		}
	}

	// are we checking VDP interrupt?
	if (ad == 0x02)
		ret=0;											// that's the only int we have

	// no other devices at this time

	return(ret);
}

///////////////////////////////////////////////////////////////////////
// Equivalent to CALL FILES(n)
///////////////////////////////////////////////////////////////////////
void do_files(int n) 
{										/* call files(n) */
	Byte *p;
	int adr = 0x4000 - n*518 - 534;		/* first address of disk data (FIXME: ignore previously allocated space) */

	if (v9938)
	{
		adr |= (VDPREG[14] & 0x07) << 14;
	}
	wrword(0x8370, (adr-1) & 0x3fff);	/* last free byte in VDP RAM */
	p = &VDP[adr];
	*p++ = 0xaa;						/* ident */
	*p++ = 0x3f;						/* link (last entry) */
	*p++ = 0xff;
	*p++ = 0x11;						/* CRU page */
	*p++ = n;							/* max. number of files */
	for (; n; n--) 
	{
		*(p+6) = 0;						/* clear first byte of FDS to indicate unused buffer */
		p += 518;						/* next buffer */
	}
}

///////////////////////////////////////////////////////////////////////
// Base entry for standard DSR call
///////////////////////////////////////////////////////////////////////
void do_dsrlnk()
{
	// performs file i/o using the PAB passed as per a normal
	// dsrlnk call. *Only* does LOAD image reads from
	// DSK1, because I don't need more than that for SSA
	// DSK1 is defined as a subdirectory off the current
	// path called "DSK1"
	
	// address of length byte passed in CPU >8356
	
	int PAB;
	char filename[256];
	char temp[1024];

	PAB = romword(0x8356)-14;			// base address of PAB in VDP RAM
	
	if (v9938)
	{
		PAB |= (VDPREG[14] & 0x07) << 14;
	}

	buildfilename(filename, PAB);
	if (0==strlen(filename)) 
	{
		setfileerror(PAB, 0xe0);		// file error
		return;
	}

	switch (VDP[PAB]) 
	{
	case 0x01:		/* close */
		// we'll support CLOSE (0x01) just because it happens in case of Errors
		// but we won't do anything about it
		sprintf(temp,"Closing %s", filename);
		debug_write(temp);
	break;
	
	case 0x05:		/* load */
		load(filename, PAB);
	break;
	
	case 0x06:		/* save */
		save(filename, PAB);			// SAVE, just to be nice
	break;
	
	default:
		// bad opcode. Boy, this DSR sucks ;)
		// set return byte in PAB and exit
		sprintf(temp,"unsupported DSRLNK opcode %d on file %s", VDP[PAB], filename);
		debug_write(temp);
		setfileerror(PAB, 0x040);		// Bad open attribute
	break;
	}
}

///////////////////////////////////////////////////////////////////////
// Load a PROGRAM image file
///////////////////////////////////////////////////////////////////////
void load(char *filename, int PAB) 
{
	int buffer, max_bytes;
	FILE *fp;
	Byte drive;
	char temp[1024];

	drive=VDP[PAB+13] - '0';						// drive number
	buffer = ((VDP[PAB+2]<<8) | VDP[PAB+3])&0x3fff;	// VDP buffer
	if (v9938)
	{
		buffer |= (VDPREG[14] & 0x07) << 14;
	}
	max_bytes=(VDP[PAB+6]<<8) | VDP[PAB+7];			// maximum bytes to read

	if (disktype[drive] == 'i') 
	{
		sprintf(temp,"loading on image disk %d not yet supported", drive);
		debug_write(temp);
		setfileerror(PAB, 0xe0);					// file error
		return;
	}

	fp=fopen(filename, "rb");
	if (NULL == fp)
	{
		sprintf(temp,"cant load 0x%X bytes drive %d file %s", max_bytes, drive, filename);
		debug_write(temp);
		// couldn't open the file
		setfileerror(PAB, 0xe0);					// file error
		return;
	}

	if (disktype[drive] == 'v' || disktype[drive] == 't') /* skip header */
	{
		fseek(fp, 128, SEEK_SET);
	}

	max_bytes = fread(&VDP[buffer], 1, max_bytes, fp);
	sprintf(temp,"loading 0x%X bytes drive %d file %s", max_bytes, drive, filename);
	debug_write(temp);

	setnofileerror(PAB);

	fclose(fp);										// all done
}

///////////////////////////////////////////////////////////////////////////////
// Save a PROGRAM image file
///////////////////////////////////////////////////////////////////////////////
void save(char *filename, int PAB)
{
	int buffer, max_bytes;
	FILE *fp;
	Byte drive;
	char temp[1024];
	
	// Execute a SAVE opcode (Program image)
	
	drive=VDP[PAB+13] - '0';	// drive number
	buffer = ((VDP[PAB+2]<<8) | VDP[PAB+3])&0x3fff;	// VDP buffer
	if (v9938)
	{
		buffer |= (VDPREG[14] & 0x07) << 14;
	}
	max_bytes=(VDP[PAB+6]<<8) | VDP[PAB+7];			// maximum bytes to read

	sprintf(temp,"saving 0x%X bytes file %s", max_bytes, filename);
	debug_write(temp);

	if (disktype[drive] == 'i') 
	{
		debug_write("saving on image disks not supported");
		setfileerror(PAB, 0xe0);					// file error
		return;
	}

	fp=fopen(filename, "wb");
	if (NULL == fp)
	{
		// couldn't open the file
		debug_write("cant open for writing");
		setfileerror(PAB, 0xe0);					// file error
		return;
	}

	if (disktype[drive] == 't') 
	{	/* create TIFILES header */
		Byte h[128];								// header

		memset(h, 0, 128);							// initialize
		h[0] = 7;
		h[1] = 'T';
		h[2] = 'I';
		h[3] = 'F';
		h[4] = 'I';
		h[5] = 'L';
		h[6] = 'E';
		h[7] = 'S';
		h[8] = (max_bytes+255) >> 16;				// length in sectors HB
		h[9] = (max_bytes+255) >> 8;				// LB 
		h[10] = 0x01;								// program 
		h[11] = 0;									// records/sector
		h[12] = max_bytes & 0xff;					// # of bytes in last sector
		h[13] = 0;									// record length 
		h[14] = 0;									// # of records(FIX)/sectors(VAR) LB!
		h[15] = 0;									// HB 
		fwrite(h, 1, 128, fp);
	}
	else 
	{
		if (disktype[drive] == 'v') 
		{	/* create v9t9 header */
			Byte h[128];							// header 
			Byte *p;

			memset(h, 0, 128);						// initialize 
			
			for (p=h; p < (h+VDP[PAB+9]); p++) 
			{
				*p = VDP[PAB+15+(p-h)];
			}
			while (p<h+10)
			{
				*p++ = ' ';							// fill with blanks 
			}
			h[12] = 0x01;							// program 
			h[13] = 0;								// records/sector 
			h[14] = (max_bytes+255) >> 16;			// length in sectors HB 
			h[15] = (max_bytes+255) >> 8;			// LB 
			h[16] = max_bytes & 0xff;				// # of bytes in last sector 
			h[17] = 0;								// record length 
			h[18] = 0;								// # of records(FIX)/sectors(VAR) LB! 
			h[19] = 0;								// HB 
			fwrite(h, 1, 128, fp);
		}
	}

	if (fwrite(VDP+buffer, 1, max_bytes, fp) == (unsigned int)max_bytes)
	{
		setnofileerror(PAB);
	}
	else
	{
		setfileerror(PAB, 0xe0);					// file error
	}

	fclose(fp);										// all done
}

/////////////////////////////////////////////////////////////////////////
// Set the error bit in the PAB
/////////////////////////////////////////////////////////////////////////
void setfileerror(int PAB, Byte errcode) 
{
	VDP[PAB+1] = errcode;		// file error
	CPU[0x807c] |= 0x20;		// are we supposed to set this bit?
}

/////////////////////////////////////////////////////////////////////////
// Set no error in the PAB
/////////////////////////////////////////////////////////////////////////
void setnofileerror(int PAB) 
{
	VDP[PAB+1] = 0x00;		// no errors
	CPU[0x807c] &= 0xdf;		// clear COND bit
}

/////////////////////////////////////////////////////////////////////////
// Assemble a filename to find the file
/////////////////////////////////////////////////////////////////////////
void buildfilename(char *filename, int PAB) 
{
	char *p;
	int idx;
	Byte len, drive;

	len=VDP[PAB+9] - 5;								// Filename length minus 'DSK1.' bit
	drive=VDP[PAB+13] - '0';						// drive number
  
	if (drive<1 || drive>9 || len<1 || 0==strlen(&diskpath[drive][0])) 
	{
		strcpy(filename, "");
		return;
	}

	p = (char *) stpcpy(filename, &diskpath[drive][0]); // points to end
	for (idx=0; idx<=len; idx++) 
	{												// search end, convert name 
		if (idx==len || VDP[PAB+15+idx]==' ') 
		{
			*p = 0;
			break;
		} else 
		{
			if (VDP[PAB+15+idx] == '/')
			{
				*p++ = '_';
			}
		    else
			{
				*p++ = tolower(VDP[PAB+15+idx]);
			}
		}
	}
	
	//  VDP[0x3eeb] = drive;						// last accessed drive
}

/////////////////////////////////////////////////////////////////////////
// Perform low level disk call
/////////////////////////////////////////////////////////////////////////
void do_sbrlnk() 
{
	// performs low level disk i/o using the PAB passed as per a normal
	// dsrlnk call.

	// address of length byte passed in CPU >8356

	int PAB;
	Byte drive;
	int sect;
	int opcode;
	char temp[1024];
	//  (moved so that MSVC doesn't complain) ;)  -  sorry, but not there (what did it complain?)
	int buffer = romword(0x834e)&0x3fff;	// VDP buffer for sector
	Byte n = CPU[0x804c];


	opcode = (PC & 0xff)>>1;
	PAB = romword(0x8356);	// base address of PAB in VDP RAM
	if (v9938)
	{
		PAB |= (VDPREG[14] & 0x07) << 14;
	}
	drive = CPU[0x804c];
	sect = romword(0x8350);

	if (drive<1 || drive>9 || 0==strlen(&diskpath[drive][0])) 
	{
	  sprintf(temp,"SBRLNK call to invalid drive %d", drive);
	  debug_write(temp);
	  CPU[0x8050] = 0x06;	/* hardware error */
	}

	switch (opcode) 
	{
	case 0x10: 				/* sector read/write */
		if (v9938)
		{
			buffer |= (VDPREG[14] & 0x07) << 14;
		}
		sprintf(temp,"%sing drive %d sector %d", CPU[0x804d] ? "read" : "writ", drive, sect);
		debug_write(temp);
		if (CPU[0x804d]==0) 
		{
			CPU[0x8050] = 0x34;	/* write protected */
		} else 
		{
			read_sect(drive, sect, (char*)&VDP[buffer]);
		}
	break;
		
	case 0x16: 				/* call files */
		if (n>0 && n<=16) 
		{
			do_files(n);
			CPU[0x8050] = 0;		/* no error */
		} else	  
		{
			CPU[0x8050] = 0xff;		/* error */
		}
	break;
	
	default:
		sprintf(temp,"unsupported SBRLNK opcode 0x%x", opcode);
		debug_write(temp);
		CPU[0x8050] = 0x06;	/* hardware error */
	break;
	}
	
	//	VDP[0x3eeb] = drive;			/* last accessed drive */
}

/////////////////////////////////////////////////////////////////////////
// Return true if the filename does not start with a dot
/////////////////////////////////////////////////////////////////////////
#ifdef __linux__
int nodot(const struct dirent *d) 
{
	return (*(d->d_name) != '.');
}
#endif 

/////////////////////////////////////////////////////////////////////////
// Read a sector (for directories and the like)
/////////////////////////////////////////////////////////////////////////
void read_sect(Byte drive, int sect, char *buffer) 
{
	FILE *fp;
	char temp[1024];

	switch (disktype[drive]) 
	{
	case 'i':
		fp = fopen(&diskpath[drive][0], "rb");
		if (fp) 
		{
			fseek(fp, sect<<8, SEEK_SET);
			if (fread(buffer, 1, 256, fp)==256)
			{
				CPU[0x8050] = 0;			/* no error */
			}
			else
			{
				CPU[0x8050] = 0x06;			/* hardware error */
			}
			fclose(fp);
		} else 
		{
			sprintf(temp,"cant open %s", &diskpath[drive][0]);
			debug_write(temp);
			memset(buffer, 0, 256);
			CPU[0x8050] = 0x06;				/* hardware error */
		}
	break;

	case 'v':								/* v9t9: simulated directory sector access */
	case 't':								/* TIFILES: simulated directory sector access */
	case 'r':								/* raw: simulated directory sector access */
		CPU[0x8050] = 0;					/* no error */
		memset(buffer, 0, 256);				/* init buffer */
		if (sect==0) 
		{									/* sector 0 */
			char *p;
			if (strlen(&diskpath[drive][0]) < 10) 
			{
				p = (char *) stpcpy(buffer, &diskpath[drive][0]); /* points to end */
				while (p < buffer+10)
				{
					*p++ = ' ';
				}
			} else 
			{
				strcpy(buffer, &diskpath[drive][0] + strlen(&diskpath[drive][0]) - 10);
			}
			p = buffer+0x0a;
			*p++ = 0x05;
			*p++ = (unsigned char)0xa0;
			*p++ = 0x12;
			*p++ = 'D';
			*p++ = 'S';
			*p++ = 'K';
			*p++ = 'P';
			*p++ = 0x28;
			*p++ = 0x02;
			*p++ = 0x02;
			for (p=buffer+0x38; p<buffer+256; p++)
			{
				*p = (unsigned char)0xff;
			}
		} else 
		{			/* other sector */
#ifdef __linux__
			struct dirent **eps;
#endif
			int n;

#ifdef WIN32
			// TODO - get a directory
			n=0;
#endif
#ifdef __linux__
			n = scandir(&diskpath[drive][0], &eps, nodot, alphasort);
#endif

			if (sect==1) 
			{
				int i;
				for (i=0; i<n && i<127; i++)
				{
					buffer[2*i+1] = i+2;
				}
			}
			else 
			{
				if (n>0 && sect>=2 && sect<n+2) 
				{
					char *p, *q;
					int len;
					char tifbuf[128];
	
#ifdef WIN32
					// TODO: malloc space for filename
					p= (char*)malloc(5);
#endif
#ifdef __linux__
					p = (char *) malloc(strlen(&diskpath[drive][0]) + strlen(eps[sect-2]->d_name) + 1);
#endif
					strcpy(p, &diskpath[drive][0]);
#ifdef WIN32
					// TODO: copy a filename into the buffer
#endif
#ifdef __linux__
					strcat(p, eps[sect-2]->d_name);
#endif
					fp=fopen(p, "rb");
					switch (disktype[drive]) 
					{
						case 'v': 		/* v9t9: simulated directory sector access */
							fread(buffer, 1, 0x1b, fp); /* file information without link map */
							/*  debug_write("v9t9 FDS %s", p); */
						break;

						case 'r': 		/* raw: simulated directory sector access */
							sprintf(temp,"raw FDS %s", p);
							debug_write(temp);
#ifdef WIN32
							// TODO - copy the filename and set q to point the end of the string
							q=buffer;
#endif
#ifdef __linux__
							q = (char *) stpncpy(buffer, eps[sect-2]->d_name, 10);
#endif
							while (q < buffer+10)
							{
								*q++ = ' ';
							}
							fseek(fp, 0, SEEK_END);
							len = ftell(fp);
							buffer[0x0c] = 0x09;	/* always PROGRAM */
							buffer[0x0e] = (len+255) >> 16;
							buffer[0x0f] = (len+255) >> 8;
							buffer[0x10] = len & 0xff;
						break;

						case 't': 		/* raw: simulated directory sector access */
							/* debug_write("TIFILES FDS %s", p); */
#ifdef WIN32
							// TODO - copy the filename and set q to point the end of the string
							q=buffer;
#endif
#ifdef __linux__
							q = (char *) stpncpy(buffer, eps[sect-2]->d_name, 10);
#endif
							while (q < buffer+10)
							{
								*q++ = ' ';
							}
							fread(tifbuf, 1, 128, fp); /* TIFILES header */
							q += 2;					/* start at offset 0x0c */
							*q++ = tifbuf[10];
							*q++ = tifbuf[11];
							*q++ = tifbuf[8];
							*q++ = tifbuf[9];
							*q++ = tifbuf[12];
							*q++ = tifbuf[13];
							*q++ = tifbuf[14];
							*q = tifbuf[15];
						break;
					} /* switch */
					fclose(fp);
					free(p);
				} else 
				{
					debug_write("attempt to read not simulated sector");
					CPU[0x8050] = 0x06;	/* hardware error */
				}
#ifdef __linux__
				free(eps);
#endif
			}
		}
	break;

	default:
		debug_write("sector access only implemented for images and v9t9");
		CPU[0x8050] = 0x06;		/* hardware error */
	} /* switch */
}


/////////////////////////////////////////////////////////////////////////
// Read (part of) a file from a disk image
/////////////////////////////////////////////////////////////////////////
#if 0
void read_image_file(int PAB, char *buffer, int offset, int len) 
{
	Byte bufs1[256], buffds[256], fn[10];
	Byte drive;
	Byte *p1, *p2, *p3;
	Word Start, Length;
	int found, fds, i;
	char temp[1024];

	drive=VDP[PAB+13] - '0';	// drive number
	read_sect(drive, 1, bufs1);
	p1 = stpncpy(fn, VDP+PAB+15, VDP[PAB+9]-5);
	while (p1<fn+10)				/* fill with spaces for comparing */
	{
		*p1++ = ' ';
	}
	p1=bufs1, p2=bufs1+256;
	found=0;
	
	do 
	{
		p = p1 + (((p2-p1)/2) & ~1);
		fds = (*p)<<8 | *(p+1);
		if (!fds) 
		{
			p2 = p;
			continue;
		}
		read_sect(drive, fds, buffds);
		if (0 == (cmp=strncmp(fn, buffds, 10))) 
		{
			found=1;
			break;
		} else 
		{
			if (cmp>0)
			{
				p1=p+2;
			}
			else
			{
				p2=p;
			}
		}
	} while (p1<p2);

	if (found) 
	{
		size = 256;
		Length = 0;
		Prog = buffds[12] & 1;
		Dis = (buffds[12] & 2) ? 0 : 1;
		Var = buffds[12] >> 7;
		Len = buffds[0x11];
		sprintf(temp,"Copying %d %s %s %s%s%s%d", o, argv[o], filename, Prog ? "P": " ", !Prog && Dis ? "D": " ", !Prog && Var ? "V": " ", Len);
		debug_write(temp);
		for (p3 = buffds+0x1c; *p3 | *(p3+1); p3 += 3) 
		{		/* all chunks */
			Start = *p3 | ((*(p3+1) & 0x0f) << 8);
			Length = (((*(p3+1) & 0xf0) >> 4) | (*(p3+2) << 4)) - Length;
			for (i=0; i<=Length; i++) 
			{ /* read every sector */
				read_sect(Start++, buffer);
				if (Prog && i==Length && (*(p3+3) | *(p3+4)) == 0)
				{
					size =  buffds[0x10] ?  buffds[0x10] : 256;
				}
				/* last sector of P-File: shorter */
				fwrite(buffer, size, 1, outfile);
			} /* for i */
		} /* for p3 */
	} /* found / for p */
}
#endif

/////////////////////////////////////////////////////////////////////////
// Write a line to the debug buffer displayed on the debug screen
// or to stderr in the Linux build
/////////////////////////////////////////////////////////////////////////
void debug_write(char *s)
{
	memcpy(&lines[0][0], &lines[1][0], 2900);						// scroll data
	strncpy(&lines[29][0], s, 99);									// copy in new line
	memset(&lines[29][strlen(s)], 0x20, 99-strlen(s));				// clear rest of line
	lines[29][99]=0;												// zero terminate
	
#ifdef WIN32
	OutputDebugString(s);
	OutputDebugString("\n");
#endif
#ifdef __linux__
	fprintf(stderr, "%s\n", s);
#endif
}

///////////////////////////////////////////////////////////////////
// Console interrupt simulator
///////////////////////////////////////////////////////////////////
void ConsoleInterrupt()
{

	Byte x1;											// temp var
	int t1,t2;											// temp vars

	interrupt_needed=0;
	x1=CPU[0x80c2];
	if ((x1&0x80)==0)									// if console routine on
	{ 
		if ((x1&0x40)==0)								// update sprites
	    {												// tables are fixed in memory
			for (num_sprites=0; num_sprites<(CPU[0x807a]&0x1f); num_sprites++)		// mask off to 32 sprites max TODO - this is really 31
			{	
				offset=num_sprites<<2;
				y_speed=(Byte)VDP[0x780+offset];
				x_speed=(Byte)VDP[0x781+offset];
				t1=(Byte)VDP[0x782+offset];
				t2=(Byte)VDP[0x783+offset];
				t1a=(y_speed&0xf0)>>4;
				t1b=(y_speed&0xf);
				t2a=(x_speed&0xf0)>>4;
				t2b=(x_speed&0xf);
			
				if (t1a>0x7)
				{	
					t1a-=0xf;
					if (t1b==0)
						t1a-=1;
				}
				if (t2a>0x7)
				{	
					t2a-=0xf;
					if (t2b==0)
						t2a-=1;
				}
	
				t1b=(y_speed&0xf);
				if ((y_speed>0x7f)&&(t1b))
					t1b=(t1b-0x10);
				if ((x_speed>0x7f)&&(t2b))
					t2b=(t2b-0x10);

				sprite_y=(Byte)VDP[0x300+offset];
				sprite_x=(Byte)VDP[0x301+offset];
				
				ty=sprite_y;
				tx=sprite_x;
			
				t1+=t1b;
				t2+=t2b;

				sprite_y+=t1a;
				sprite_x+=t2a;
	
				if (t1>0xf)
				{	
					t1-=0xf;
					sprite_y++;
				}
				if (t1<0)
				{	
					t1+=0xf;
					sprite_y--;
				}

				if (t2>0xf)
				{	
					t2-=0xf;
					sprite_x++;
				}
				if (t2<0)
				{	
					t2+=0xf;
					sprite_x--;
				}

				sprite_x&=0xff;
				sprite_y&=0xff;

				if ((sprite_y>=0xc0)&&(sprite_y<=0xe0))
				{	
					if (y_speed>0x7f)
						sprite_y-=0x20;
					else
						sprite_y+=0x20;
				}
	
				VDP[0x300+offset]=(Byte)sprite_y;
				VDP[0x301+offset]=(Byte)sprite_x;
				VDP[0x782+offset]=(Byte)t1;
				VDP[0x783+offset]=(Byte)t2;
			
				if ((ty!=sprite_y)||(tx!=sprite_x))
					redraw_needed=1;

			}
		}
    
		if ((x1&0x20)==0)								// update sound list
	    { 
	      if (CPU[0x80ce])
		  {		CPU[0x80ce]--;
				if (CPU[0x80ce]==0)
				{	
					snd_address=romword(0x83cc);
					if (CPU[0x80fd] & 0x01) {
					  if (v9938)
						sndlist = &VDP[((VDPREG[14] & 0x07) << 14) |
									  (snd_address&0x3fff)];
					  else
						sndlist=(&VDP[snd_address]);	// VDP sound list
					} else
					{
						sndlist=(&GROM[snd_address]);	// GROM sound list
					}
					data=*sndlist++;
					snd_address+=data;
					if ((data!=0)&&(data!=0xff))
					{	
						while (data--)
						{	
							wsndbyte(*sndlist++);
						}
						data=*sndlist++;
						snd_address+=2;
						wrword(0x83cc,snd_address);
						CPU[0x80ce]=data;
					}	
					else
					{	
						if (data==0xff)
						{	
							if (CPU[0x80fd] & 0x1)
								CPU[0x80fd]&=0xfe;
							else
								CPU[0x80fd]|=0x01;
						}
						CPU[0x80cc]=*sndlist++;
						CPU[0x80cd]=*sndlist++;
						CPU[0x80ce]=0x01;
					}
				}
			}
		}
	
	    if ((x1&0x10)==0)								// Check QUIT key
		{ 
#ifdef __linux__
	      if ((key[KEY_ALT]||key[KEY_ALTGR])&&key[KEY_EQUALS]) {
			while (key[KEY_EQUALS])
			  ; /* wait for key release */
#endif
#ifdef WIN32
			if (((key[VK_MENU])||(key[VK_CONTROL]))&&(key[187]))
			{ 
				while (key['=']);						// wait for key release
#endif
				ST=(ST&0xfff0);							// disable interrupts
				WP=romword(0x0000);						// Reset = BLWP @>0000
				wrword(WP+26,0xffff);
				wrword(WP+28,PC);
				wrword(WP+30,ST);
				PC=romword(0x0002);
				grmaccess=0;		// No GROM Access yet
				vdpaccess=0;		// No VDP address writes yet 
				interrupt_needed=0;	// No interrupt missed yet
			}
	    }
	}

	// the following parts can't be turned off, except by LIMI or disabling VDP ints 
  
  	wrword(0x83d6,romword(0x83d6)+2);	// this blanks the screen when it 
										// wraps around to zero 
	
	if (romword(0x83d6)==0)				// we're wrapped					
	{
		t1=CPU[0x80d4];					// read the byte...					
		t1=(t1|0x20)&0xbf;				// int turns ON VDP interrupts, and	
										// turns OFF display bit			
		VDPREG[1]=(Byte)t1;				// write it							
	}
  
//	CPU[0x8079]+=CPU[0x80fc];			// VDP Int Timer - not sure how it	
	CPU[0x8079]++;						// gets this, but here it is. Add	
										// byte from R14 of the GPL WP at
										// >83e0. Can't trace what it is.	
										// E/A manual calls it System Status

	CPU[0x807b]=VDPS[0];				// VDP Status register - copied to CPU RAM		
										// This is where global sprite COINC goes		
    
	if (romword(0x83C4)) 				// call the user interrupt routine 
	{	// pretend that we really were called with BLWP by saving the values	
		// into the appropriate workspace (0x83c0)
		
		wrword(0x83da,WP);				// WP in R13 
		wrword(0x83dc,PC);				// PC in R14 
		wrword(0x83de,ST);				// ST in R15 

		ST=(ST&0xfff0);					// disable interrupts

		/* now load the correct workspace, and perform a branch and link to the address */
		WP=0x83e0;				/* interrupt workspace is 0x83e0						*/
		wrword(0x83f6,0x0ab8);	/* write into new R11 hard-coded return address in ROM	*/
		PC=romword(0x83c4);		/* off we go.. TI should return normally now			*/
	}
}

// these three functions are all thread routines for WIN32
#ifdef WIN32
//////////////////////////////////////////////////////////////
// 'Retrace' counter for timing - runs at 50 or 60 hz
//////////////////////////////////////////////////////////////
void __cdecl TimerThread()
{
	char buf[1024];

	if (TIMERR_NOERROR != timeBeginPeriod(1))
		fail("Can't start timer");

	time(&STARTTIME);

	myTimer=timeSetEvent(hzRate==50 ? 20:16, 1, (void (__stdcall*)(unsigned int, unsigned int, unsigned long, unsigned long, unsigned long))Counting, NULL, TIME_PERIODIC | TIME_CALLBACK_FUNCTION);	// 16 for 60hz (62.5), 20 for 50hz
	debug_write("Started Timer Thread");
	
	while (quitflag==0)
	{	// now we just wait for the exit flag
		Sleep(100);
	}

	time(&ENDTIME);

	sprintf(buf, "Seconds: %ld, ticks: %ld", (long)ENDTIME-STARTTIME, ticks);
	debug_write(buf);

	debug_write("Ending Timer Thread");
}

////////////////////////////////////////////////////////////////
// Timer calls this function
////////////////////////////////////////////////////////////////
void __stdcall Counting(unsigned int a, unsigned int b, unsigned long c, unsigned long d, unsigned long e)
{
	ticks++;
	retrace_count++;
	interrupt_needed=1;
	VDPS[0] |= 0x80;
}

////////////////////////////////////////////////////////////////
// restart the timer with a (probably) new rate
////////////////////////////////////////////////////////////////
void ResetTimer()
{
	char buf[1024];

	sprintf(buf, "Timer restart to %dhz", hzRate);
	debug_write(buf);
	
	timeKillEvent(myTimer);
	myTimer=timeSetEvent(hzRate==60 ? 16:20, 1, (void (__stdcall*)(unsigned int, unsigned int, unsigned long, unsigned long, unsigned long))Counting, NULL, TIME_PERIODIC | TIME_CALLBACK_FUNCTION);	// 16 for 60hz (62.5), 20 for 50hz
}

////////////////////////////////////////////////////////////////
// stpcpy - strcpy then return the address of the \0
////////////////////////////////////////////////////////////////
char *stpcpy(char *dest, char *src)
{
	char *r;

	strcpy(dest, src);
	r=dest;
	while ('\0' != *r)
		r++;

	return r;
}

////////////////////////////////////////////////////////////////
// stpncpy - strncpy then return the address of the \0
////////////////////////////////////////////////////////////////
char *stpncpy(char *dest, char *src, int n)
{
	char *r;

	strncpy(dest, src, n);
	r=dest;
	while ('\0' != *r)
		r++;

	return r;
}
#endif


// Local Variables: //
// tab-width:4 //
// comment-column:32 //
// fill-column:100 //
// End: //
