#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/fd.h>
#include <linux/fdreg.h>
#include <sys/ioctl.h>
#include <fcntl.h>

unsigned char buffer[256], bufs0[256], bufs1[256], buffds[256];
int akt_track=-1, max_track, max_head, max_sect, max_t_sect;
int density=1, sides=1;
int size=256, fd, drivenr=0;
struct floppy_raw_cmd raw_cmd;

void read_sect(int sect, void *buffer) {
  int track, head, t_sect, tmp;

  if ( sect<0 || sect>max_sect ) {
    perror("sect");
    exit(1);
  }
  track = sect / max_t_sect;
  t_sect = sect % max_t_sect;
  if (track>=max_track) {
    track = 2*max_track - 1 - track;
    head = 1;
  } else {
    head = 0;
  }

  if (track != akt_track) {
    raw_cmd.length= 256;
    raw_cmd.data = buffer;
    raw_cmd.rate = 2;
    raw_cmd.track = 0;
    raw_cmd.cmd_count = 3;
    raw_cmd.cmd[0] = FD_SEEK;
    raw_cmd.cmd[1] = drivenr;
    raw_cmd.cmd[2] = track << 1;
    raw_cmd.flags = FD_RAW_INTR;

    tmp = ioctl( fd, FDRAWCMD, & raw_cmd );
    if ( tmp < 0 ){
      perror("raw cmd: seek");
      exit(1);
    }
    akt_track = track;
  }

  raw_cmd.length= 256;
  raw_cmd.data = buffer;
  raw_cmd.rate = 2;
  raw_cmd.track = 0;
  raw_cmd.cmd_count = 9;
  raw_cmd.cmd[0] = FD_READ;
  raw_cmd.cmd[1] = drivenr + (head << 2);
  raw_cmd.cmd[2] = track;
  raw_cmd.cmd[3] = head;
  raw_cmd.cmd[4] = t_sect;
  raw_cmd.cmd[5] = 1;		/* length 256 */
  raw_cmd.cmd[6] = 18;		/* FIXME: SPS */
  raw_cmd.cmd[7] = 0x2a;	/* gap */
  raw_cmd.cmd[8] = 0x80;	/* sizecode2 */
  raw_cmd.flags = FD_RAW_READ | FD_RAW_INTR;

  tmp = ioctl( fd, FDRAWCMD, & raw_cmd );
  if ( tmp < 0 ){
    perror("raw cmd: read");
    exit(1);
  }
}


void main(int argc, char **argv)
{
  int i, startsect=-1, directory=0;
  char *drive,*s,*end, *copy=NULL;
  int number=1;
  int c;
     
  drive="/dev/fd0";
  if ((s = getenv("drive") ) || (s = getenv("DRIVE")) )
    drive = s;

  opterr = 0;
  while ((c = getopt (argc, argv, "SsDdil:r:n:c:")) != -1)
    switch (c) {
    case 's':
      sides = 0;
      break;
    case 'S':
      sides = 1;
      break;
    case 'd':
      density = 0;
      break;
    case 'D':
      density = 1;
      break;
    case 'l':
      drive = optarg;
      break;
    case 'r':
      startsect = strtoul(optarg, 0, 0);
      break;
    case 'n':
      number = strtoul(optarg, 0, 0);
      break;
    case 'i':			/* directory */
      directory = 1;
      break;
    case 'c':			/* copy */
      copy = optarg;
      break;
    default:
      perror(argv[0]);
      exit(1);
    }

  if (!strcmp(drive, "/dev/fd1")) drivenr=1;
  fd = strtoul(drive, &end, 0);
  if ( *end ){
    fd = open( drive, O_ACCMODE | O_NDELAY);
    if ( fd < 0 ){
      perror("open floppy");
      exit(1);
    }
  }

  max_track = 40;
  max_head = sides + 1;
  max_sect = (1 << (sides+density)) * 360;
  max_t_sect = 9 * (density+1);

  if (startsect>=0) {
    if (startsect>max_sect ) {
      perror("startsect");
      exit(1);
    }

    for (i=startsect; i<startsect+number && i<max_sect; i++) {
      read_sect(i, buffer);

      write(1, buffer, size );
    
#if 0
      fprintf(stderr,"\n");
      fprintf(stderr,"remaining= %lu\n", raw_cmd.length );
      
      for( i=0; i< raw_cmd.reply_count; i++ )
	fprintf(stderr,"%d: %x\n", i, raw_cmd.reply[i] );
#ifdef FD_RAW_DISK_CHANGE
      if( raw_cmd.flags & FD_RAW_DISK_CHANGE)
	fprintf(stderr,"disk change\n");
      else
	fprintf(stderr,"no disk change\n");
#endif
#endif
    }
  } /* startsect>=0 */

  if (directory) {
    unsigned char *p;
    char type[8];
    int fds;
    int NumSect, Free=0;

    read_sect(0, bufs0);
    NumSect = bufs0[0x0a]<<8 | bufs0[0x0b];
    for (p=bufs0+0x38; p<bufs0+256; p++)
      for (i=0; i<8; i++)
	if ((*p & (1 << i)) == 0)
	  Free++;
    bufs0[0x0a] = 0;		/* terminate name */
    printf("%-10s Free: %-4d Used: %-4d %c\n",
	   bufs0,		/* name */
	   Free,
	   NumSect - Free,
	   bufs0[0x10]		/* Protected */
	   );
    read_sect(1, bufs1);
    for (p=bufs1; (fds = (*p)<<8 | *(p+1)) ; p+=2) {
      read_sect(fds, buffds);
      if (buffds[12] & 1) {
	strcpy(type, "Program");
      } else {
	strcpy(type, (buffds[12] & 2) ? "Int/" : "Dis/");
	strcpy(type+4, (buffds[12] & 0x80) ? "Var" : "Fix");
      }
      printf("%-10s %s %-5ld %4d %s\n",
	     buffds,		/* name */
	     type,
	     (buffds[12] & 1) ? /* P: length, otherwise setlength */
	     (long) (buffds[0x0e]<<16 | buffds[0x0f]<<8) - 256 +
	     (buffds[0x10] ?  buffds[0x10] : 256)
	     : buffds[0x11],
	     (buffds[0x0e]<<8 | buffds[0x0f]) + 1,
	     (buffds[12] & 8) ? "P" : " " /* Protected */
	     );
    } /* for p */
  } /* directory */
  
  if (copy) {
    unsigned char *p, *p1, *p2;
    int fds, found=0, cmp;
    int Prog=0, Dis=1, Var=0, Len=0;
    int Start, Length=-1;
    int j;
    char NewLine = '\n';

    read_sect(1, bufs1);
    p1=bufs1, p2=bufs1+256;
    do {
      p = p1 + (((p2-p1)/2) & ~1);
      fds = (*p)<<8 | *(p+1);
      read_sect(fds, buffds);
      for (i=0; i<10; i++)	/* search end */
	if (buffds[i]==' ') {
	  buffds[i] = 0;
	  break;
	}
      if (!fds)
	p2 = p;
      else if (0 == (cmp=strcasecmp(copy, buffds))) {
	found=1;
	break;
      } else if (cmp>0)
	p1=p+2;
      else
	p2=p;
    } while (p1<p2);
    if (found) {
      if (buffds[12] & 1) {
	Prog=1;
      } else {
	if (buffds[12] & 2) Dis=0;
	if (buffds[12] & 0x80) Var=1;
	Len = buffds[0x11];
      }
      for (p = buffds+0x1c;
	   *p | *(p+1);
	   p += 3) {		/* all chunks */
	Start = *p | ((*(p+1) & 0x0f) << 8);
	Length = (((*(p+1) & 0xf0) >> 4) | (*(p+2) << 4)) - Length;
	for (i=0; i<Length; i++) { /* read every sector */
	  read_sect(Start++, buffer);
	  if (Dis && Var && Len==80) { /* Textfiles */
	    p1 = buffer;
	    while ((j = (unsigned int) *p1++) < 255) {
	      if ((p1+j) > (buffer+255)) {	/* no endflag */
		fprintf(stderr,
			"%s: %s is no regular textfile\n", argv[0], copy);
		exit(1);
	      }
	      write(1, p1, j);
	      write(1, &NewLine, 1);
	      p1 += j;
	    }
	  } else {		/* all other files */
	    if (Prog && i==Length-1 && (*(p+3) | *(p+4)) == 0)
	      size =  buffds[0x10] ?  buffds[0x10] : 256;
	    /* last sector of P-File: shorter */
	    write(1, buffer, size );
	  } /* no textfile */
	} /* for i */
      }	/* for p */
    } /* found */
  } /* copy */
  
  close(fd);
  exit(0);
}
