///////////////////////////////////////////////////////
// Ami99 VDP Routines
// M.Brent
// includes the MESS 6x8 font
//
// TODO: extend sprite buffer to hold all sprites
// Make buffers static and change VDP write byte routine
// to update the vdp_drawn flags as appropriate rather than
// redrawing every frame (in case of register or
// color changes much more will have to be regenerated)
// generation of sprite and pattern tables can stay
// in the functions as now but the 'vdp_drawn' array can
// simply be updated
///////////////////////////////////////////////////////

#define FONTHEIGHT 8
#define BORDER 40
#define TIPALETTE 41

// TI Palette 
#ifdef WIN32
//int red[16]={ 1,2,32,96,32,64,160,64,224,224,192,192,32,192,160,224 };
//int green[16]= { 1,2,192,224,32,96,32,192,32,96,192,192,128,64,160,224 };
//int blue[16]= { 1,2,32,96,224,224,32,224,32,96,32,128,32,160,160,224 };
int red[16]={ 0,0,8,24,8,16,40,16,56,56,48,48,8,48,40,56 };
int green[16]= { 0,0,48,56,8,24,8,48,8,24,48,48,32,16,40,56 };
int blue[16]= { 0,0,8,24,56,56,8,56,8,24,8,32,8,40,40,56 };
struct RGB TIPAL[16];
struct RGB vdp_pal[256];
#endif
#ifdef __linux__
BITMAP *VDPBUFFER, *SPRITEBUFFER, *SCREENBUFFER;
RGB TIPAL[16];
PALETTE vdp_pal;
int red[16]={ 0,0,8,24,8,16,40,16,56,56,48,48,8,48,40,56 };
int green[16]= { 0,0,48,56,8,24,8,48,8,24,48,48,32,16,40,56 };
int blue[16]= { 0,0,8,24,56,56,8,56,8,24,8,32,8,40,40,56 };
#endif 

Word SIT;					// Screen Image Table
Word CT;					// Color Table
Word PDT;					// Pattern Descriptor Table
Word SAL;					// Sprite Allocation Table
Word SDT;					// Sprite Descriptor Table
int SITsize, CTsize, PDTsize, SALsize, SDTsize;

int col = 41;				// Color index
int i1,i2;					// temp vars
unsigned int ad;			// temp vars
Byte chr,fgcol,bgcol,x;		// temporary vars for drawing display

Word p_add;					// Address

//////////////////////////////////////////////////////////
// 9900 disassembly routines by John Butler
//////////////////////////////////////////////////////////
#include "9900dism.cpp"

//////////////////////////////////////////////////////////
// Font table from MESS
//////////////////////////////////////////////////////////
#include "fontstuf.cpp"

//////////////////////////////////////////////////////////
// Get table addresses from Registers
//////////////////////////////////////////////////////////
void gettables()
{
	if (v9938) 
	{
		// default table base addresses
		/* Screen Image Table / Pattern Name Table */
		SIT=(VDPREG[2]<<10)&0x1ffff;
		/* Colour Table */
		CT=((VDPREG[3]<<6)|(VDPREG[10]<<14))&0x1ffff;
		/* Pattern Descriptor Table / Pattern Generator Table */
		PDT=(VDPREG[4]<<11)&0x1ffff;
		/* Sprite Attribute List / Table */
		SAL=((VDPREG[5]<<7)|(VDPREG[11]<<15))&0x1ffff;
		/* Sprite Descriptor Table / Sprite Pattern Generator Table */
		SDT=(VDPREG[6]<<11)&0x1ffff;

		// default sizes (G1)
		SITsize = 0x300;
		CTsize = 0x20;
		PDTsize = 0x800;
		SALsize = 0x80;
		SDTsize = 0x400;

		// corrections
		switch (mode) 
		{
		case G2:
		case G3:					/* FIXME: sprite mode 2 */
			CTsize = 0x1800;
			PDTsize = 0x1800;
		break;
		case G4:					/* FIXME: sprite mode 2 */
		case G5:					/* FIXME: sprite mode 2 */
			SIT &= 0x18000;
			SAL &= 0x1fc00;
			PDTsize = 0x6a00;
			SITsize = CTsize = 0;		/* bitmapped */
			SALsize = SDTsize = 0;	/* FIXME: no sprites */
		break;
		case G6:					/* FIXME: sprite mode 2 */
		case G7:					/* FIXME: sprite mode 2 */
			SIT = (VDPREG[2]<<11) & 0x10000;
			SAL &= 0x1fc00;
			PDTsize = 0xd400;
			SITsize = CTsize = 0;		/* bitmapped */
			SALsize = SDTsize = 0;	/* FIXME: no sprites */
		break;
		case T1:
			SITsize = 0x3c0;
			CTsize = 0;
			SALsize = SDTsize = 0;	/* no sprites */
		break;
		case T2:
			SIT &= 0x1f000;
			CT &= 0x1fe00;
			SITsize = 0x870;
			CTsize = 0x10e;
			SALsize = SDTsize = 0;	/* no sprites */
		break;
		default:
		break;
		} /* switch mode */
	} else 
	{						/* !v9938 */
		/* Screen Image Table */
		SIT=(VDPREG[2]<<10)&0x3fff;
		/* Colour Table */
		CT=(VDPREG[3]<<6)&0x3fff;
		/* Pattern Descriptor Table */
		PDT=(VDPREG[4]<<11)&0x3fff;
		/* Sprite Attribute List */
		SAL=(VDPREG[5]<<7)&0x3fff;
		/* Sprite Descriptor Table */
		SDT=(VDPREG[6]<<11)&0x3fff;
	} // !v9938

	// The normal math for table addresses isn't quite right in bitmap mode
	// The PDT and CT have different math
	if (mode==G2 || mode==G3) {
	  PDT &= 0x1e000;
	  CT &= 0x1e000;
	}
}

//////////////////////////////////////////////////////////
// Set VDP palette
//////////////////////////////////////////////////////////
void VDPsetpalette() 
{
  int i1, i2, t, t1, t2;

	vdp_pal[BORDER].r=TIPAL[0].r;
	vdp_pal[BORDER].g=TIPAL[0].g;
	vdp_pal[BORDER].b=TIPAL[0].b;

	switch (mode) 
	{
	case G1:
		for (i1=0; i1<32; i1++) 
		{
			t=VDP[CT+i1];
			t1=t&0xf;
			t2=t>>4;
			i2=i1+128;
			vdp_pal[i1].r=TIPAL[t1].r;
			vdp_pal[i1].g=TIPAL[t1].g;
			vdp_pal[i1].b=TIPAL[t1].b;
			vdp_pal[i2].r=TIPAL[t2].r;
			vdp_pal[i2].g=TIPAL[t2].g;
			vdp_pal[i2].b=TIPAL[t2].b;
		}
	break;
	case G2:
	case G3:
	case G4:
	case G5:
	case G6:
		for (i1=0; i1<16; i1++) 
		{	
			vdp_pal[i1].r=TIPAL[i1].r;
			vdp_pal[i1].g=TIPAL[i1].g;
			vdp_pal[i1].b=TIPAL[i1].b;
		}
	break;
	case G7:
		for (i1=0; i1<256; i1++) 
		{
			vdp_pal[i1].r = (i1&0x1c)<<1;
			vdp_pal[i1].g = (i1&0xe0)>>2;
			vdp_pal[i1].b = (i1&0x03)<<4;
		}
	break;
	case T1: 
		i1=0;
		t=VDPREG[7];
		t1=t&0xf;
		t2=t>>4;
		i2=i1+128;
		vdp_pal[i1].r=TIPAL[t1].r;
		vdp_pal[i1].g=TIPAL[t1].g;
		vdp_pal[i1].b=TIPAL[t1].b;
		vdp_pal[i2].r=TIPAL[t2].r;
		vdp_pal[i2].g=TIPAL[t2].g;
		vdp_pal[i2].b=TIPAL[t2].b;
	break;
	case T2:
		t=VDPREG[7];
		t1=t&0xf;
		t2=t>>4;
		vdp_pal[0].r=TIPAL[t1].r;
		vdp_pal[0].g=TIPAL[t1].g;
		vdp_pal[0].b=TIPAL[t1].b;
		vdp_pal[1].r=TIPAL[t2].r;
		vdp_pal[1].g=TIPAL[t2].g;
		vdp_pal[1].b=TIPAL[t2].b;
		t=VDPREG[12];					/* blink color */
		t1=t&0xf;
		t2=t>>4;
		vdp_pal[2].r=TIPAL[t1].r;
		vdp_pal[2].g=TIPAL[t1].g;
		vdp_pal[2].b=TIPAL[t1].b;
		vdp_pal[3].r=TIPAL[t2].r;
		vdp_pal[3].g=TIPAL[t2].g;
		vdp_pal[3].b=TIPAL[t2].b;
	break;
	default:
	break;
	}	/* switch mode */
 
#ifdef WIN32
	// fix up penbuffer
	HDC myDC;
	int idx;

	if (PenBuffer)
	{
		PenBuffer->GetDC(&myDC);

		for (idx=0; idx<256; idx++)
			SetPixel(myDC, idx, 0, RGB((vdp_pal[idx].r)<<2, (vdp_pal[idx].g)<<2, (vdp_pal[idx].b)<<2));

		PenBuffer->ReleaseDC(myDC);
	}

#endif

#ifdef __linux__
	set_palette(vdp_pal);
#endif

	setpalette_needed = 0;
}

//////////////////////////////////////////////////////////
// Perform drawing of a single frame
// Determines which screen mode to draw
// Bitmap gets priority, then text, then graphics
// Multicolor not currently supported
//////////////////////////////////////////////////////////
void VDPdisplay()
{
	char szTemp[1024];

#ifdef WIN32
	RECT myrect;
	HBRUSH mybrush;
	HDC myDC;
#endif

	rebuild_patterns_needed = 1;

	if (v9938) 
	{
		if (!(VDPREG[1] & 0x40)) 
		{
			// screen is blanked
#ifdef WIN32
			myrect.top=0;
			myrect.left=0;
			myrect.right=320;
			myrect.bottom=200;
		
			mybrush = CreateSolidBrush(RGB(TIPAL[0].r, TIPAL[0].g, TIPAL[0].b));
			BackBuffer->GetDC(&myDC);
			FillRect(myDC, &myrect, mybrush);
			BackBuffer->ReleaseDC(myDC);
			DeleteObject(mybrush);
#endif
#ifdef __linux__
			rectfill(screen, 0, 0, wti + 39, hti + 39, BORDER);
#endif
			return;
		}

		if (mode!=oldmode || debug!=olddebug) 
		{	/* graphics mode or debug state changed */
			
			gettables();

			switch (mode) 
			{
			case G1:
				debug_write("Entering Graphics 1 mode");
			break;		   
			case G2:			   
				debug_write("Entering Graphics 2 mode");
			break;		   
			case G3:			   
				debug_write("Entering Graphics 3 mode");
			break;		   
			case G4:			   
				debug_write("Entering Graphics 4 mode");
			break;		   
			case G5:			   
				debug_write("Entering Graphics 5 mode");
			break;		   
			case G6:			   
				debug_write("Entering Graphics 6 mode");
			break;		   
			case G7:			   
				debug_write("Entering Graphics 7 mode");
			break;		   
			case T1:			   
				debug_write("Entering Text 1 mode");
			break;		   
			case T2:			   
				debug_write("Entering Text 2 mode");
			break;		   
			case MC:			   
				debug_write("Entering Multicolor mode");
			break;		   
			default:			   
				sprintf(szTemp, "Entering unknown 0x%2x mode", mode);
				debug_write(szTemp);
			break;
		} /* switch */

		sprintf(szTemp, "debug o%s\n", debug ? "n" : "ff");
		debug_write(szTemp);

		if (mode==T2 || mode==G5 || mode==G6 || mode==G7) 
		{
			wti = 512;											/* width of TI screen */
			hti = (VDPREG[9]&0x80) ? 424 : 384;					/* interlace mode not yet supported */
			wscr = debug ? 1024 : 640;							/* width of window / hor. resolution */
			hscr = debug ? 768  : 480;
		} else 
		{
			wti = 256;											/* width of TI screen */
			hti = (VDPREG[9]&0x80) ? 212 : 192;					/* interlace mode not yet supported */
			wscr = debug ? 640 : 320;							/* width of window / hor. resolution */
			hscr = debug ? 480 : 200;
		}	

		// change resolution if necessary
		if (wscr != oldwscr) 
		{
#ifdef WIN32
			// TODO: resize window if it's smaller than wscr, hscr?
#endif
#ifdef __linux__
			if (set_gfx_mode(GFX_AUTODETECT, wscr, hscr, 0, 0) < 0)
			{
				sprintf(szTemp, "Can't set %dx%dx8 graphics mode!", wscr, hscr);
				debug_write(szTemp);
			}
			clear_keybuf();	/* maybe didnt get release event */
#endif
		}

		VDPsetpalette();
#ifdef WIN32
		myrect.top=0;
		myrect.left=0;
		myrect.right=wscr;
		myrect.bottom=hscr;
		
		mybrush = CreateSolidBrush(RGB(TIPAL[0].r, TIPAL[0].g, TIPAL[0].b));
		BackBuffer->GetDC(&myDC);
		FillRect(myDC, &myrect, mybrush);
		BackBuffer->ReleaseDC(myDC);
		DeleteObject(mybrush);
#endif
#ifdef __linux__
		rectfill(screen, 0, 0, wscr-1, hscr-1, 0); /* clear screen (border!) */
#endif

		oldmode = mode;
		olddebug = debug;
		oldwscr = wscr;

	} /* if mode!=oldmode... */
	
	if (setpalette_needed)
	{
		VDPsetpalette();
	}

	switch (mode) 
	{
	default:					/* all other modes not yet supported */
								/* use G1 so that at least something is visible... */
	case G1:
		VDPgraphics();
		break;
	case G2:
	case G3:												/* cant handle sprite mode 2 anyway :) */
		VDPbitmap();
		break;
	case G4:
		VDPgraphics4(0, 0, wti-1, hti-1);
		break;
	case G6:
		VDPgraphics6(0, 0, wti-1, hti-1);
		break;
	case G7:
		VDPgraphics7(0, 0, wti-1, hti-1);
		break;
	case T1:
		VDPtext();
		break;
	case T2:
		VDPtext2();
		break;
	}

    redraw_needed=0;
	} else 
	{						/* !v9938 */

		/* determine which video mode to display */
		/* we only support Bitmap, Text, and Graphics, not Multicolor */
		/* right or not, I give Bitmap priority, then text, then graphics */

		int i1;

		gettables();
		VDPsetpalette();
		rebuild_patterns_needed = 1;

		if (!(VDPREG[1] & 0x40))
		{
			// screen is blanked
			for (i1=0; i1<16; i1++)
			{	
				vdp_pal[i1].r=TIPAL[i1].r;
				vdp_pal[i1].g=TIPAL[i1].g;
				vdp_pal[i1].b=TIPAL[i1].b;
			}
			vdp_pal[BORDER].r=TIPAL[0].r;
			vdp_pal[BORDER].g=TIPAL[0].g;
			vdp_pal[BORDER].b=TIPAL[0].b;

#ifdef WIN32
			// fix up penbuffer
			int idx;

			if (PenBuffer)
			{
				PenBuffer->GetDC(&myDC);

				for (idx=0; idx<256; idx++)
					SetPixel(myDC, idx, 0, RGB((vdp_pal[idx].r)<<2, (vdp_pal[idx].g)<<2, (vdp_pal[idx].b)<<2));

				PenBuffer->ReleaseDC(myDC);
			}

			myrect.top=0;
			myrect.left=0;
			myrect.right=320;
			myrect.bottom=200;
		
			mybrush = CreateSolidBrush(RGB(TIPAL[0].r, TIPAL[0].g, TIPAL[0].b));
			BackBuffer->GetDC(&myDC);
			FillRect(myDC, &myrect, mybrush);
			BackBuffer->ReleaseDC(myDC);
			DeleteObject(mybrush);
#endif
#ifdef __linux__
			set_palette(vdp_pal);
			rectfill(screen, 0, 0, 319, 199, 0);
#endif
			return;
		}

		redraw_needed=0;

		if (VDPREG[0] & 0x02)	// MODE BIT 3
		{
			VDPbitmap();
			return;
		}

		if (VDPREG[1] & 0x10)	// MODE BIT 2
		{
			VDPtext();
			return;
		}

		if (VDPREG[1] & 0x80)	// MODE BIT 1
		{
			//fail("Video mode MULTICOLOR not supported");
			// just ignore this mode
		}

		VDPgraphics();
	}

}

////////////////////////////////////////////////////////////
// Startup and run VDP graphics interface
////////////////////////////////////////////////////////////
void __cdecl VDPmain()
{	

#ifdef WIN32
	DDBLTFX ddbltfx;
	RECT fill;
	DWORD ret;
	HDC myDC;
	int idx;

	MaintainAspect=0;

	if (DirectDrawCreate(NULL, &lpdd, NULL) != DD_OK)				// create DD object
	{
		fail("Can't make Direct Draw Object");
	}

	// get the primary surface
	lpdd->SetCooperativeLevel(myWnd, DDSCL_NORMAL);					// set co-op level for windowed mode
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	if (lpdd->CreateSurface(&ddsd, &ScreenBuffer, NULL) != DD_OK)
	{
		fail("Can't get primary surface");
	}

	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 768;
	ddsd.dwHeight = 768;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;

	// get the backbuffer - full 768x768 size for debug info
	if (lpdd->CreateSurface(&ddsd, &BackBuffer, NULL) != DD_OK)
	{
		debug_write("* Backbuffer in RAM - *VERY* Bad!");
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = 768;
		ddsd.dwHeight = 768;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

		if (lpdd->CreateSurface(&ddsd, &BackBuffer, NULL) != DD_OK)
			fail("Can't get back buffer");
	}
	else debug_write("- Backbuffer in Video RAM");

	// clear the backbuffer
	ZeroMemory(&ddbltfx, sizeof(ddbltfx));
	ddbltfx.dwSize = sizeof(ddbltfx);
	ddbltfx.dwFillColor=RGB(0xff, 0xff, 0xff);	// white (I hope)
	fill.top=0;
	fill.left=0;
	fill.right=640;
	fill.bottom=480;
	BackBuffer->Blt(&fill, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &ddbltfx);

	// The PenBuffer has a copy of each color to eliminate the need for GDI locking
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 256;
	ddsd.dwHeight = 1;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;			// video memory

	if (lpdd->CreateSurface(&ddsd, &PenBuffer, NULL) != DD_OK)
	{
		debug_write("* Pen Buffer in RAM - HORRIBLE!!");
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = 256;
		ddsd.dwHeight = 1;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY; // anywhere

		if (lpdd->CreateSurface(&ddsd, &PenBuffer, NULL) != DD_OK)
			fail("Can't get Pen Buffer");
	}
	else debug_write("- Pen Buffer in Video RAM");

	// now we fill in the pen buffer
	PenBuffer->GetDC(&myDC);

	for (idx=0; idx<256; idx++)
		SetPixel(myDC, idx, 0, RGB(vdp_pal[idx].r<<2, vdp_pal[idx].g<<2, vdp_pal[idx].b<<2));

	PenBuffer->ReleaseDC(myDC);

	// VDPBUFFER is a surface used to draw the character set for  
	// blitting to the screen. 
	// Gives a bit of speed boost, & reduces duplication of work.
	// We only draw needed chars 
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 2048;
	ddsd.dwHeight = 32;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;

	if (lpdd->CreateSurface(&ddsd, &VDPBuffer, NULL) != DD_OK)
	{
		debug_write("* VDP Buffer in RAM - BAD!");
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = 2048;
		ddsd.dwHeight = 32;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

		if (lpdd->CreateSurface(&ddsd, &VDPBuffer, NULL) != DD_OK)
			fail("Can't get VDP buffer");
	}
	else debug_write("- VDP Buffer in Video RAM");

	// get sprite drawing buffer
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 16;
	ddsd.dwHeight = 16;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;

	if (lpdd->CreateSurface(&ddsd, &SpriteBuffer, NULL) != DD_OK)
	{
		debug_write("* Sprite Buffer in RAM - not good");
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = 16;
		ddsd.dwHeight = 16;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

		if (lpdd->CreateSurface(&ddsd, &SpriteBuffer, NULL) != DD_OK)
			fail("Can't get Sprite Buffer");
	}
	else debug_write("- Sprite Buffer in Video RAM");

	// The FontBuffer is used to render the debugger font for quick blits
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 1024;
	ddsd.dwHeight = 8;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;			// video memory

	if (lpdd->CreateSurface(&ddsd, &FontBuffer, NULL) != DD_OK)
	{
		debug_write("* Font buffer in RAM - affects debug speed");
		ZeroMemory(&ddsd, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.dwWidth = 1024;
		ddsd.dwHeight = 8;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY; // anywhere

		if (lpdd->CreateSurface(&ddsd, &FontBuffer, NULL) != DD_OK)
			fail("Can't get Font Buffer");
	}
	else debug_write("- Font Buffer in Video RAM");
	
	RenderFont();

	for (i1=0; i1<16; i1++)
	{ 
		TIPAL[i1].r=red[i1];
	    TIPAL[i1].b=blue[i1];
		TIPAL[i1].g=green[i1];
	}

	for (i1=0; i1<256; i1++)
	{ 
		vdp_pal[i1].r=0x80;
		vdp_pal[i1].g=0x80;
		vdp_pal[i1].b=0x80;
	}

	for (i1=0; i1<16; i1++)
	{	
		vdp_pal[i1+TIPALETTE].r=TIPAL[i1].r;
		vdp_pal[i1+TIPALETTE].g=TIPAL[i1].g;
		vdp_pal[i1+TIPALETTE].b=TIPAL[i1].b;
	}

	redraw_needed=1;

	// now we create a waitable object and sit on it - the main thread
	// will tell us when we should redraw the screen.
	DisplayEvent=CreateEvent(NULL, false, false, NULL);
	if (NULL == DisplayEvent)
		debug_write("Video Event Creation failed");
	BlitEvent=CreateEvent(NULL, false, false, NULL);
	if (NULL == BlitEvent)
		debug_write("Blit Event Creation failed");

	debug_write("Starting video loop");

	while (quitflag==0)
	{
		if ((ret=WaitForMultipleObjects(2, Video_hdl, false, 100)) != WAIT_TIMEOUT)
		{
			if (WAIT_FAILED==ret)
				ret=GetLastError();

			if (WAIT_OBJECT_0 == ret)
				VDPdisplay();

			if (debug)
			{
				draw_debug();								// draw the screen if in debug mode
			}

			doBlit();
		}
	}

	CloseHandle(DisplayEvent);
#endif

#ifdef __linux__
	set_color_depth(8);
  
	if (debug)
	{  
		if (set_gfx_mode(GFX_DIRECTX,640,480,0,0)<0)
			fail("Can't set 640x480x8 graphics mode!");
	}
	else
	{
		if (set_gfx_mode(GFX_DIRECTX,320,200,0,0)<0)
			fail("Can't set 320x200x8 graphics mode!");
	}

	set_display_switch_mode(SWITCH_BACKGROUND);
	set_display_switch_callback(SWITCH_IN, restore_video);

	/* VDPBUFFER points to a bitmap sized to be 256 characters by */
	/* 1 character high. It's used to draw the character set for  */
	/* blitting to the screen. We draw 256 chars instead of 768.  */
	/* Gives a bit of speed boost, & reduces duplication of work  */
  
	/* Now drawing only needed chars */

	VDPBUFFER=create_bitmap(2048,32);		// for WIN32 compatibility, changed from 4096x16 to 2048x32
	if (VDPBUFFER==NULL) fail("Can't create VDP workspace.");

	SPRITEBUFFER=create_bitmap(16,16);
	if (SPRITEBUFFER==NULL) fail("Can't create VDP Sprite workspace.");

	SCREENBUFFER=create_bitmap(256,192);
	if (SCREENBUFFER==NULL) fail("Can't create VDP screen buffer.");

	clear_to_color(screen,BORDER);

	for (i1=0; i1<16; i1++)
	{ 
		TIPAL[i1].r=red[i1];
	    TIPAL[i1].b=blue[i1];
		TIPAL[i1].g=green[i1];
	}

	for (i1=0; i1<256; i1++)
	{ 
		vdp_pal[i1].r=0x80;
		vdp_pal[i1].g=0x80;
		vdp_pal[i1].b=0x80;
	}

	for (i1=0; i1<16; i1++)
	{	
		vdp_pal[i1].r=TIPAL[i1].r;
		vdp_pal[i1].g=TIPAL[i1].g;
		vdp_pal[i1].b=TIPAL[i1].b;
	}

	redraw_needed=1;
	/* screen should be set up and open */
  
	VDPREG[8]=0xff;   /* clear hacked-in keyboard buffer */
#endif
}

//////////////////////////////////////////////////////
// Draw a debug screen around the main one
//////////////////////////////////////////////////////
#ifdef __linux__
#define textout(s,x,y) textout(screen, font, s, x, y, 1);
#endif

void draw_debug()
{
	char buf1[80], buf2[80], temp[1024];
	int myPC, idx;
	int tmpPC;
	int i, xoff, yoff;

	xoff=(wti<320 ? wti+64 : wti+16);
	yoff=hti+48;

	// eventually track a starting point and put the current PC near the middle of the screen
	myPC=PC;

	for (idx=0; idx<(hti<256?30:50); idx++)
	{
		tmpPC = myPC;
		myPC += Dasm9900(buf2, myPC);
		sprintf(buf1,"%04x   %-32s",tmpPC,buf2);
		textout(buf1, xoff, idx*FONTHEIGHT);
	}

	// draw the current register settings
	idx=0;

	sprintf(buf1,"PC     - %04x",PC);
	textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);

	sprintf(buf1,"WP     - %04x",WP);
	textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);

	sprintf(buf1,"ST     - %04x",ST);
	textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);

	if (v9938)
	{
		strcpy(buf1, "VDPST  - ");
		for (i=0; i<5; i++)
		{
			sprintf(temp, "%02x ", VDPS[i]);
			strcat(buf1, temp);
		}
		textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);
		strcpy(buf1, "       - ");
		for (i=0; i<5; i++)
		{
			sprintf(temp, "%02x ", VDPS[i+5]);
			strcat(buf1, temp);
		}
		textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);
	}
	else
	{
		sprintf(buf1, "VDPST  - %02x", VDPS[0]);
		textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);
	}
	
	sprintf(buf1,"VDPADD - %04x",VDPADD);
	textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);

	sprintf(buf1,"GRMADD - %04x",GRMADD);
	textout(buf1, 0, yoff+(idx++)*FONTHEIGHT);

	// Registers and VDP registers
	for (idx=0; idx<16; idx++)
	{
		if (v9938) 
		{
			if (idx<15)
				sprintf(buf1,"R%-2d=%04x - VDPREG %-2d=%02x, %-2d=%02x, %-2d=%02x",
				  idx, romword(WP+idx*2),
				  idx, VDPREG[idx],
				  idx+16, VDPREG[idx+16],
				  idx+32, VDPREG[idx+32]
				  );
			else
				sprintf(buf1,"R%-2d=%04x - VDPREG %-2d=%02x, %-2d=%02x",
				  idx, romword(WP+idx*2),
				  idx, VDPREG[idx],
				  idx+16, VDPREG[idx+16]
				  );
		} 
		else 
		{												/* !v9938 */
			if (idx<8)
				sprintf(buf1,"R%-2d    - %04x     VDPREG %1d    - %02x",idx, romword(WP+idx*2), idx, VDPREG[idx]);
			else
				sprintf(buf1,"R%-2d    - %04x",idx, romword(WP+idx*2));
		}

		textout(buf1, 0, 64+yoff+idx*FONTHEIGHT);
	}

	// debug text
	for (idx=0; idx<30; idx++)
	{
		textout(&lines[idx][0], xoff-(wti>320?160:0), (idx+(hti<256?30:55))*FONTHEIGHT);
	}

	i=0;
}
#undef textout

#ifdef __linux__
/////////////////////////////////////////////////////////
// Called when returning to this screen
/////////////////////////////////////////////////////////
void restore_video()
{
	// DOESN'T WORK?
}
#endif

/////////////////////////////////////////////////////////
// Draw graphics mode
/////////////////////////////////////////////////////////
void VDPgraphics()
{
	int xx,t,o;						// temp variables
	int i1,i2,i3;					// temp variables
	int vdp_drawn[256];				// flags for which characters have been drawn
#ifdef WIN32
	RECT myrect;
#endif
	int tcol;
	int c1, c2;

	gettables();

	// clear the vdp_drawn[] array
	if (rebuild_patterns_needed) 
	{
		memset(vdp_drawn,0,sizeof(vdp_drawn));
		rebuild_patterns_needed = 0;
	}

	// now blit each tile to the screen buffer
	// calculate position in VDPBUFFER of base of table

	o=0;						// offset in SIT

	for (i1=0; i1<192; i1+=8)									// y loop
	{ 
		for (i2=0; i2<256; i2+=8)								// x loop
		{ 
			if (vdp_drawn[VDP[SIT+o]]==0)
			{	
				// Now we draw a single character into VDPBUFFER. Then we set a flag		
				// that is has been done, so that we don't do it twice this frame. We only	
				// need to draw the characters we actually use this way.					

				xx=VDP[SIT+o]<<3;
				p_add=PDT+xx;
		
				for (i3=0; i3<8; i3++)
				{	
					Byte c = VDP[SIT+o]>>3;
					t=VDP[p_add++];
		
					pixel(VDPBuffer,xx,i3,c|(t&0x80 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+1,i3,c|(t&0x40 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+2,i3,c|(t&0x20 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+3,i3,c|(t&0x10 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+4,i3,c|(t&0x08 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+5,i3,c|(t&0x04 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+6,i3,c|(t&0x02 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+7,i3,c|(t&0x01 ? 0x80 : 0 ));
				}
				vdp_drawn[VDP[SIT+o]]=1;
			}
#ifdef WIN32
			myrect.top=0;
			myrect.bottom = 8;
			myrect.left=VDP[SIT+o++]<<3;
			myrect.right=myrect.left + 8;
			BackBuffer->BltFast(i2+32, i1+4, VDPBuffer, &myrect, DDBLTFAST_NOCOLORKEY);
#endif
#ifdef __linux__
			blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++]<<3),0,i2+32,i1+4,8,8);
#endif
		}
	}

	DrawSprites();
}

////////////////////////////////////////////////////////////////////////
// Draw graphics mode 4
////////////////////////////////////////////////////////////////////////
void VDPgraphics4(int x1, int y1, int x2, int y2) 
{ /* this function to display the graphics once */
	int i1,i2;
	Byte *p1;

	for (i1=y1+4; i1<=y2+4; i1++) 
	{	/* y loop */
		p1 = &VDP[SIT & 0x18000] + ((i1-4)<<7) + (x1>>1);
		for (i2=x1+32; i2<=x2+32; i2+=2) 
		{	/* x loop */
			pixel(BackBuffer, i2, i1, *p1 >> 4);
			pixel(BackBuffer, i2+1, i1, *p1++ & 0x0f);
		}
	}
}

////////////////////////////////////////////////////////////////////////
// Draw graphics mode 6
////////////////////////////////////////////////////////////////////////
void VDPgraphics6(int x1, int y1, int x2, int y2) 
{ /* this function to display the graphics once */
	int i1,i2;
	Byte *p1, *p2;

	for (i1=y1+4; i1<=y2+4; i1+=2) 
	{	/* y loop */
		p1 = &VDP[SIT&0x10000] + ((i1-4)<<7) + (x1>>1);
		if (VDPREG[9] & 0x0c == 0x0c) /* interlace */
			p2 = p1 - 0x10000;		/* &VDP[0]; FIXME: handle p1=&VDP[0] */
		else
			p2 = p1;

		for (i2=x1+32; i2<=x2+32; i2+=2) 
		{	/* x loop */
			pixel(BackBuffer, i2, i1, *p1 >> 4);
			pixel(BackBuffer, i2+1, i1, *p1++ & 0x0f);
			pixel(BackBuffer, i2, i1+1, *p2 >> 4);
			pixel(BackBuffer, i2+1, i1+1, *p2++ & 0x0f);
		}
	}
}

////////////////////////////////////////////////////////////////////////
// Draw graphics mode 7
////////////////////////////////////////////////////////////////////////
void VDPgraphics7(int x1, int y1, int x2, int y2) 
{ /* this function to display the graphics once */
	int i1,i2;
	Byte *p1, *p2;

	for (i1=y1+4; i1<=y2+4; i1+=2) 
	{	/* y loop */
		p1 = &VDP[SIT&0x10000] + ((i1-4)<<7) + (x1>>1);
		if (VDPREG[9] & 0x0c == 0x0c) /* interlace */
			p2 = p1 - 0x10000;		/* &VDP[0]; FIXME: handle p1=&VDP[0] */
		else
			p2 = p1;

		for (i2=x1+32; i2<=x2+32; i2+=2) 
		{	/* x loop */
			pixel(BackBuffer, i2, i1, *p1);
			pixel(BackBuffer, i2+1, i1, *p1++);
			pixel(BackBuffer, i2, i1+1, *p2);
			pixel(BackBuffer, i2+1, i1+1, *p2++);
		}
	}
}

////////////////////////////////////////////////////////////////////////
// Draw text mode 40x24
////////////////////////////////////////////////////////////////////////
void VDPtext()
{ 
	int xx,t,o;
	int i1,i2,i3;
	int vdp_drawn[256];		// flags for which characters have been drawn
	int c1, c2;
#ifdef WIN32
	DDBLTFX blit;
	RECT myrect;
#endif

	gettables();

	t=VDPREG[7];
	c1=t&0xf;
	c2=t>>4;

	// clear the vdp_drawn[] array
	if (rebuild_patterns_needed) 
	{
		memset(vdp_drawn, 0, sizeof(vdp_drawn));
		rebuild_patterns_needed = 0;
	}

	// now blit each tile to the screen buffer
	// calculate position in VDPBUFFER of base of table

	// erase border area
#ifdef WIN32
	ZeroMemory(&blit, sizeof(blit));
	blit.dwSize = sizeof(blit);
	blit.dwFillColor=RGB(vdp_pal[c2].r<<2, vdp_pal[c2].g<<2, vdp_pal[c2].b<<2);	
	myrect.top=0;
	myrect.left=0;
	myrect.right=39;
	myrect.bottom=199;
	BackBuffer->Blt(&myrect, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &blit);
	
	ZeroMemory(&blit, sizeof(blit));
	blit.dwSize = sizeof(blit);
	blit.dwFillColor=RGB(vdp_pal[c2].r<<2, vdp_pal[c2].g<<2, vdp_pal[c2].b<<2);
	myrect.top=0;
	myrect.left=280;
	myrect.right=319;
	myrect.bottom=199;
	BackBuffer->Blt(&myrect, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &blit);
#endif
#ifdef __linux__
	rectfill(screen, 0, 0, 39, 199, 0);
	rectfill(screen, 280, 0, 319, 199, 0);
	rectfill(screen, 0, 196, 319, 199, 0);
#endif

	o=0;						// offset in SIT

	for (i1=0; i1<192; i1+=8)					// y loop
	{ 
		for (i2=0; i2<240; i2+=6)				// x loop
		{ 
			if (vdp_drawn[VDP[SIT+o]]==0)
			{	
				// Now we draw a single character into VDPBUFFER. Then we set a flag		
				// that is has been done, so that we don't do it twice this frame. We only	
				// need to draw the characters we actually use this way.					
				xx=VDP[SIT+o]<<3;
				p_add=PDT+xx;
		
				for (i3=0; i3<8; i3++)		// 6 pixels wide
				{	
					t=VDP[p_add++];
					pixel(VDPBuffer,xx,i3,(t&0x80 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+1,i3,(t&0x40 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+2,i3,(t&0x20 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+3,i3,(t&0x10 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+4,i3,(t&0x08 ? 0x80 : 0 ));
					pixel(VDPBuffer,xx+5,i3,(t&0x04 ? 0x80 : 0 ));
				}
				vdp_drawn[VDP[SIT+o]]=1;
			}
#ifdef WIN32			
			myrect.top=0;
			myrect.bottom = 8;
			myrect.left=VDP[SIT+o++]<<3;
			myrect.right=myrect.left + 6;
			BackBuffer->BltFast(i2+40, i1+4, VDPBuffer, &myrect, DDBLTFAST_NOCOLORKEY);
#endif
#ifdef __linux__
			blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++])<<3,0,i2+40,i1+4,6,8);
#endif
		}
	}
	// no sprites in text mode
}

/////////////////////////////////////////////////////
// Draw Text mode 2
/////////////////////////////////////////////////////
void VDPtext2()
{ /* this function to display the text once */
	int xx,t, o=0;
	int i1,i2,i3,i5,t1;
	int vdp_drawn[256];		/* flags for which characters have been drawn */
#ifdef WIN32
	DDBLTFX blit;
	RECT myrect;
#endif

	if (rebuild_patterns_needed) 
	{
		// initialize VDPBUFFER
#ifdef WIN32
		ZeroMemory(&blit, sizeof(blit));
		blit.dwSize = sizeof(blit);
		blit.dwFillColor=RGB(vdp_pal[0].r<<2, vdp_pal[0].g<<2, vdp_pal[0].b<<2);	
		myrect.top=0;
		myrect.left=0;
		myrect.right=2047;
		myrect.bottom=16;
		VDPBuffer->Blt(&myrect, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &blit);
		ZeroMemory(&blit, sizeof(blit));
		blit.dwSize = sizeof(blit);
		blit.dwFillColor=RGB(vdp_pal[2].r<<2, vdp_pal[2].g<<2, vdp_pal[2].b<<2);	
		myrect.top=16;
		myrect.left=0;
		myrect.right=2048;
		myrect.bottom=32;
		VDPBuffer->Blt(&myrect, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &blit);
#endif
#ifdef __linux__
		rectfill(VDPBUFFER, 0, 0, 2047, 15, 0);
		rectfill(VDPBUFFER, 0, 16, 2047, 31, 2);
#endif
		memset(vdp_drawn, 0, sizeof(vdp_drawn));
		rebuild_patterns_needed = 0;
	}

	/* now blit each tile to the screen buffer */
	/* calculate position in VDPBUFFER of base of table */

	for (i1=0; i1< (VDPREG[9]&0x80 ? 212 : 192); i1+=8)         /* y loop */
	{ 
		for (i2=0; i2<480; i2+=6)       /* x loop */
		{ 
			if (vdp_drawn[VDP[SIT+o]]==0)
			{	/* Now we draw a single character into VDPBUFFER. Then we set a flag		*/
				/* that is has been done, so that we don't do it twice this frame. We only	*/
				/* need to draw the characters we actually use this way.					*/

				xx=VDP[SIT+o]<<3;
				p_add=PDT+xx;
		
				for (i3=0; i3<8; i3++)		// 6 pixels wide
				{	
					t=VDP[p_add++];
					t1 = i3<<1;

					if (t&0x80) 
					{
						pixel(VDPBuffer,xx,t1,1);
						pixel(VDPBuffer,xx,t1+1,1);
					}
					if (t&0x40) 
					{
						pixel(VDPBuffer,xx+1,t1,1);
						pixel(VDPBuffer,xx+1,t1+1,1);
					}
					if (t&0x20) 
					{
						pixel(VDPBuffer,xx+2,t1,1);
						pixel(VDPBuffer,xx+2,t1+1,1);
					}
					if (t&0x10) 
					{
						pixel(VDPBuffer,xx+3,t1,1);
						pixel(VDPBuffer,xx+3,t1+1,1);
					}
					if (t&0x08) 
					{
						pixel(VDPBuffer,xx+4,t1,1);
						pixel(VDPBuffer,xx+4,t1+1,1);
					}
					if (t&0x04) 
					{
						pixel(VDPBuffer,xx+5,t1,1);
						pixel(VDPBuffer,xx+5,t1+1,1);
					}
					/* same character with blink bit on */
					if (t&0x80) 
					{
						pixel(VDPBuffer,xx,t1+16,3);
						pixel(VDPBuffer,xx,t1+17,3);
					}
					if (t&0x40) 
					{
						pixel(VDPBuffer,xx+1,t1+16,3);
						pixel(VDPBuffer,xx+1,t1+17,3);
					}
					if (t&0x20) 
					{
						pixel(VDPBuffer,xx+2,t1+16,3);
						pixel(VDPBuffer,xx+2,t1+17,3);
					}
					if (t&0x10) 
					{
						pixel(VDPBuffer,xx+3,t1+16,3);
						pixel(VDPBuffer,xx+3,t1+17,3);
					}
					if (t&0x08) 
					{
						pixel(VDPBuffer,xx+4,t1+16,3);
						pixel(VDPBuffer,xx+4,t1+17,3);
					}
					if (t&0x04) 
					{
						pixel(VDPBuffer,xx+5,t1+16,3);
						pixel(VDPBuffer,xx+5,t1+17,3);
					}
				} /* for i3 */
		
				vdp_drawn[VDP[SIT+o]]=1;
			} /* if !drawn */

			i3 = VDP[CT + (o>>3)];				/* significant byte in color table */
			i3 &= 1 << (7 - (o&0x07));			/* mask significant bit */
			i5=0;
			if (i3)
			{
				i5=16;							/* buffer offset if CT bit set */
			}
#ifdef WIN32
			myrect.top=i5;
			myrect.bottom = i5+16;
			myrect.left=(VDP[SIT+o++])<<3;
			myrect.right=myrect.left+6;
			BackBuffer->BltFast(i2+40,2*i1+4, VDPBuffer, &myrect, DDBLTFAST_NOCOLORKEY);
#endif
#ifdef __linux__
			blit(VDPBUFFER,screen/*SCREENBUFFER*/,(VDP[SIT+o++])<<3,i5,i2+40,2*i1+4,6,16);
#endif
		}	/* for i2 */
	} /* for i1 */

// no sprites in text mode
}

/////////////////////////////////////////////////////
// Draw Bitmap Mode
/////////////////////////////////////////////////////
void VDPbitmap()
{
	int xx,t,bank,o;
	int i1,i2,i3;
	Word c_add;
	int c1, c2;
#ifdef WIN32
	RECT myrect;
#endif
	int vdp_drawn[256];		// flags for which characters have been drawn

//  	PDT=((VDPREG[4]&0x04) << 11)&0x3fff;	// pattern descriptor table
//  	CT=((VDPREG[3]&0x80) << 6)&0x3fff;		// color table

//  	SIT=(VDPREG[2]<<10)&0x3fff;		// Screen Image Table
//  	SAL=(VDPREG[5]<<7)&0x3fff;		// Sprite Attribute List
//  	SDT=(VDPREG[6]<<11)&0x3fff;		// Sprite Descriptor Table 

	o=0;						// offset in SIT

	for (bank=0; bank<3; bank++)				// 3 banks in bitmap mode
	{
		// clear the vdp_drawn[] array
		memset(vdp_drawn, 0, sizeof(vdp_drawn));

		// now blit each tile to the screen buffer
		// calculate position in VDPBUFFER of base of table
		for (i1=0; i1<64; i1+=8)				// y loop
		{ 
			for (i2=0; i2<256; i2+=8)			// x loop
			{ 
				if (vdp_drawn[VDP[SIT+o]] == 0)
				{
					xx=VDP[SIT+o]<<3;
					p_add=(PDT+xx+(bank<<11))&0x3fff;
					c_add=(CT+xx+(bank<<11))&0x3fff;
					
					for (i3=0; i3<8; i3++)
					{	
						t=VDP[p_add++];
						c1=(VDP[c_add])>>4;
						c2=(VDP[c_add++]) & 0xf;

						pixel(VDPBuffer, xx, i3, t&0x80 ? c1 : c2);
						pixel(VDPBuffer, xx+1, i3, t&0x40 ? c1 : c2);
						pixel(VDPBuffer, xx+2, i3, t&0x20 ? c1 : c2);
						pixel(VDPBuffer, xx+3, i3, t&0x10 ? c1 : c2);
						pixel(VDPBuffer, xx+4, i3, t&0x08 ? c1 : c2);
						pixel(VDPBuffer, xx+5, i3, t&0x04 ? c1 : c2);
						pixel(VDPBuffer, xx+6, i3, t&0x02 ? c1 : c2);
						pixel(VDPBuffer, xx+7, i3, t&0x01 ? c1 : c2);
					}
					vdp_drawn[VDP[SIT+o]]=1;
				}
				
#ifdef WIN32
				myrect.top=0;
				myrect.bottom = 8;
				myrect.left=VDP[SIT+o++]<<3;
				myrect.right=myrect.left + 8;
				BackBuffer->BltFast(i2+32, i1+4+(bank<<6), VDPBuffer, &myrect, DDBLTFAST_NOCOLORKEY);
#endif
#ifdef __linux__
				blit(VDPBUFFER,screen/*SCREENBUFFER*/,VDP[SIT+o++]<<3,0,i2+32,i1+4+(bank<<6),8,8);
#endif
			}
		}
	}

	DrawSprites();
}

////////////////////////////////////////////////////////////////
// Stretch-blit the backbuffer into the active window
////////////////////////////////////////////////////////////////
#ifdef WIN32
void doBlit()
{
	RECT myrect, myrect2;
	DDBLTFX blit;

	if (NULL==ScreenBuffer)
		return;				// video isn't ready

	if ((!DrawInactive) && (myWnd != GetForegroundWindow()))
		return;

	GetWindowRect(myWnd, &myrect);
	GetClientRect(myWnd, &myrect2);
	i1=((myrect.right - myrect.left)-(myrect2.right - myrect2.left))/2;
	myrect.left += i1;
	myrect.right = myrect.left + myrect2.right;
	myrect.bottom -= i1;
	myrect.top = myrect.bottom - myrect2.bottom;

	ZeroMemory(&blit, sizeof(blit));
	blit.dwSize = sizeof(blit);

	if (debug)
	{	
		myrect2.top=0;
		myrect2.left=0;
		myrect2.right=(wti>320 ? 768:640);
		myrect2.bottom=(hti>256 ? 768:480);
	}
	else
	{
		myrect2.top=4;
		myrect2.left=32;
		myrect2.right=32+wti;
		myrect2.bottom=4+hti;
	}
	
	ScreenBuffer->Blt(&myrect, BackBuffer, &myrect2, DDBLT_ASYNC, &blit);
}
#endif

/////////////////////////////////////////////////////////////////////////
// Write a string of text to the video card
/////////////////////////////////////////////////////////////////////////
#ifdef WIN32
void textout(char *s, int x, int y)
{
	DDBLTFX blit;
	RECT rect1, rect2;

	ZeroMemory(&blit, sizeof(blit));
	blit.dwSize = sizeof(blit);

	while (*s)
	{
		while ((*s)&&(*s<32)) s++;

		rect1.top=y;
		rect1.left=x;
		rect1.bottom=y+8;
		rect1.right=x+8;

		rect2.left=(*s)*8;
		rect2.top=0;
		rect2.right=rect2.left + 8;
		rect2.bottom=8;

		BackBuffer->BltFast(x, y, FontBuffer, &rect2, DDBLTFAST_NOCOLORKEY);

		s++;
		x+=8;
	}

}
#endif

//////////////////////////////////////////////////////////
// Draw Sprites into the backbuffer
//////////////////////////////////////////////////////////
void DrawSprites()
{
	int i1, i3, xx, yy, pat, col, p_add, t;
	int highest;
	int curSAL;
#ifdef WIN32
	DDCOLORKEY key;
	COLORREF TRANS;
	DDBLTFX blit;
	RECT myrect;

	TRANS=RGB(255,255,255);
#endif

	highest=31;

	// find the highest active sprite
	for (i1=0; i1<32; i1++)			// 32 sprites 
	{
		yy=VDP[SAL+(i1<<2)];
		if (yy==0xd0)
		{
			highest=i1-1;
			break;
		}
	}

	for (i1=highest; i1>=0; i1--)	
	{	
		curSAL=SAL+(i1<<2);
		yy=VDP[curSAL++]+1;				// sprite Y, it's stupid, cause 255 is line 0 
		if (yy>255) yy=0;
		xx=VDP[curSAL++];				// sprite X 
		pat=VDP[curSAL++];				// sprite pattern
		if (VDPREG[1] & 0x2)
			pat=pat&0xfc;				// if double-sized, it must be a multiple of 4
		col=VDP[curSAL]&0xf;			// sprite color 
	
		if (VDP[curSAL++]&0x10)			// early clock
			xx-=32;

		// cannot happen, otherwise would lead to an endless loop...
//  		if (yy==0xd1)
//  			i1=32;						// 0xd0 ends sprite list, add 1 cause I did above

		// clear the spritebuffer
		// We only need to clear it once 
#ifdef WIN32
		ZeroMemory(&blit, sizeof(blit));
		blit.dwSize = sizeof(blit);
		blit.dwFillColor=TRANS;
		myrect.top=0;
		myrect.left=0;
		myrect.right=16;
		myrect.bottom=16;
		SpriteBuffer->Blt(&myrect, NULL, NULL, DDBLT_COLORFILL | DDBLT_ASYNC, &blit);
#endif
#ifdef __linux__
		clear(SPRITEBUFFER);
#endif

		if ((yy<192)&&(col))			// if on-screen and not transparent
		{	
			p_add=SDT+(pat<<3);
			col+=TIPALETTE;
			
			for (i3=0; i3<8; i3++)
			{	
				t=VDP[p_add++];

				if (t&0x80)
					pixel(SpriteBuffer, 0, i3, col);
				if (t&0x40)
					pixel(SpriteBuffer, 1, i3, col);
				if (t&0x20)
					pixel(SpriteBuffer, 2, i3, col);
				if (t&0x10)
					pixel(SpriteBuffer, 3, i3, col);
				if (t&0x08)
					pixel(SpriteBuffer, 4, i3, col);
				if (t&0x04)
					pixel(SpriteBuffer, 5, i3, col);
				if (t&0x02)
					pixel(SpriteBuffer, 6, i3, col);
				if (t&0x01)
					pixel(SpriteBuffer, 7, i3, col);

				if (VDPREG[1]&0x02)		// double-size sprites, need to draw 3 more chars 
				{	
					t=VDP[p_add+7];
					if (t&0x80)
						pixel(SpriteBuffer, 0, i3+8, col);
					if (t&0x40)
						pixel(SpriteBuffer, 1, i3+8, col);
					if (t&0x20)
						pixel(SpriteBuffer, 2, i3+8, col);
					if (t&0x10)
						pixel(SpriteBuffer, 3, i3+8, col);
					if (t&0x08)
						pixel(SpriteBuffer, 4, i3+8, col);
					if (t&0x04)
						pixel(SpriteBuffer, 5, i3+8, col);
					if (t&0x02)
						pixel(SpriteBuffer, 6, i3+8, col);
					if (t&0x01)
						pixel(SpriteBuffer, 7, i3+8, col);

					t=VDP[p_add+15];
					if (t&0x80)
						pixel(SpriteBuffer, 8, i3, col);
					if (t&0x40)
						pixel(SpriteBuffer, 9, i3, col);
					if (t&0x20)
						pixel(SpriteBuffer, 10, i3, col);
					if (t&0x10)	
						pixel(SpriteBuffer, 11, i3, col);
					if (t&0x08)
						pixel(SpriteBuffer, 12, i3, col);
					if (t&0x04)
						pixel(SpriteBuffer, 13, i3, col);
					if (t&0x02)
						pixel(SpriteBuffer, 14, i3, col);
					if (t&0x01)
						pixel(SpriteBuffer, 15, i3, col);

					t=VDP[p_add+23];
					if (t&0x80)
						pixel(SpriteBuffer, 8, i3+8, col);
					if (t&0x40)
						pixel(SpriteBuffer, 9, i3+8, col);
					if (t&0x20)
						pixel(SpriteBuffer, 10, i3+8, col);
					if (t&0x10)
						pixel(SpriteBuffer, 11, i3+8, col);
					if (t&0x08)
						pixel(SpriteBuffer, 12, i3+8, col);
					if (t&0x04)
						pixel(SpriteBuffer, 13, i3+8, col);
					if (t&0x02)
						pixel(SpriteBuffer, 14, i3+8, col);
					if (t&0x01)
						pixel(SpriteBuffer, 15, i3+8, col);
				}
			}

#ifdef WIN32		
			if (VDPREG[1]&0x01)			// magnified sprites
			{	
				myrect.top=yy+4;
				myrect.left=xx+32;
				myrect.right=myrect.left+32;
				myrect.bottom=myrect.top+32;
			}
			else
			{	
				myrect.top=yy+4;
				myrect.left=xx+32;
				myrect.right=myrect.left+16;
				myrect.bottom=myrect.top+16;
			}

			ZeroMemory(&blit, sizeof(blit));
			blit.dwSize = sizeof(blit);

			key.dwColorSpaceLowValue=TRANS;
			key.dwColorSpaceHighValue=TRANS;

			i3=SpriteBuffer->SetColorKey(DDCKEY_SRCBLT, &key);
				
			BackBuffer->Blt(&myrect, SpriteBuffer, NULL, DDBLT_ASYNC | DDBLT_KEYSRC, &blit);
#endif
#ifdef __linux__
			if (VDPREG[1]&0x01)
			{	/* magnified sprites */
				stretch_sprite(screen/*SCREENBUFFER*/, SPRITEBUFFER, xx+32, yy+4, 32, 32);
			}
			else
			{	draw_sprite(screen/*SCREENBUFFER*/,SPRITEBUFFER,xx+32,yy+4);
			}
#endif
		}
	}
}

////////////////////////////////////////////////////////////
// Copy of Pixel from the PenBuffer to the passed surface
// This, it turns out, is way faster than using GDI calls
////////////////////////////////////////////////////////////
#ifdef WIN32
void pixel(LPDIRECTDRAWSURFACE src, int x, int y, int c)
{
	RECT rect1, rect2;
	DDBLTFX blit;

	rect1.left=c;
	rect1.right=c+1;
	rect1.top=0;
	rect1.bottom=1;

	rect2.left=x;
	rect2.right=x+1;
	rect2.top=y;
	rect2.bottom=y+1;

	ZeroMemory(&blit, sizeof(blit));
	blit.dwSize = sizeof(blit);

	src->Blt(&rect2, PenBuffer, &rect1, DDBLT_ASYNC, &blit);
}
#endif

// Local Variables: //
// tab-width:4 //
// comment-column:32 //
// fill-column:100 //
// End: //
